#ifndef __IO_UTILITIES__
#define __IO_UTILITIES__

/*! \file IOUtilities.h
 *  \brief Header File for Input/Output Utility Routines
 */

#include <string>
#include <set>

#include "VectorUtilities.h"
#include "ManipVector.h"

//! Training Configuration
class CTrainingConfig
{
public:
  string name;	//!< Name of the Training Configuration
  INT_VCTR_1D index; //!< Index of the Individuals (or the Markers) used for the Training Phase

  void Print();
  int size();
};

using namespace std;

//! Saving Genotype Data to a STRUCTURE Format File
/*! \param bbase boolean=true if user chose to assign some individuals to the base cluster (Kth cluster)
 *  \param isBase isBase[i]=1 if individual i is compulsorily assigned to the base cluster (if bbase=false isBase[i]=0 for all i)
 *  \param Xtild Environmental Data
 *  \param X Genotype Data
 *  \param File The Output File Name
 *  \return Result of the Action (Success/Failure)
 *  \author Chibiao Chen
 */
bool SaveData(const bool bbase,const INT_VCTR_1D& isBase, const DBL_VCTR_2D& Xtild, const INT_VCTR_3D& X, const string& File);

//! Loading Genotype and Covariates Data
/*! \param bbase boolean=true if user chose to assign some individuals to the base cluster (Kth cluster)
 *  \param isBase isBase[i]=1 if individual i is compulsorily assigned to the base cluster (if bbase=false isBase[i]=0 for all i)
 *  \param C Coordinate of Individuals
 *  \param Xtild Environmental Data
 *  \param XQual Qualitative Variables (eg Language, Habitat)
 *  \param LevelPopSize Number of Individuals in Each Class of Each Qualitative Variable
 *  \param nQualLevels
 *  \param X Genotype Data
 *  \param recAllele
 *  \param ER Number of Extra Rows
 *  \param EC Number of Extra Columns
 *  \param File Input Data
 *  \param specialFormat If true, individuals are stored in "one line per individual" format
 *  \param nbQuantVar Number of Quantitative Variable (other than spatial coordinates)
 *  \param nb_var_regr Number of Spatial Coordinates used for the Trend(0,1,2)
 *  \param isRecAllele
 *  \param trendDegree Degree of the Spatial Trend (if -1 data do NOT contained spatial coord.)
 *  \return Result of the Action (Success/Failure)
 *  \author Durand Eric, Jay Flora 
 */

bool LoadData(const bool bbase, INT_VCTR_1D& isBase,DBL_VCTR_2D& C, DBL_VCTR_2D& Xtild, INT_VCTR_2D& XQual, INT_VCTR_2D& LevelPopSize, INT_VCTR_1D& nQualLevels, INT_VCTR_3D& X, INT_VCTR_1D& recAllele, int ER, int EC, const string& File, const bool& specialFormat, const int& nbQuantVar, const int& nb_var_regr, const bool& isRecAllele, const int& trendDegree);

//! To Interpret Input Data As Qualitative Covariate Variables
/*! 
 *  \author Jay Flora
 */

bool TransformQualData(INT_VCTR_2D& XQual, DBL_VCTR_2D& XQuant, DBL_VCTR_2D& Xtild, INT_VCTR_1D& nQualLevels, INT_VCTR_2D& LevelPopSize);

//! Saving Configurations for "Continuing Previous Run"
/*! \param z Membership of Individuals (Hard Clustering)
 *  \param p Estimated Allele Frequencies
 *  \param File The Output File Name
 *  \return Result of the Action (Success/Failure)
 *  \author Chibiao Chen
 */
bool SaveCF(const INT_VCTR_1D& z, const DBL_VCTR_3D& p, const string& File);

//! Loading Configurations for "Continuing Previous Run"
/*! \param z Membership of Individuals (Hard Clustering)
 *  \param p Estimated Allele Frequencies
 *  \param File The Input File Name
 *  \return Result of the Action (Success/Failure)
 *  \author Chibiao Chen
 */
bool LoadCF(INT_VCTR_1D& z, DBL_VCTR_3D& p, const string& File);


//! Saving an object of type DBL_VTR_2D in a File
/*! \param V Vector
 *  \param File The Output File Name
 *  \author Jay Flora
*/
void SaveToMyFile(const DBL_VCTR_2D& V, const string& File);
//! Saving an object of type INT_VTR_2D in a File
/*! \param V Vector
 *  \param File The Output File Name
 *  \author Jay Flora
*/
void SaveToMyFile(const INT_VCTR_2D& V, const string& File);
//! Saving an object of type DBL_VTR_1D in a File
/*! \param V Vector
 *  \param File The Output File Name
 *  \author Jay Flora
*/
void SaveToMyFile(const DBL_VCTR_1D& V, const string& File);
//! Saving an object of type INT_VTR_1D in a File
/*! \param V Vector
 *  \param File The Output File Name
 *  \author Jay Flora
*/
void SaveToMyFile(const INT_VCTR_1D& V, const string& File);

//! Saving the Posterior Distribution Regression Coefficients in a File (no more than 1000 sweeps are kept)
/*! \param Beta Vector containing the Regression Coefficients
 *  \param File The Output File Name
 *  \author Jay Flora
*/
void SaveBetaToMyFile(const DBL_VCTR_3D& Beta, const string &File, const int BurnInPeriod);

//! Saving an object of type DBL_VTR_2D in a File in a Format compatible with CLUMPP
/*! \param V Vector
 *  \param File The Output File Name
 *  \param header boolean: Write or not an header (title + cluster labels)
 *  \param title
 *  \author Jay Flora
*/
void SaveToMyFileClumppFormat(const DBL_VCTR_2D& V, const string& File, const bool header=false, const string title="");

//! Writting an object of type DBL_VTR_2D in an Output File Stream in a Format compatible with CLUMPP
/*! \param V Vector
 *  \param ofs Output File Stream 
 *  \author Jay Flora
*/
void WriteInOfstreamClummpFormat(const DBL_VCTR_2D V, ofstream& ofs);


// Load Index for Training Data 
bool LoadTraining(vector<CTrainingConfig> &vTrainingConfig,vector<CTrainingConfig> &vCompTrainingConfig, bool &indivValidation, const int N, const int L, const string &File);



//bool LoadDataFromIndivTrainingConfig(const bool bbase, INT_VCTR_1D& isBase, DBL_VCTR_2D& Xtild, INT_VCTR_3D& X, INT_VCTR_2D& QualVar, INT_VCTR_2D& LevelPopSize, INT_VCTR_1D& isBase_valid, DBL_VCTR_2D& Xtild_valid, INT_VCTR_3D& X_valid, INT_VCTR_2D& QualVar_valid, INT_VCTR_2D& LevelPopSize_valid, const INT_VCTR_1D trainingIndiv,int ER, int EC, const string& File);

//void LoadOneIndividualData(const bool bbase, INT_VCTR_1D& isBase, DBL_VCTR_2D& Xtild, INT_VCTR_3D& X, INT_VCTR_2D& QualVar, INT_VCTR_2D& LevelPopSize, int EC, ifstream &ifs, int i);

//bool LoadDataFromLociTrainingConfig(const bool bbase, INT_VCTR_1D& isBase, DBL_VCTR_2D& Xtild, INT_VCTR_3D& X, INT_VCTR_3D& Xval, INT_VCTR_2D& QualVar, INT_VCTR_2D& LevelPopSize, const INT_VCTR_1D trainingLoci, int ER, int EC, const string& File);




bool LoadQualInfoFromData(const int N, const int nbQualVar, const int nbQuantVar, const int nbLoci, const int A, const int ER, const int EC, const bool& specialFormat, const bool& bbase, const bool& isRecessiveAllele, const bool& hasSpatialCoord, vector< set< int > > & levelS,  const string& File );

bool LoadNewData(DBL_VCTR_2D& C, DBL_VCTR_2D& Xtild, int ER, int EC, const int& nbQuantVar, const bool& hasSpatialCoord,  vector< set< int > > levelS, const string& File);



#endif//__IO_UTILITIES__
