#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <fstream>
#include <iostream>
#include "MatrixUtilities.h"

void DimensionErrorMsg() {
  printf("Error in dimensions, most likely due to a mismatch between the input datafile format and command line parameters. \nPlease check whether you have correctly input:\n");
  printf("%s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n",
	 "Number of Individuals",
	 "Number of Dummy Individuals",
	 "Number of Loci",
	 "Number of Qualitative Covariates",
	 "Number of Quantitative Covariates (except spatial coordinates)",
	 "Presence or Absence of Spatial Coordinates",
	 "Number of Extra Columns",
	 "Number of Extra Rows",
	 "Ploidy",
	 "Special Format Option for Data Stored in 1 Line per Individual instead of 2");
  printf( "For more details: ./popsAdm or ./pops \n");
}

void choldc(DBL_VCTR_2D& matrix, const DBL_VCTR_2D mat){
  int i,j,k;                         /* loop variables */
  double sum;                         /* Temp variable */
  int n=mat.size();
  if (n!=(int)mat[0].size()){
    printf("Error in cholewsky : problem in matrix dimension");
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  if ( ((int)matrix.size()!=0) && (((int)matrix.size()!=n)||((int)matrix[0].size()!=n)) ){
    printf("choldc : error, cholesky matrix is already initialized with wrong dimensions. Exiting...\n");
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  CopyVctr(mat,matrix);
  for( i=0; i<n; i++ )
    {
      for( j=i; j<n; j++ )
	{
	  sum = matrix[i][j];
	  k = i;
	  while ( --k >= 0 ){
	    sum -= matrix[i][k] * matrix[j][k];
	  }
	  if ( i==j )
	    {
	      if (sum <= 0.0)
		{
		  printf("Choldc Failed! Exiting...\n");
		  DimensionErrorMsg();
		  exit(EXIT_FAILURE);
		}
	      else 
		{
		  matrix[i][i] = sqrt( sum );
		  }
	    }
	  else
	    {
	      matrix[j][i] = sum / matrix[i][i];
	    }
	}
    }
  for(i=0; i<n; i++ ){
    for(j=i+1; j<n; j++){
      matrix[i][j]=0;     
    }
  }
}

void Mult(DBL_VCTR_1D& Res, const DBL_VCTR_2D& A, const DBL_VCTR_1D& B){
  // A%*%B
  double sum;
  int i,j,k;
  if (A[0].size()!=B.size()){
    cout<<"Mult: Matrices have inadequate dimensions to be multiplied. Exiting..."<<endl;
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  int L=(int)A.size();
  if ((int)Res.size()==0){
    // Res was not initialized
    for (k=0; k<L; k++)
      Res.push_back(0.0);
  }   
  else if((int)Res.size()!=L){
    cout<<"Mult: Res matrix has inadequate dimensions to contain the result of the matrices multiplication. Exiting..."<<endl;
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  for (i=0; i<L; i++){
    sum=0.0;
    for (j=0; j<(int)B.size(); j++){
      sum+=A[i][j]*B[j];
    }
    Res[i]=sum;
  }
}
void Mult(DBL_VCTR_2D& Res, const DBL_VCTR_2D& A, const DBL_VCTR_2D& B){
  // A%*%B
  int lig,col,tmp;
  if (A[0].size()!=B.size()){ 
    cout<<A[0].size()<<" vs "<<B.size()<<endl;
    cout<<"Mult: Matrices have inadequate dimensions to be multiplied. Exiting..."<<endl;
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  int L=(int)A.size();
  int C=(int)B[0].size();
  int T=(int)B.size();
  if ((int)Res.size()==0){
    // Res was not initialized
    InitVctr2D(Res,L,C,0.0);
  }else if ( ((int)Res.size()!=L) || ((int)Res[0].size()!=C) ){
    cout<<"Mult: Res matrix has inadequate dimensions to contain the result of the matrices multiplication. Exiting..."<<endl;
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  for (lig=0; lig<L; lig++){
    for (col=0; col<C; col++){
      Res[lig][col]=0.0;
      for (tmp=0; tmp<T; tmp++){
	Res[lig][col]+=A[lig][tmp]*B[tmp][col];
      }
    }
  }  
}
void Mult_t(DBL_VCTR_2D& Res, const DBL_VCTR_2D& A, const DBL_VCTR_2D& B){
  // A%*%t(B)
  int lig,col,tmp;
  if (A[0].size()!=B[0].size()){ //on multiplie avec la ligne puisqu'on veut la transposee
    cout<<A[0].size()<<" vs "<<B[0].size()<<endl;
    cout<<"Mult_t: Matrices have inadequate dimensions to be multiplied. Exiting..."<<endl;
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  int L=(int)A.size();
  int C=(int)B.size();
  int T=(int)B[0].size();
  if ((int)Res.size()==0){
    // Res was not initialized
    InitVctr2D(Res,L,C,0.0);
  }else if ( ((int)Res.size()!=L) || ((int)Res[0].size()!=C) ){
    cout<<"Mult_t: Res matrix has inadequate dimensions to contain the result of the matrices multiplication. Exiting..."<<endl;
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
    //return;
  }

    for (lig=0; lig<L; lig++){
      for (col=0; col<C; col++){
	Res[lig][col]=0.0;
      for (tmp=0; tmp<T; tmp++){
	Res[lig][col]+=A[lig][tmp]*B[col][tmp];
      }
    }
  }  
}
void Mult_t_left(DBL_VCTR_2D& Res, const DBL_VCTR_2D& A, const DBL_VCTR_2D& B){
  // t(A)%*%B
  int lig,col,tmp;
  if (A.size()!=B.size()){ 
    cout<<A.size()<<" vs "<<B.size()<<endl;
    cout<<"Mult_t_left: Matrices have inadequate dimensions to be multiplied. Exiting..."<<endl;
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  int L=(int)A[0].size();
  int C=(int)B[0].size();
  int T=(int)B.size();
  if ((int)Res.size()==0){
    // Res was not initialized
    InitVctr2D(Res,L,C,0.0);
  }else if ( ((int)Res.size()!=L) || ((int)Res[0].size()!=C) ){
    cout<<"Mult_t_left: Res matrix has inadequate dimensions to contain the result of the matrices multiplication. Exiting..."<<endl;
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  for (lig=0; lig<L; lig++){
    for (col=0; col<C; col++){
      Res[lig][col]=0.0;
      for (tmp=0; tmp<T; tmp++){
	Res[lig][col]+=A[tmp][lig]*B[tmp][col];
      }
    }
  }  
}

void Inverse(DBL_VCTR_2D& res, const DBL_VCTR_2D& A){
  if (A.size()!=A[0].size()){
    printf("inverse : error, matrix is not square. Exiting...\n");
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  int n=(int)A.size();
  //printf("dimensions n=%d\n",n);
  int i,j,k;
  DBL_VCTR_2D MM,M1,Mi1;
  CopyVctr(A,MM);
  CopyVctr(A,M1);
  //Mi1 and res initialized with identity matrix 
  InitVctr2D(Mi1,n,n,0.0);
  for (i=0; i<n; i++){
    Mi1[i][i]=1.0;
  }
  CopyVctr(Mi1,res);
  for (i=0;i<n;i++){
    for (j=0;j<n;j++){
      if (MM[i][i]==0){
	printf("inverse : inversion with gauss method impossible, division by 0\n");
	DimensionErrorMsg();
	exit(EXIT_FAILURE);
      }
      M1[i][j]=MM[i][j]/MM[i][i];
      Mi1[i][j]=res[i][j]/MM[i][i];
    }
    //printf("MM.size() %d", (int)MM.size());
    CopyVctr(M1,MM);
    //printf("res.size() %d", (int)res.size());
    CopyVctr(Mi1,res);
    for (k=0;k<n;k++)// mise zros de la colonne
      if (k!=i)
	for (j=0;j<n;j++){
	  M1[k][j]=MM[k][j]-MM[i][j]*MM[k][i];
	  Mi1[k][j]=res[k][j]-res[i][j]*MM[k][i];
	}
    CopyVctr(M1,MM);
    CopyVctr(Mi1,res);    
  }
  //printf("dim de res: %d, %d\n",(int)res.size(),(int)res[0].size());
}
void Transpose(DBL_VCTR_2D& res, const DBL_VCTR_2D& A){
  int i,j;
  int L =(int)A.size();
  int C=(int)A[0].size();
  if ((int)res.size()==0){
    InitVctr2D(res,C,L,0.0);
  }else if ((int)res.size()!=C || (int)res[0].size()!=L){
    printf("Transpose : wrong dimensions. Exiting...\n");
    DimensionErrorMsg();
    exit(EXIT_FAILURE);
  }
  for (i=0; i<C; i++){
    for (j=0; j<L; j++){
      res[i][j]=A[j][i];
    }
  }
}

double FindBound(const DBL_VCTR_1D& Wi, int k){
  //search for max(W[i,-k],0)
  int j;
  double max=0.0;
  for (j=0; j<(int)Wi.size(); j++){
    if ((j!=k) && (Wi[j]>max))
      max=Wi[j];
  }
  return(max);
}


void InverseWithChol(DBL_VCTR_2D& res, const DBL_VCTR_2D& A)
{
    if (A.size()!=A[0].size())
    {
          printf("inverse : error, matrix is not square. Exiting...\n");
          DimensionErrorMsg();
	  exit(EXIT_FAILURE);
    }
    int N=(int)A.size();

    if ((int)res.size()==0){
      InitVctr2D(res,N,N,0.0);
    } else if ((int)res.size()!=N || (int)res[0].size()!=N ){
      printf("InverseWithChol : wrong dimensions for Initialized matrix res. Exiting...\n");
      DimensionErrorMsg();
      exit(EXIT_FAILURE);
    }
    gsl_matrix* gslMat = gsl_matrix_alloc(N,N);
    gsl_matrix* gslResult = gsl_matrix_alloc(N,N);
    gsl_matrix_set_identity(gslResult);
    gsl_vector* gslVect = gsl_vector_alloc(N);
    int i,j;
    //copy A to gslMat
    for (i = 0; i < N; i++)
      {
	for (j = 0; j < N; j++)
	  {
	    double x = A[i][j];
	    gsl_matrix_set(gslMat,i,j,x);
	  }
      }
    gsl_linalg_cholesky_decomp(gslMat);
    for (i = 0; i < N; i ++)
      {
	for (j = 0; j < N; j ++)
	  {
	    gsl_vector_set(gslVect,j,gsl_matrix_get(gslResult,i,j));
	  }
	gsl_linalg_cholesky_svx(gslMat, gslVect);
	for (j = 0; j < N; j ++)
	  {
	    gsl_matrix_set(gslResult,i,j,gsl_vector_get(gslVect,j));
	  }
      }
    //Copy gslMat to result
    for (i = 0; i < N; i++)
      {
	for (j = 0; j < N; j ++)
	  {
	    res[i][j] = gsl_matrix_get(gslResult,i,j);
	  }
      }  
    gsl_matrix_free(gslMat);
    gsl_vector_free(gslVect);
    gsl_matrix_free(gslResult);
    
}
