#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <fstream>
#include <iostream>
#include <sstream>

#include <limits>

#include "StatUtilities.h"



/********************************************************
 ***                  STATS UTILITIES                 ***
 ********************************************************/

void MultivariateGauss(DBL_VCTR_1D& mu, DBL_VCTR_2D& v_chol, DBL_VCTR_1D& Y, gsl_rng*& rng){
  DBL_VCTR_1D X; 
  int k,i,j;
  int n=v_chol.size();
  const double infini = numeric_limits< double >::infinity();
  //std::numeric_limits< double >::infinity();

  for(k=0; k<n; k++)
    X.push_back(gsl_ran_ugaussian(rng));
  
  if ((int)Y.size()==0){
    // Y has not been initialized yet
    for (i =0;i<n;i++){
      Y.push_back(mu[i]);
      for (j=0;j<n;j++){
	Y[i]+=v_chol[i][j]*X[j];
      }
    }
  }else{
    for (i =0;i<n;i++){
      Y[i]=mu[i];
      for (j=0;j<n;j++){
	Y[i]+=v_chol[i][j]*X[j];
      }
    }
  }
  if (Y[0]==infini)
    {
      cout<<"Y[0] == infini in Multivariate Gaussian"<<endl;
      PrintVctr(mu);
      PrintVctr(X);
      PrintVctr(v_chol);
    }
}
double TruncGauss(double mu, double sigma, double lower, double upper, gsl_rng*& rng){
  double FU,FL;
  if (upper<lower) {
    cout<<"Error lower bound is greater than upper bound in TruncGauss. Exiting..."<<endl;
    exit(1);
  }
  FL=gsl_cdf_ugaussian_P((lower-mu)/sigma);
  FU=gsl_cdf_ugaussian_P((upper-mu)/sigma);
  return (mu+sigma*gsl_cdf_ugaussian_Pinv(gsl_rng_uniform(rng)*(FU-FL)+FL));
}
double TruncGauss_Left(double mu, double sigma, double lower, gsl_rng*& rng){
  // ie upper =  + Inf => FU==1
  double FL;
  FL=gsl_cdf_ugaussian_P((lower-mu)/sigma);
  return (mu+sigma*gsl_cdf_ugaussian_Pinv(gsl_rng_uniform(rng)*(1-FL)+FL));
}
double TruncGauss_Right(double mu, double sigma, double upper, gsl_rng*& rng){
  // ie lower = - Inf => FL==0
  double FU;
  FU=gsl_cdf_ugaussian_P((upper-mu)/sigma);
  return ( mu + sigma*gsl_cdf_ugaussian_Pinv(gsl_rng_uniform(rng)*FU) );
}

void SimulateXtild(DBL_VCTR_2D& Xtild, int N, int beta_length, gsl_rng*& rng){
  int i,j;
  if ((int)Xtild.size()!=0){
    printf("Xtild exists already. No new simulation.\n");
    return;
  }
  // the first colonn contains 1 (for intercept)
  InitVctr2D(Xtild,N,beta_length,1.0);
  for (i=0; i<N; i++){
    for (j=1; j<beta_length; j++){
      //Xtild[i][j]=gsl_rng_uniform(rng);
      //Xtild[i][j]=-5+10*gsl_rng_uniform(rng); //runif(-5,5)
      Xtild[i][j]=-1+2*gsl_rng_uniform(rng); //runif(-1,1)
    }
  }
}

void MultiProbitModel(INT_VCTR_1D& Z, DBL_VCTR_2D& Xtild, DBL_VCTR_2D& Beta_vrai, int M, int base, gsl_rng*& rng){
  int i,j,k,ind;
  double max;
  int P=(int)Beta_vrai[0].size();
  int K=(int)Beta_vrai.size()+1;
  DBL_VCTR_2D Pred;
  if ((int)Z.size()>0){
    printf("Warning : Z has already been initialized. Exit from MultiProbitModel...");
    return;
  }
  if ((int)Xtild.size()==0){
    // we have to simulate a matrix Xtild
    // the first colonn contains 1 (for intercept)
    InitVctr2D(Xtild,M,P,1.0);
    for (i=0; i<M; i++){
      for (j=1; j<P; j++){
	//Xtild[i][j]=gsl_rng_uniform(rng);
	//Xtild[i][j]=-5+10*gsl_rng_uniform(rng); //runif(-5,5)
	Xtild[i][j]=-1+2*gsl_rng_uniform(rng); //runif(-1,1)
      }
    }
  }
  InitVctr2D(Pred,M,K-1,0.0);
  Mult_t(Pred,Xtild,Beta_vrai); //Xtild%*%t(Beta_vrai)
  for (i=0; i<M; i++){
    for (k=0; k<(K-1); k++){
      Pred[i][k]+=gsl_ran_ugaussian(rng);
    }
  }
  for (i=0; i<M;i++){
    max=0;
    ind=base; //Z[i] <= base if max<0 
    for (k=0; k<base; k++){
      if (Pred[i][k]>max){
	max=Pred[i][k];
	ind=k;
      }
    }
    for (k=base; k<K-1; k++){
      if (Pred[i][k]>max){
	max=Pred[i][k];
	ind=k+1; //the real class label is k+1 because we have to count the base class
      }
    }
    Z.push_back(ind);
  }
}


// return a string of stars to print (stars number will represent the "significance level")
string Significance(const double* V, const int size) {

  if ((0>=quantile(V,size,.05)) && (0<=quantile(V,size,.95))) {
    //cout << quantile(V,size,.05) << " to " << quantile(V,size,.95)<< endl;
    return "no";  //0 in IC 90%
  } else if ((0>=quantile(V,size,.025)) && (0<=quantile(V,size,.975))) {
    //cout << quantile(V,size,.025) << " to " << quantile(V,size,.975)<< endl;
    return "*";  //0 in IC 95 % but not in IC 90%
  } else if ((0>=quantile(V,size,.005)) && (0<=quantile(V,size,.995))) {
    //cout << quantile(V,size,.005) << " to " << quantile(V,size,.995)<< endl;
    return "**";  //0 in IC 99% but not in IC 95%
  } else if ((0>=quantile(V,size,.0005)) && (0<=quantile(V,size,.9995))) {
    //cout << quantile(V,size,.0005) << " to " << quantile(V,size,.9995)<< endl;
    return "***";  //0 in IC 99.9% but not in IC 99%
  } else {
    //cout << "0 not in " << quantile(V,size,.0005) << " to " << quantile(V,size,.9995)<< endl;
    return "****";  //0 not in IC 99.9%
  }

}

// return confidence interval
string CI(const double* V, const int size, const double inf, const double sup) {
  stringstream infCI,supCI;
  infCI << quantile(V,size,inf);
  supCI << quantile(V,size,sup);
  string stringCI = "[ " + infCI.str() + " , " + supCI.str() + " ]";
  return stringCI;

}

double Correlation(const DBL_VCTR_2D& X, const DBL_VCTR_2D& Y) {
  
  // since the sum of each row is 1
  // the average value of a matrix is 1*N/(k*N) ie 1/K
  if ( (X.size()!=Y.size()) ||(X[0].size()!=Y[0].size()) ){
    cout<<"Correlation: Matrices have inadequate dimensions to be multiplied. Exiting..."<<endl;
    exit(1);
  }

  int N = static_cast<int>(X.size());
  int K = static_cast<int>(X[0].size());
  double propMean = N/K; // mean*sampleSize
  double cov = 0; // cov = N*cov(X,Y)
  double varX = 0;
  double varY = 0;
  int i,k;

  for (i=0; i <N; i++) {
    for (k=0; k<K; k++) {
      cov += X[i][k]*Y[i][k];
      varX += X[i][k]*X[i][k];
      varY += Y[i][k]*Y[i][k];
    }
  }
  cov -= propMean; 
  varX -= propMean; 
  varY -= propMean; 
  
   return(cov / (sqrt(varX)*sqrt(varY)));
  
}
