
/*! \file pops.cpp
*  \brief The POPS MCMC Algorithms
*/

#include <algorithm>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_log.h>
#include <gsl/gsl_sf_gamma.h>

//#include "windows.h"
#include <string.h>
#include <sstream>
#include <iostream>
#include <set>
#include <string>
#include <limits>

#include "IOUtilities.h"
#include "PlotUtilities.h"
#include "RNGUtilities.h"
#include "VectorUtilities.h"
#include "MiscUtilities.h"
#include "StatUtilities.h"
#include "ranker.h"
#include "MatrixUtilities.h"
#include "ManipVector.h"
#include "PredictUtilities.h"



//! The Negative Limit of Logarithm (double)
#define N_LIMIT (-707.0)

//! The Positive Limit of Logarithm (double)
#define P_LIMIT (+707.0)

//! Sweep Parameters
class CSweepParams
{
public:
	INT_VCTR_1D z;	//!< Cluster Label of Individuals
	DBL_VCTR_3D p;	//!< Allele Frequencies
	DBL_VCTR_2D w; //!< Latent Variables in the Probit Model
	DBL_VCTR_2D beta; //!< Coefficients of Regression
};

//! Criteria of a Run
class CRunCriteria
{
public:

	int KEst;
	double dAL;
	double dAL_valid;
	double DIC;
	double DIC2;
	DBL_VCTR_1D dAIC;
	DBL_VCTR_1D DICregr;
	DBL_VCTR_1D DICregr2;
	double DIC2_tot;
	double BF;
	double LHmax;
	double cor;

	//void Print();

};

/*void CRunCriteria::Print() 
{

}*/


using namespace std;

//! Showing the Usage of the Program
/*! The compulsory/required options are indicated by upper case characters; the optional
*  options are indicated by lower case characters.
*  \author Chibiao Chen and Eric Durand and Flora Jay
*/
void ShowUsage()
{
        cout << "To use PREDICTION UTILITIES for NEW values of covariates, please run pops -predict"<< endl ;
	
	cout << "Here, we explain command-line usage for classical POPS utilities."<< endl;
	cout << "Options can be specified in any order." << endl;
	cout << endl;

	cout << "Typical Usages:" << endl;
	cout << "     * No Validation and No Covariates:\n     pops -FF -NN -AA -LL -KK -DD -SS -BB -T-1 -X0 -XL0" << endl;
	cout << endl;
	cout << "     * No Validation and Covariates:\n     pops -FF -NN -AA -LL -KK -DD -SS -BB -TT -XX -XLXL" << endl;
	cout << endl;
	cout << "     * Split-Sample Validation:\n     pops -FF -vv -NN -AA -LL -KK -DD -SS -BB -TT -XX -XLXL" << endl;
	cout << endl;
	cout << "     * 2-fold Cross-Validation:\n     pops -FF -vv -2fold -NN -AA -LL -KK -DD -SS -BB -TT -XX -XLXL" << endl;
	cout << endl;
	cout << endl;

	cout << "Required Parameters:" << endl;
	cout << "-FF: F = File Name of Input Data File" << endl;
	cout << "         If file name contains spaces, input it as \"F\"." << endl;
	cout << "-NN: N = Number of Individuals" << endl;
	cout << "-AA: A = Ploidy (1 = Haploid, 2 = Diploid, ...)" << endl;
	cout << "-LL: L = Number of Loci" << endl;
	cout << "-KK: K = Number of Clusters" << endl;
	cout << "-XLXL: XL = Number of Qualitative Covariates" << endl;
	cout << "-XX: X = Number of Quantitative Covariates (other than spatial coordinates)" << endl;
	cout << "-TT :    T = Degree of Trend" << endl;
	cout << "             -1: NO Spatial Coord. in datafile F" << endl; 
	cout << "              0: Spatial Coord. NOT used" << endl;
	cout << "             {1,2,3}: Degree 1, 2 or 3" << endl;
	cout << "-DD: D = Parameter of Dirichlet Allele Frequency Model" << endl;
	cout << "-SS: S = Total Number of Sweeps of MCMC" << endl;
	cout << "-BB: B = Burn In Number of Sweeps of MCMC" << endl;
	cout << endl;
	cout << "Optional Parameters:" << endl;
	cout << "-rr:     r = Number of Extra Rows in Data File, Default: 0" << endl;
	cout << "-cc:     c = Number of Extra Columns in Data File, Default: 0" << endl;
	cout << "-ii:     i = Folder Name of Input Data File, Default: Current Folder" << endl;
	cout << "             If folder name contains spaces, input it as \"i\"." << endl;
	cout << "-oo:     o = Folder Name of Output Result Files, Default: Current Folder" << endl;
	cout << "             If folder name contains spaces, input it as \"o\"." << endl;
	cout << "-orunorun: orun = String to Append to Output Result Files Names, e.g. a Run Number" << endl;
	cout << "-spsp: sp = SPecial data format: one individual = one row" << endl;
	cout << "       sp = y, Yes" << endl;
	cout << "       sp = n, No (Default)" << endl;
	cout << "-gg -hh: Launch multiple runs for file F. Outfiles will be indexed from _RUNg to _RUNh" <<endl;
	cout << "-vv:     v = Partition File Name for Split-Sample Validation (no option -2fold) or Cross-Validation (option -2fold)." << endl;
	cout << "             Should contained the Loci Index (or the Individuals Index) for the Training Data" << endl;
	cout << "             Not compatible with sp format."   << endl;         
	cout << endl;
	cout << "Options:" << endl;
	cout << "-2fold : Run with 2-fold cross-validation. A Partition File must be specified with the option -v." << endl;
	cout << "         Not compatible with sp format." << endl;
	cout << "-ref :   The 1st non-extra column of datafile F indicates Individuals to assigned to the Reference Cluster" << endl;
	cout << "         (value 1 for ref. indiv., 0 for others)" << endl;
}




//! Translating the XY Coordinates
/*! Translate the XY coordinates to all positive values to circumvent the quirks of
*  ChartDirector in the XYChart.
*  \param C Coordinates To Be Translated/Translated (Input & Output)
*  \return Result of the Action (Success/Failure)
*  \author Chibiao Chen
*/
bool TranslateCoordinates(DBL_VCTR_2D& C)
{
	int N = int(C.size());
	if (N == 0 || C[0].size() != 2)
	{
		return false;
	}
	double xMin = +1.0E+300;
	double xMax = -1.0E+300;
	double yMin = +1.0E+300;
	double yMax = -1.0E+300;
	int i;
	for (i = 0; i < N; i++)
	{
		if (C[i][0] < xMin)
		{
			xMin = C[i][0];
		}
		if (C[i][0] > xMax)
		{
			xMax = C[i][0];
		}
		if (C[i][1] < yMin)
		{
			yMin = C[i][1];
		}
		if (C[i][1] > yMax)
		{
			yMax = C[i][1];
		}
	}
	double xOffset = -xMin + (xMax - xMin) * 2.0;
	double yOffset = -yMin + (yMax - yMin) * 2.0;
	for (i = 0; i < N; i++)
	{
		C[i][0] += xOffset;
		C[i][1] += yOffset;
	}

	return true;
}


//! The POPS MCMC Algorithms
/*! The compulsory/required options are indicated by upper case characters; the optional
*  options are indicated by lower case characters.
*  \author Flora Jay and Eric Durand and Chibiao Chen
*/
int pops(int argc, char* argv[], string run,ofstream& ofsCriteria, CRunCriteria &runCriteria, const bool bFIT, const bool bValOnIndiv, CTrainingConfig &trainingConfig)
{

	gsl_set_error_handler_off();

	// Required Parameters
	string F;			// Name of Training Data File
	int N;				// Number of Samples/Individuals in Training data
	int NV=0;			// Number of Samples/Individuals in Validation data
	int A;				// Haploid, Diploid, ...
	int L;				// Number of Loci for Training Data
	int LV = 0;			// Number of Loci for Validation Data
	INT_VCTR_1D J;                  // Number of Distinct Alleles Observed at Every Locus
	int K;				// Number of Cluster
	int beta_length;		// Number of Regression Coefficients (in one cluster)
	double lambda;		        // Parameter of Dirichlet Allele Frequency Model
	int TotalSweep;		        // Total Number of Sweeps
	int BurnIn=0;			// Burn In Number of Sweeps
	int trendDegree = -1;           //Degree of trend (default, no spatial coordinates)
	int nbQuantVar=0;               // Number of Quantitative Covariates other than Spatial Coordinates
	int nbQualVar=0;			// Number of Qualitative Covariates


	// Optional Parameters
	int ER = 0;			// Number of Extra Rows
	int EC = 0;			// Number of Extra Columns
	string FoI;			// Folder of Input Data File
	string FoO;			// Folder of Output Result Files
	int BurnInStructure=0;          // Number of Sweeps using No Covariates (basic STRUCTURE model), will be fixed if not specified
	INT_VCTR_1D nQualLevels;        // Number of Categories for each Qualitative Variable	

	bool isRecAllele = false;       // Is there recessive alleles in the data set ? For compatibility with tessAdmCov, not used in this version
	bool specialFormat = false;     // Is the data in special format? (one line per individual, even diploid). Then diploid individual: one locus = 2 columns
	int nd = 0;                     // Number of dummy individuals
	int nb_var_regr = 2;            // How many variables (latitude,longitude) in linear regressions ? For compatibility with tessAdmCov

	// Auxiliary Variables
	//m int totalNbQualLevels=0;		// Number of Categories for ALL Qualitative Variable (tocalculate the size of Xtild)

	bool bF = false;
	bool bN = false;
	bool bA = false;
	bool bL = false;
	bool bK = false;
	bool bD = false;
	bool bS = false;
	bool bB = false;
	bool bBS = false;
	bool bX = false;
	bool bT = false;
	bool bbase=false;

	if (argc < 9)
	{
	        cout << "Only " << argc-1 << " arguments given, this is not enough." << endl;
		ShowUsage();
		cout << "Only " << argc-1 << " arguments given, this is not enough." << endl;
		exit(EXIT_FAILURE);
	}
	else
	{
		// Parse the command line parameters.
		for(int ac = 1; ac < argc; ac++)
		{
			switch (argv[ac][1])
			{
			  
			case 'X':
				if (argv[ac][2]=='L')
				{
				  //m nQualLevels.push_back(atoi((const char*) &argv[ac][3]));
				  //m totalNbQualLevels += atoi((const char*) &argv[ac][3]);
				  nbQualVar = atoi((const char*) &argv[ac][3]);
				  bX = true;
				}
				else
				{
					nbQuantVar = atoi((const char*) &argv[ac][2]);
					bX = true;
				}
				break;
				
			case 'F':
				F = (const char*) &argv[ac][2];
				bF = true;
				break;

			case 'N':
				N = atoi((const char*) &argv[ac][2]);
				bN = true;
				break;
				
			case 'A':
				A = atoi((const char*) &argv[ac][2]);
				bA = true;
				break;

			case 'L':
				L = atoi((const char*) &argv[ac][2]);
				bL = true;
				break;

			case 'K':
				K = atoi((const char*) &argv[ac][2]);
				bK = true;
				break;

			case 'D':
				lambda = atof((const char*) &argv[ac][2]);
				bD = true;
				break;

			case 'S':
				TotalSweep = atoi((const char*) &argv[ac][2]);
				bS = true;
				break;

			case 'B':
				if (argv[ac][2]=='S'){ 
					BurnInStructure = atoi((const char*) &argv[ac][3]);
					bBS = true;
				}else{
					BurnIn = atoi((const char*) &argv[ac][2]);
					bB = true;
				}
				break;
				
			case 'T':
				trendDegree = atoi((const char*) &argv[ac][2]);
				bT = true;
				break;

			case 'g':
				break;
			case 'h':
				break;
			case 'v':
				break;
			case '2':
				break;
				
				
			case 'r':
			
				if (argv[ac][2] == 'e')
				// option -ref
				{
					if (argv[ac][3] == 'f')
					{
						bbase=true;
						cout<<"Individuals chosen by user will be assigned to the reference cluster"<<endl;	
					}
					else
					{
						cout << "Unknown Option - " << argv[ac] << " For more details: ./pops " << endl;
						exit(EXIT_FAILURE);
					}
				}
				else if (argv[ac][2] == 'c')
				{
					// recessive alleles ?
					if (argv[ac][3] == 'y')
					{
						isRecAllele = true;
					}
					if (argv[ac][3] == 'n')
					{
						isRecAllele = false;
					}
					if (argv[ac][3] != 'y' && argv[ac][3] != 'n')
					{
						cout << "Unknown Option - " << argv[ac] << " For more details: ./pops " << endl;
						exit(EXIT_FAILURE);
					}	
				}
				else
				{
					ER = atoi((const char*) &argv[ac][2]);

				}
				break;

			case 'c':
				EC = atoi((const char*) &argv[ac][2]);
				break;

			case 'd':
				nd = atoi((const char*) &argv[ac][2]);
				break;
				
			case 'i':
				FoI = (const char*) &argv[ac][2];
				break;

			case 'o':
			  if (argv[ac][2] == 'r' && argv[ac][3] == 'u' && argv[ac][4] == 'n') {
			    run = (const char*) &argv[ac][5];
			  } else {
			    FoO = (const char*) &argv[ac][2];
			  }
			  break;
		
			case 's':
				// Special data format
				if (argv[ac][2] == 'p')
				{
					if (argv[ac][3] == 'y')
					{
						specialFormat = true;
					}
					else if (argv[ac][3] == 'n')
					{
						specialFormat = false;
					}
					else
					{
						cout << "Unknown Option - " << argv[ac] << " For more details: ./pops " << endl;
						exit(EXIT_FAILURE);
					}
				}
				else
				{
					cout << "Unknown Option - " << argv[ac] << " For more details: ./pops " << endl;
					exit(EXIT_FAILURE);
				}
				break;
				
			case 'n':
			        nb_var_regr = atoi((const char*) &argv[ac][2]);
				break;

			
			default:  // Unknown parameters.
				cout << "Unknown Option - " << argv[ac] << " For more details: ./pops " << endl;
				exit(EXIT_FAILURE);
			}
		}
	}


	if (!bF || !bN || !bA || !bL || !bK || !bD || !bS || !bB  || !bT )
	{
	        string missingopt="";
	        if (!bF) missingopt+=" -F" ;
		if (!bN) missingopt+=" -N" ;
		if (!bA) missingopt+=" -A" ;
		if (!bL) missingopt+=" -L" ;
		if (!bK) missingopt+=" -K" ;
		if (!bD) missingopt+=" -D" ;
		if (!bS) missingopt+=" -S" ;
		if (!bB) missingopt+=" -B" ;
		if (!bT) missingopt+=" -T" ;
		cout << "\n Mandatory arguments are missing, please check details for: " << missingopt << " with pops | more " << endl;
    		exit(EXIT_FAILURE);
	}
	
	if (isRecAllele && A != 2)		// for compatibility with tessAdmCov
	{
		cout << "Dominant markers allowed for diploid individuals only ..." << endl;
		exit(EXIT_FAILURE);
	}
	
	if (trendDegree < -1 || trendDegree > 3)
	{
		cout << "The trend degree must be in {-1,0,1,2,3}." << endl;
		cout << "(value -1 if the datafile do NOT contained spatial coordinates)" << endl;
		ShowUsage();
		exit(EXIT_FAILURE);
	}



	// Expected total number of lines given command line
	int expectedLines, expectedCols;
	int rec = isRecAllele? 1:0;	
	if (!specialFormat)
	{
	        // A-ploids individuals stored on A lines.
	        expectedLines = rec + ER + A*N + nd;
	}
	else
	{
		// individuals stored on one line.
		expectedLines = rec + ER + N + nd;
	}

	// Expected total number of columns given command line
	int coord =  (trendDegree ==-1)? 0:2; //Two Columns for XY Coordinates (for popsAdm trendDegree always !=1)
	int refCol = bbase? 1:0;
	if (!specialFormat)
	{
		// One column per locus
		expectedCols = EC + refCol + coord + nbQuantVar + nbQualVar + L;
	}
	else
	{
		// A column(s) per locus for A-ploids individuals
		expectedCols = EC + refCol + coord + nbQuantVar + nbQualVar + A*L;
	}
	printf("\n**Comment:**\n According to your parameter settings, the input datafile should have *%d* lines and *%d* columns.\nIf this is not correct please check your command line again.\n\n",expectedLines,expectedCols);



	string FIFN = !FoI.empty() ? FoI + "/" + F : F;
	string FOFN = !FoO.empty() ? FoO + "/" + F : F;
	
	if (!run.empty())
	  {
	    FOFN=FOFN+"_RUN"+run+"_";
	    cout<<"RUN= "<<run<<endl;
	  }
	
	int NumOfDigits = (int) log10((double) TotalSweep) + 1;

	// Loop Index Variables
	int i, a, l, j, k, predict, var;

	//m nbQualVar = nQualLevels.size();	
	InitVctr1D(nQualLevels,nbQualVar,0);
	
	int trendSize;
	//Degree of trend
	// The intercept is NOT taken into account in trendSize
	if (trendDegree == -1)
	{
		trendSize=0;
	}
	else 
	{
		trendSize = (trendDegree +1)*(trendDegree +2)/2 - 1;
	}

	/*if (nbQuantVar==0 && nbQualVar==0 && trendDegree==-1) // no covariates
	{
		beta_length=0;
	}
	else
	{
		//m beta_length = 1 for the intercept + trend + nbQuantVar quantitative covariates + (totalNbQualLevels-nbQualVar) levels for qualitative covariates 
		beta_length = 1+trendSize + nbQuantVar + totalNbQualLevels-nbQualVar; 
	}*/


	// Load Data
	if (bFIT)
	{
		if (bValOnIndiv)
		{
			NV = N-trainingConfig.size();
			N = trainingConfig.size();
			LV=0;
			cout <<"Training Config Code: " << trainingConfig.name <<"   Nb of Individuals,   in the Training Set:" << N << "   in the Validation Set:"<< NV << endl;
		}
		else
		{
			LV = L-trainingConfig.size();
			L = trainingConfig.size();
			NV=N;
			cout <<"Training Configuration Code: " << trainingConfig.name << "   Nb of Loci,   in the Training Set:" << L << "   in the Validation Set:"<< LV << endl;

		}
	}
	// Training  Data
	INT_VCTR_1D isBase;
	DBL_VCTR_2D Xtild; // Covariates Values
	INT_VCTR_2D XQual; 	// Classes of Each Individual for Each Qual. Variable 
							// eg XQual[i][n]: class of individual i for the nth qual. variable
	INT_VCTR_2D LevelPopSize; 	// Number of Individuals in Each Class for Each Qual. Variable
								// eg LevelPopSize[n][c]: nb of individuals in the cth class of the nth qual. variable
	INT_VCTR_3D X; // Genotype Data 
	InitVctr1D(isBase,N,0);
	//m InitVctr2D(Xtild,N,beta_length,1.0);
	InitVctr2D(XQual,N,nbQualVar,0);
	//m InitVctr2D(LevelPopSize,nbQualVar,nQualLevels,0);
	InitVctr3D(X, N, A, L, 0);
	// Validation Data
	INT_VCTR_1D isBase_valid;
	DBL_VCTR_2D Xtild_valid; // Covariates Values
	INT_VCTR_2D XQual_valid; // Classes of Each Individual for Each Qual. Variable 
	INT_VCTR_2D LevelPopSize_valid; // Number of Individuals in Each Class for Each Qual. Variable
	INT_VCTR_3D XsameLoci; // Genotype in Validation Data if Validation Loci are sampled at the Same Locations as Training Loci 
	INT_VCTR_3D XnewLoci; // Genotype Data if Validation Loci Locations are New
	bool bValidation = true; // bValidation=true if there are Validation Data
	bool bSameLoci; // bSameLoci=true if Validation Loci are sampled at the Same Locations as Training Loci 
	// ie for Validation on Individuals
	INT_VCTR_1D ZCurr_valid;
	
	
	INT_VCTR_1D recAllele;						// For compatibility with tessAdmCov, not used in this version
	INT_VCTR_3D Xobs;                           // Observed genotypes ; for compatibility with tessAdmCov, not used in this version
	if (isRecAllele)
	{
		InitVctr1D(recAllele,L,0);
		InitVctr3D(Xobs,N,A,L,0);
	}

	DBL_VCTR_2D C;								// Coordinates
	InitVctr2D(C, N + nd, 2, 0.0);
	
	if (bFIT)
	{
		// Load Training and Validation Data

		bValidation=true;
		InitVctr1D(ZCurr_valid,NV,0);
		bSameLoci=bValOnIndiv;
		InitVctr1D(isBase_valid,NV,0);
		//m InitVctr2D(Xtild_valid,NV,beta_length,1.0);
		InitVctr2D(XQual_valid,NV,nbQualVar,0);
		//m InitVctr2D(LevelPopSize_valid,nbQualVar,nQualLevels,0);
		if (bValOnIndiv) {
			InitVctr3D(XsameLoci, NV, A, L, 0);
			//q LoadDataFromIndivTrainingConfig(bbase, isBase, Xtild, X, XQual, LevelPopSize, isBase_valid, Xtild_valid, XsameLoci, XQual_valid, LevelPopSize_valid, trainingConfig.index, ER, EC, FIFN);
		} 
		else
		{
			InitVctr3D(XnewLoci, N, A, LV, 0);
			//q LoadDataFromLociTrainingConfig(bbase, isBase, Xtild, X, XnewLoci, XQual, LevelPopSize, trainingConfig.index, ER, EC, FIFN);
			Xtild_valid = Xtild;
			isBase_valid = isBase;
			XQual_valid = XQual;
			LevelPopSize_valid = LevelPopSize;
		}
	}
	else
	{
		// No Validation Data
		// Load Data
	  LoadData(bbase, isBase, C, Xtild, XQual, LevelPopSize, nQualLevels, X, recAllele, ER, EC, FIFN, specialFormat, nbQuantVar, nb_var_regr, isRecAllele, trendDegree);
	  //cout << "pops. xtild:" << endl;
	  //PrintVctr(Xtild);
	  bValidation = false;
	  bSameLoci = true;
	}

	if (Xtild.size() == 0)
	  // if no covariates, (Xtild uninitialized in this case)
	  {
	    beta_length = 0;
	  } 
	else 
	  {
	    beta_length=Xtild[0].size();
	  }
	
	if (trendDegree != -1) 
	{	
	  TranslateCoordinates(C);
	}


	string test = "test.txt";
	string test_valid = "test_valid.txt";
	string test_new = "test_new.txt";

	//SaveData(bbase,isBase,Xtild,X,test);
	
	if (bValidation)
	{
		if (bSameLoci)
		{
		  //SaveData(bbase,isBase_valid,Xtild_valid,XsameLoci,test_valid);
		}
		else
		{
		  //SaveData(bbase,isBase,Xtild,XnewLoci,test_new);
		}
	}

	// Determine the actual J from the Genotype Data .
	int LN = L+LV;
	vector< set< int > > alleleS;
	set< int > emptyS;
	for (l = 0; l < LN; l++)
	{
		alleleS.push_back(emptyS);
	}


	// Training Loci Locations
	for (l = 0; l < L; l++)
	{
		for (a = 0; a < A; a++)
		{
			for (i = 0; i < N; i++)
			{
				alleleS[l].insert(X[i][a][l]);
			}
			for (i = 0; i < static_cast<int>(XsameLoci.size()); i++)
			{
				alleleS[l].insert(XsameLoci[i][a][l]);
			}			
		}
	}

	// Eventual New Locations (from Validation Data)
	for (l = 0; l < LV; l++)
	{
		for (a = 0; a < A; a++)
		{
			for (i = 0; i < static_cast<int>(XnewLoci.size()); i++)
			{
				alleleS[L+l].insert(XnewLoci[i][a][l]);
			}
		}
	}


	INT_VCTR_2D alleleV;
	INT_VCTR_1D emptyV;
	for (l = 0; l < LN; l++)
	{
		alleleV.push_back(emptyV);
	}
	for (l = 0; l < LN; l++)
	{
		set< int >::const_iterator csiB = alleleS[l].begin();
		set< int >::const_iterator csiE = alleleS[l].end();
		// Discard the missing allele.
		// Missing allele is represented by a negative integer.
		if (*csiB < 0)
		{
			++csiB;
		}
		set< int >::const_iterator csi;
		for (csi = csiB; csi != csiE; ++csi)
		{
			alleleV[l].push_back(*csi);
		}
	}

	//J.clear();
	//vector<int> emptyv; 
	//J.swap(emptyv);	
	//cout << "*11" <<endl;
	for (l = 0; l < LN; l++)
	{
		J.push_back(static_cast<int>(alleleV[l].size()));
	}
	//int JMax = *max_element(J.begin(), J.end());
 		//cout << "*12" <<endl;

	// If there are Validation Data with New Loci
	// We will calculate the cross-validation score as
	// Pr(X_valid | Z) = integral( Pr(X_valid | Z, p_newLoci)) * Pr(p_newLoci | Z) d(p_newLoci) )
	// which has an explicit formula (see Multivariate Polya Distribution)
	// with a part independant of the sweep (LHConstPart)
	double LHConstPart = 0.0; 
	double sumJl = 0.0;
	if (bValidation && !bSameLoci)
	{
		for (l=0; l<LV; l++)
		{
			LHConstPart += gsl_sf_lngamma(J[l+L]*lambda);
			sumJl += J[l+L];
		} 
		cout << endl;
		LHConstPart -= sumJl * gsl_sf_lngamma(lambda);
		LHConstPart *= K;
	}


	const double infinite_value = std::numeric_limits< double >::infinity();
		//cout << "*14" <<endl;
	int max_nb_proposal =0;
	int tmp_nb_proposal=0;
	int loop_sup_PROPOSAL_MAX=0;
	const int PROPOSAL_MAX=15000;
	INT_VCTR_1D proposal;
	InitVctr1D(proposal,K,0);
	double proba_proposal_genotype[K];
	gsl_ran_discrete_t *sample_struct = NULL;

	clock_t s_clock = clock();
	const gsl_rng_type *T;
	gsl_rng_env_setup();
	T = gsl_rng_default;
	gsl_rng* rng = NULL;
	AllocRNG(rng, T);
	
	DBL_VCTR_2D DirParams;
	InitVctr2D(DirParams, L, J, lambda);
	DBL_VCTR_2D DirParamsValid;
	InitVctr2D(DirParamsValid, LV, &J[L], lambda);

	CSweepParams spSweep;
	InitVctr1D(spSweep.z, N, 0);
	InitVctr3D(spSweep.p, K, L, J, 0.0);
	InitVctr2D(spSweep.w, N, K-1, 0.0);
	InitVctr2D(spSweep.beta, K-1, beta_length, 0.0);
	double upper,lower,max;
	DBL_VCTR_2D logpr;
	InitVctr2D(logpr,N,K,0.0);
	DBL_VCTR_2D Pz;
	InitVctr2D(Pz,N,K,0.0);	
	DBL_VCTR_1D maxlogPr;
	InitVctr1D(maxlogPr,N,-infinite_value);
	DBL_VCTR_1D sumlogPr;
	InitVctr1D(sumlogPr,N,0.0);
	double LCurr_valid = 0.0; // Log-likelihood of Validation Data at a sweep
	//double LCurrValidMax = -infinite_value; // For a given sweep and individual, will contain max_k (log(P(Genotype of ind.i| k)))
	//double LCurrValidTmp = 0.0; // Temp variable for calculating log-likelihood of Validation Data
	int b,zi;
	//int Nrep = 100;
	int pas_predict=10;
	int nb_predict=10;

	INT_VCTR_1D sizeClust;
	InitVctr1D(sizeClust,K,0);
	// To Stock Predictions of Training Data Individuals
	DBL_VCTR_3D nClusterQualVar;
	InitVctr3D(nClusterQualVar,nbQualVar,nQualLevels,K,0.0);
	DBL_VCTR_2D nClusterIndiv;
	InitVctr2D(nClusterIndiv,N,K,0.0);
	// To Stock Predictions of Validation Data Individuals
	DBL_VCTR_3D nClusterQualVar_valid;
	InitVctr3D(nClusterQualVar_valid,nbQualVar,nQualLevels,K,0.0);
	DBL_VCTR_2D nClusterIndiv_valid;
	InitVctr2D(nClusterIndiv_valid,NV,K,0.0);
	bool accept;
	DBL_VCTR_2D Pred;
	InitVctr2D(Pred, N, K-1,0.0);
	DBL_VCTR_2D Pred_valid;
	InitVctr2D(Pred_valid, NV, K-1,0.0);
	double tmpPred;
	double tmpPred_valid;
	DBL_VCTR_2D beta_mean;
	InitVctr2D(beta_mean, K-1, beta_length, 0.0);
	// Warning: Beta History is not stored as in TessAdmCOv
	// (for compatibility with the function Significance)
	DBL_VCTR_3D BH;// Beta History 
	InitVctr3D(BH,K-1,beta_length,TotalSweep,0.0);
	//INT_VCTR_2D pop_stock;
	//InitVctr2D(pop_stock,TotalSweep+1,N,-1);
	DBL_VCTR_2D V,V_chol; //V:matrice variance-covariance
	DBL_VCTR_2D V_tmp,Mat_tmp; //for V and beta_mean calcul
	DBL_VCTR_2D V_tXtild; //for beta_mean calcul
	DBL_VCTR_1D Constant; //for initial mean of beta
	InitVctr1D(Constant,beta_length,0.0);
			//cout << "*18" <<endl;
	// If there are covariates (ie beta_length!=0)
	// Calcul of V, V_chol and V_tXtild
	// in R : V=solve(t(Xtild)%*%Xtild)
	// V=V_chol%*%t(V_chol)
	// V_tXtild=V%*%t(Xtild)
	if (beta_length!=0)
	  {
	    Mult_t_left(V_tmp,Xtild,Xtild); //t(Xtild)%*%Xtild
	    InverseWithChol(V,V_tmp);
	    choldc(V_chol,V);
	    Mult_t(V_tXtild,V,Xtild);
	    // Initialize beta
	    for (k=0; k<(K-1); k++)
	      {
		MultivariateGauss(Constant,V_chol,spSweep.beta[k],rng);
	      }
	  }


	// Initialize z and p using configuration file or priors.
	if (bbase)
	//base chosen by user
	{
		for (i = 0; i < N; i++)
		{
			if (isBase[i] != 1)
			{
				spSweep.z[i] = gsl_rng_uniform_int(rng, K-1); 
			}
			else
			{
				spSweep.z[i] = K-1;
			}
			
		}
	}else{
		for (i = 0; i < N; i++)
		{
			spSweep.z[i] = gsl_rng_uniform_int(rng, K);
		}
	}
	
	for (k = 0; k < K; k++)
	{
		for (l = 0; l < L; l++)
		{
			gsl_ran_dirichlet(rng, J[l], &DirParams[l][0], &spSweep.p[k][l][0]);
		}
	}
	//	pop_stock[0]=spSweep.z;

	// Initialize auxiliary variable n.
	INT_VCTR_3D n;
	InitVctr3D(n, K, L, J, 0);
	INT_VCTR_3D nValid;
	InitVctr3D(nValid, K, LV, &J[L], 0);
	INT_VCTR_2D nValidSum;
	InitVctr2D(nValidSum, K, LV, 0); // sum_{j = 0 .. J[l+L]-1} (nValid[k][l][j])
	int nSum_kl;
	int nSum;

	// auxiliary variable 
	INT_VCTR_1D nIndInClust; // Number of Individuals (from Training Data) classified in Each Cluster
	InitVctr1D(nIndInClust,K,0);
	DBL_VCTR_1D lognIndInClust; // Log of nIndInClust
	InitVctr1D(lognIndInClust,K,0.0);
	
	// Initialize auxiliary variable logp.
	DBL_VCTR_3D logp;
	InitVctr3D(logp, K, L, J, 0.0);

	// Initialize auxiliary variable for Validation Phase.
	DBL_VCTR_3D pCurr_valid;
	InitVctr3D(pCurr_valid, K, LV, &J[L], 0.0);
	DBL_VCTR_3D logpCurr_valid;
	InitVctr3D(logpCurr_valid, K, LV, &J[L], 0.0);

	// Auxiliary variable to calculate Logarithms
	gsl_sf_result logpTmp;
	//gsl_sf_result logpTmp_valid;
	gsl_sf_result lognIICTmp;
	
	int status;
	int nReplaceLogp = 0;
	//int nReplaceLogp_valid = 0;
	
	// Initialize clustering probabilities cpSweep and cpAverage.
	DBL_VCTR_2D cpAverage;
	InitVctr2D(cpAverage, N, K, 0.0);

	CSweepParams spAverage;
	InitVctr1D(spAverage.z, N, 0);
	InitVctr3D(spAverage.p, K, L, J, 0.0);
	InitVctr2D(spAverage.beta,K-1,beta_length,0.0);
	InitVctr2D(spAverage.w,N,K-1,0.0);
	
	int KeptSweep;
	//mod
	if (beta_length==0)
	// Structure Only -> BurnInStructure = TotalSweep
	{
	  cout << "No covariates" << endl;
		KeptSweep = TotalSweep - BurnIn - BurnInStructure;
		BurnInStructure = TotalSweep; // For all sweeps updates are simular to the updates in STRUCTURE model (without covariate)
		BurnIn = 0; // No Burn-In Sweeps using Information of Covariates
	}
	else 
	{
		if (!bBS)
		{
			//User did not choose a value for BurnInStructure so we fixe it to max(50, 0.05 * BurnIn) //5 percents
			//cout << "User did not choose BS" << endl;
			BurnInStructure=(int)floor(0.05*BurnIn);  
			if (BurnInStructure > 50) 
			{
				BurnInStructure=50;
			}
			BurnIn = BurnIn - BurnInStructure;
		}
		KeptSweep = TotalSweep - BurnIn - BurnInStructure;
	}

	//cout << "Sweeps times:  " << BurnInStructure << "\t" << BurnIn << "\t" << KeptSweep << "\t" << TotalSweep << endl;
	
	DBL_VCTR_1D LH(TotalSweep, 0.0);  // Log-Likelihood History of Training Data
	DBL_VCTR_1D LH_valid(KeptSweep, 0.0);  // Log-Likelihood History of Validation Data
	DBL_VCTR_2D AIC;  // AIC History (for regressions in cluster 1 to K-1 and the last column contains the sum over all clusters)
	InitVctr2D(AIC, TotalSweep, K, 0.0);
	DBL_VCTR_2D RSS;  // Residual Sum of Squares (for regressions in cluster 1 to K-1 and the last column contains the sum over all clusters)
	InitVctr2D(RSS, TotalSweep, K, 0.0);
	double tmpRSS=0.0; //residual sum of squares

	DBL_VCTR_2D varBeta;
	InitVctr2D(varBeta,K-1,beta_length,0.0);
	
	// Now, let's start our Markov chain Monte Carlo.
	for (int s = 0; s < TotalSweep; s++)
	{
		
		// Update p.
		for (k = 0; k < K; k++)
		{
			// cout<<"cluster "<<k<<" :"<<endl;
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					nSum = 0;
					for (i = 0; i < N; i++)
					{
						for (a = 0; a < A; a++)
						{
							if (X[i][a][l] == alleleV[l][j] && spSweep.z[i] == k)
							{
								nSum++;
							}
						}
					}
					n[k][l][j] = nSum;
					//cout<<n[k][l][j];
				}
				//cout<<endl;
			}
			//cout<<endl;
		}
		for (k = 0; k < K; k++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					DirParams[l][j] = n[k][l][j] + lambda;
				}
				gsl_ran_dirichlet(rng, J[l], &DirParams[l][0], &spSweep.p[k][l][0]);
			}
		}
		
		// Update logp.
		for (k = 0; k < K; k++)
		{
			//cout<<"cluster "<<k<<" :"<<endl;
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					status = gsl_sf_log_e(spSweep.p[k][l][j], &logpTmp);
					if (status)
					{
						logp[k][l][j] = N_LIMIT;
						nReplaceLogp++;
					}
					else
					{
						logp[k][l][j] = logpTmp.val;
					}
				}
			}
		}
		
		
		for (i = 0; i < N; i ++)
		{
			//maxPr[i] = 0.0;
			maxlogPr[i]=-infinite_value;
			sumlogPr[i]=0.0;
			for (k = 0; k < K; k ++)
			{
				logpr[i][k] = 0.0;
				for (a = 0; a < A; a++)
				{
					for (l = 0; l < L; l++)
					{
						for (j = 0; j < J[l]; j ++)
						{
							if (X[i][a][l] == alleleV[l][j])
							{
								logpr[i][k] += logp[k][l][j];
								break;
							}
						}
						
					}
				}
				/*double tmp = exp(logpr[i][k]);
			if (tmp > maxPr[i])
			{
			maxPr[i] = tmp;
			}*/
				if (logpr[i][k]>maxlogPr[i])
				{
					maxlogPr[i]=logpr[i][k];
				}
				sumlogPr[i]+=logpr[i][k];
			}
		}

		if ( s < BurnInStructure ) 
		{
			// Update z.
			//if (s==0) {cout<< "update z with old version (logRatio ...)"<<endl;}
			for (i = 0; i < N; i++)
			{
				if ( (!bbase) || (isBase[i]!=1) ) 
				{
					int kOld = spSweep.z[i];
					int kNew = gsl_rng_uniform_int(rng, K);
					if (kNew != kOld)
					{
						double dLogRatio = 0.0;
						for (a = 0; a < A; a++)
						{
							for (l = 0; l < L; l++)
							{
								for (j = 0; j < J[l]; j++)
								{
									if (X[i][a][l] == alleleV[l][j])
									{
										dLogRatio += logp[kNew][l][j] - logp[kOld][l][j];
										break;
									}
								}       
							}
						}
						
						double uni = gsl_rng_uniform(rng);
						if (uni < exp(dLogRatio))
						{
							spSweep.z[i] = kNew;
						}
					}
					// simuler wi(0..K-2)
					//mod
					if (beta_length!=0)
					{
						for (k=0; k<(K-1); k++)
						{
							if (spSweep.z[i]==k)
							{
								lower=FindBound(spSweep.w[i],k);
								spSweep.w[i][k]=TruncGauss_Left(Pred[i][k],1,lower,rng);
								if (spSweep.w[i][k]<lower)
								{
									cout<<"sweep "<<s<<": spSweep.w[i][k]<lower"<<endl;
									cout<<"exit "<<run<<endl;
									return 0;
								}
								if (isnan(spSweep.w[i][k]) || isinf(spSweep.w[i][k]))
								{
									cout<<"sweep "<<s<<": Error in simulating rtrun( "<<Pred[i][k]<<" ,1, "<<lower<<" ,Inf ) with cdf inversion method"<<endl;
									cout<<"Exit "<<run<<endl;
									return 0;
								}
							}
							else
							{
								upper=FindBound(spSweep.w[i],k);
								spSweep.w[i][k]=TruncGauss_Right(Pred[i][k],1,upper,rng);
								if (spSweep.w[i][k]>upper)
								{
									cout<<"sweep "<<s<<": spSweep.w[i][k]>upper"<<endl;
									cout<<"exit "<<run<<endl;
									return 0;
								}
								if (isnan(spSweep.w[i][k]) || isinf(spSweep.w[i][k]))
								{
									cout<<"sweep "<<s<<": Error in simulating rtrun( "<<Pred[i][k]<<" ,1,-Inf, "<<upper<<" ) with cdf inversion method"<<endl;
									cout<<"Exit "<<run<<endl;
									return 0;
								}
							}
						}
					}
					
				}
			}
		}
		else
		{
			/**************  s>=BurnInStructure :  nouvelle version *******************/
			//if (s==BurnInStructure) cout<< "update z with new version (reject loop)"<<endl;
		  
			Mult_t(Pred,Xtild,spSweep.beta); //Xtild%*%t(Beta) Xtild= covariables
			// AIC
			for (k=0; k<(K-1); k++)
			{
				tmpRSS=0.0;
				for (i=0; i<N; i++)
				{
					tmpRSS += (spSweep.w[i][k]- Pred[i][k])*(spSweep.w[i][k]- Pred[i][k]);
				}
				AIC[s][k]= 2*beta_length + tmpRSS;
				AIC[s][K-1] += 2*beta_length + tmpRSS; //the last column contains AIC for all clusters (sum over all clusters)
				RSS[s][k] = tmpRSS;   // Residual Sum of Squares for regression in cluster k
				RSS[s][K-1] += tmpRSS; //RSS sum over all regressions 
			}
			

			// Update simultaneously W and Z
			for (i = 0; i < N; i++)
			{
				//cout<<" i="<<i<<endl;
				
				if ( (!bbase) || (isBase[i]!=1) ) 
				{
					//if there is a base, the base/outgroup individuals stay in the base population 
					//if (true) //the base/outgroup individuals can be affected to others populations
					//cout<<"isBase["<<i<<"] = "<<isBase[i]<<endl;
					accept = false;
					tmp_nb_proposal=0;
					for (k=0; k < K; k++)
					{
						proposal[k]=0;
					}
					while( !accept && (tmp_nb_proposal<PROPOSAL_MAX) )
					{
						tmp_nb_proposal++;
						max=0;
						zi=K-1; //Z[i] = base if max<0 
						
						for (k=0; k < K-1; k++)
						{
							spSweep.w[i][k]=Pred[i][k]+gsl_ran_ugaussian(rng);
							if (spSweep.w[i][k] > max)
							{
								max = spSweep.w[i][k];
								zi = k;
							}
						}
						//cout<<"proposal: "<<zi<<"  avec proba "<<  exp(logpr[i][zi]-maxlogPr[i])<<" = exp( "<<logpr[i][zi]<<"  - "<<maxlogPr[i]<<" )"<<endl;
						proposal[zi]++;
						double u = gsl_rng_uniform(rng);
						//if (u < exp(logpr[i][zi])/maxPr[i])
						if (u < exp(logpr[i][zi]-maxlogPr[i]))
						{
							accept = true;
							if (tmp_nb_proposal>max_nb_proposal){
								max_nb_proposal=tmp_nb_proposal;
							}
						}
					}
					
					if (tmp_nb_proposal==PROPOSAL_MAX)
					{
						// zi hasn't been accepted before the proposal number reachs PROPOSAL_MAX
						// so we select zi thanks to a sampling function
						loop_sup_PROPOSAL_MAX++;
						if (loop_sup_PROPOSAL_MAX==200000)
						{
							cout<<"Too much reject loops ("<<loop_sup_PROPOSAL_MAX<<") are replaced by the alternative method so we leave "<<run<<endl;
							cout<<endl;
							return 0;
						}
						for (k=0; k<K; k++)
						{
							proba_proposal_genotype[k]=exp(logpr[i][k]-maxlogPr[i]);
						}
						sample_struct=gsl_ran_discrete_preproc(K,proba_proposal_genotype);
						zi=gsl_ran_discrete(rng,sample_struct);
						cout << "cluster choosen for individual "<<i<<" with a sample function : " << zi << endl;
						//cout << "**************************************************" << endl;
					}

					spSweep.z[i] = zi;
				}
			}
		}
		/************* fin nouvelle version ***********/
		
		// Update beta
		// in R :beta_mean=V%*%t(Xtild)%*%W;
		// V_tXtild=v%*%t(Xtild) has already been calculated
		if (beta_length!=0)
		{
			Mult(Mat_tmp,V_tXtild,spSweep.w);
			Transpose(beta_mean,Mat_tmp);
			for (k=0; k<(K-1); k++)
			{
				MultivariateGauss(beta_mean[k],V_chol,spSweep.beta[k],rng);
				if (isinf(spSweep.beta[k][0]) || isnan(spSweep.beta[k][0]))
				
				{
					cout<<"Warning beta["<<k<<"][0]="<<spSweep.beta[k][0]<<endl;
					for (int p=0; p<beta_length; p++) 
					  {
					    spSweep.beta[k][p]=BH[k][p][s-1];
					  }
					cout<<"spSweep.beta replaced by previous beta, for k="<<k<<endl;
				}
			}
			
		}
		
		// Update Log-Likelihood.
		double LCurr = 0.0;
		for (i = 0; i < N; i++)
		{
			for (a = 0; a < A; a++)
			{
				for (l = 0; l < L; l++)
				{
					for (j = 0; j < J[l]; j++)
					{
						if (X[i][a][l] == alleleV[l][j])
						{
							LCurr += logp[spSweep.z[i]][l][j];
							break;
						}
					}
				}
			}
		}
		LH[s] = LCurr;
		
		for (k=0; k<(K-1); k++){
			for (int p=0; p<beta_length; p++){
				BH[k][p][s]=spSweep.beta[k][p];
			}
		}
		//		pop_stock[s+1]=spSweep.z;
		// Accumulate results after burn-in.
		//mod if (s > (BurnInStructure + BurnIn - 1))
		if  (s == (TotalSweep - KeptSweep)) 
		{
			cout << "End of Burn-in" << endl;
		}
		if (s >= (TotalSweep - KeptSweep))
		{
			// p
			for (k = 0; k < K; k++)
			{
				for (l = 0; l < L; l++)
				{
					for (j = 0; j < J[l]; j++)
					{
						spAverage.p[k][l][j] += spSweep.p[k][l][j];
					}
				}
			}

			// cp and temp variable nCurrClust

			for (k = 0; k < K; k++)
			{
				nIndInClust[k] = 0;
			}
			for (i = 0; i < N; i++)
			{
				for (k = 0; k < K; k++)
				{
					if (spSweep.z[i] == k)
					{
						cpAverage[i][k] += 1.0;
						nIndInClust[k]++;
						break;
					}
				}
			}
			for (k = 0; k < K; k++)
			{
				status = gsl_sf_log_e(nIndInClust[k], &lognIICTmp);
				if (status)
				{
					lognIndInClust[k] = N_LIMIT;
				}
				else
				{
					lognIndInClust[k] = lognIICTmp.val;
				}
			}	

			
			// beta
			if (beta_length!=0)
			{
				for (k=0; k < K-1; k++)
				{
					for (b = 0; b < beta_length ; b++)
					{
						spAverage.beta[k][b] += spSweep.beta[k][b];
						varBeta[k][b] += spSweep.beta[k][b] * spSweep.beta[k][b];
					} 
				}
			}
			
			// W
			for (i = 0; i < N; i++)
			{
				for (k = 0; k < K-1; k++)
				{
					spAverage.w[i][k] += spSweep.w[i][k];
				}
			}
			
			
			// Prediction if there are covariates
			if (s%pas_predict==0 && beta_length!=0)
			{
			  for (predict=0; predict <nb_predict; predict++)
				{
					//calcul pij
					for (i = 0; i < N; i++)
					{
						max = 0;
						zi = K-1; //Z[i] = base if max<0 
						
						for (k=0; k < K-1; k++)
						{
							tmpPred = Pred[i][k]+gsl_ran_ugaussian(rng);
							if (tmpPred > max)
							{
								max = tmpPred;
								zi = k;
							}
						}
						if (nbQualVar>0)
						{
							for (var=0; var < nbQualVar; var++)
							{
							  nClusterQualVar[var][XQual[i][var]][zi]++;
							}
						}
						nClusterIndiv[i][zi]++;
					}
				}
			}


			// Validation
			if (bValidation)
			{ 
				/*mod 
			if (beta_length != 0)
			{
			Mult_t(Pred_valid,Xtild_valid,spSweep.beta); //Xtild_valid%*%t(Beta) Xtild_valid = Covariables of the Validation Data
			}
			*/
				if (bSameLoci) 
				{
					LCurr_valid = 0.0;
					
					if (beta_length != 0)
					// Covariates -> Prior Non Uniform for z (Simulation from Probit Model)
					{
						Mult_t(Pred_valid,Xtild_valid,spSweep.beta); //Xtild_valid%*%t(Beta) Xtild_valid = Covariables of the Validation Data
						
						for (i = 0; i < NV; i++)
						{
							// Simulate k from the learned Probit Model
							max = 0;
							zi = K-1; //Z[i] = base if max<0 
							
							for (k=0; k < K-1; k++)
							{
								tmpPred_valid = Pred_valid[i][k]+gsl_ran_ugaussian(rng);
								if (tmpPred_valid > max)
								{
									max = tmpPred_valid;
									zi = k;
								}
							}
							if (nbQualVar>0)
							{
								for (var=0; var < nbQualVar; var++)
								{
									nClusterQualVar_valid[var][XQual_valid[i][var]][zi]++;
								}
							}
							nClusterIndiv_valid[i][zi]++;
							
							for (a = 0; a < A; a++)
							{
								for (l = 0; l < L; l++)
								{
									for (j = 0; j < J[l]; j++)
									{
										if (XsameLoci[i][a][l] == alleleV[l][j])
										{
											LCurr_valid += logp[zi][l][j];
											break;
										}
									}
								}
								
							}
						}
					}
					else
					// No Covariates -> Prior for z : P(z=k) = 1/K
					{
						for (i = 0; i < NV; i++)
						{
							zi = gsl_rng_uniform_int(rng, K);
							for (a = 0; a < A; a++)
							{
								for (l = 0; l < L; l++)
								{
									for (j = 0; j < J[l]; j++)
									{
										if (XsameLoci[i][a][l] == alleleV[l][j])
										{
											LCurr_valid += logp[zi][l][j];
											break;
										}
									}
								}
							}
							
						}
					}
					/* first method with DL approximation (see notes)
				{			  
				
				int ki=-1;
				for (i = 0; i < NV; i++)
				{
					LCurrValidMax = -infinite_value; // Will contain max_k { log(P(Genotype of ind.i| k)) + log(Nb of Training Indivividuals in Clust k)  }
					ki = -1;
					for (k=0; k<K; k++) 
					{
					LCurrValidTmp = 0.0;
					for (a = 0; a < A; a++)
					{
						for (l = 0; l < L; l++)
						{
						for (j = 0; j < J[l]; j++)
						{
							if (XsameLoci[i][a][l] == alleleV[l][j])
							{
							LCurrValidTmp += logp[k][l][j];
							break;
							}
						}
						}
					}
					LCurrValidTmp += lognIndInClust[k];
					if (LCurrValidTmp > LCurrValidMax)
					{
						LCurrValidMax = LCurrValidTmp;
						ki = k;
					}
					}
					LCurr_valid += LCurrValidMax;
				}
				LCurr_valid -= (double)(NV*log(N));
				}*/
					
					LH_valid[s - (TotalSweep-KeptSweep)] = LCurr_valid;
				}
				else 
				{
					for (k = 0; k < K; k++)
					{
						sizeClust[k] = 0;
					}
					for (i = 0; i < NV; i++)
					{
						// if sameIndividuals
						ZCurr_valid[i]=spSweep.z[i];
						sizeClust[spSweep.z[i]]++;
					}
					
					// 2. New Loci Number in each Cluster
					
					for (k = 0; k < K; k++)
					{
						//cout << "sizeClust[" << k << "] = " << sizeClust[k] << endl;
						for (l = 0; l < LV; l++)
						{
							nSum_kl = 0;
							// Warning: information about New Loci in J and alleleV are stocked after Training Loci (ie at index l+L)
							for (j = 0; j < J[l+L]; j++) 
							{
								nSum = 0;
								for (i = 0; i < NV; i++)
								{
									for (a = 0; a < A; a++)
									{
										if (XnewLoci[i][a][l] == alleleV[l+L][j] && ZCurr_valid[i] == k)
										{
											nSum++;
										}
									}
								}
								nValid[k][l][j] = nSum;
								nSum_kl += nSum;
							}
							nValidSum[k][l] = nSum_kl; //nSum_kl =  sizeClust[k]*A - nb missing data at locus l in clust k
						}
					}


					// 3. Calculate Log-Likelihood of Validation Data
					// by integrating over p_newLoci (alleles frequencies at new loci)
					// see explicit formula for Polya Distribution 
					LCurr_valid = 0.0;
					/* for (k = 0; k < K; k++)
				{
				for (l=0; l<LV; l++)
				{
					LCurr_valid += gsl_sf_lnfact(nValidSum[k][l]) - gsl_sf_lngamma(J[l+L]*lambda + nValidSum[k][l]);
					for (j=0; j<J[l+L]; j++)
					{
					LCurr_valid += gsl_sf_lngamma(nValid[k][l][j]+lambda) - gsl_sf_lnfact(nValid[k][l][j]);
					}
				}
				}
				*/

					for (k = 0; k < K; k++)
					{
						for (l=0; l<LV; l++)
						{
							LCurr_valid -= gsl_sf_lngamma(J[l+L]*lambda + nValidSum[k][l]);
							//cout << nValidSum[k][l] << " " << LCurr_valid << endl;
							
							for (j=0; j<J[l+L]; j++)
							{
								LCurr_valid += gsl_sf_lngamma(nValid[k][l][j]+lambda);
								//cout << nValid[k][l][j]<< " "<< gsl_sf_lngamma(nValid[k][l][j]+lambda) <<"\t";
							}
						}
					}
					//cout<<endl;
					// cout << "sweep " << s << " LCurr_valid=" << LCurr_valid << endl;
					LH_valid[s - (TotalSweep-KeptSweep)] = LCurr_valid + LHConstPart;
					
				}
			}
		}
		int spr = s + 1;
		if (spr % 100 == 0)
		{
			cout.width(NumOfDigits);
			cout.fill('0');
			cout << spr << " sweeps completed." << endl;
			//cout << loop_sup_PROPOSAL_MAX << " replaced reject loops.  " << "  max nb proposal = " << max_nb_proposal<< endl;
		}
	}
	
	// end of the MCMC 


	cout << run << " completed" << endl;
	//cout << "with " << loop_sup_PROPOSAL_MAX << " replaced reject loops > " << PROPOSAL_MAX << endl;
	//cout << "logp in Training Data replaced " << nReplaceLogp << " times by N_limit (" << N_LIMIT << ")" << endl;
	//cout << "logp in Validation Data replaced " << nReplaceLogp_valid << " times by N_limit" << endl;

	
	// Calculate the Monte Carlo averages after MCMC.
	// p
	for (k = 0; k < K; k++)
	{
		for (l = 0; l < L; l++)
		{
			for (j = 0; j < J[l]; j++)
			{
				spAverage.p[k][l][j] /= KeptSweep;
			}
		}
	}
	
	// cp
	for (i = 0; i < N; i++)
	{
		for (k = 0; k < K; k++)
		{
			cpAverage[i][k] /= KeptSweep;
		}
	}
	
	// Regression Coefficients Beta
	if (beta_length != 0) 
	  {
	    for (k = 0; k < K-1; k++)
	      {
		for (b = 0; b < beta_length; b++)
		  {
		    spAverage.beta[k][b] /= KeptSweep;
		    varBeta[k][b] /= KeptSweep;
		  }
	      }
	    
	    
	    string BHat = FOFN + "betaHat.txt";
	    SaveToMyFile(spAverage.beta,BHat);
	    string strBH = FOFN + "BH.png";
	    PlotBetaHist(strBH.c_str(),BH, nQualLevels, nbQuantVar, trendDegree);
	  }	

	// W
	for (i = 0; i < N; i++)
	{
		for (k = 0; k < K-1; k++)
		{
			spAverage.w[i][k] /= KeptSweep;
		}
	}
	
	// Hard Clustering
	INT_VCTR_1D hc(N, 0);
	for (i = 0; i < N; i++)
	{
		double dMax = -1.0;
		for (k = 0; k < K; k++)
		{
			if (cpAverage[i][k] > dMax)
			{
				dMax = cpAverage[i][k];
				hc[i] = k;
			}
		}
	}
	
	// 	string strHC = FOFN + "HC.png";
	// 	PlotHC(strHC.c_str(), vptC, vbKeep, vdDataMinMax, veVoronoi, hc);

	// 	string strCF = FIFN + "CF";
	// 	SaveCF(hc, spAverage.p, strCF);

	// pi
	DBL_VCTR_1D pi(K, 0.0);
	for (k = 0; k < K; k++)
	{
		for (i = 0; i < N; i++)
		{
			pi[k] += cpAverage[i][k];
		}
	}
	for (k = 0; k < K; k++)
	{
		pi[k] /= N;
	}


	// Calculate, Stock and Plot Predicted coefficients, for Training and Validation Data

	if (beta_length != 0)
	// There are Covariates -> Predicted coefficients
	{
		// Predicted coefficients for each Individuals
		for (k=0; k<K; k++)
		{
			for (i=0; i<N; i++)
			{
				nClusterIndiv[i][k]/=(double)nb_predict*((double)KeptSweep/(double)pas_predict);
			}
			if (bValidation && bSameLoci) // There are New Individuals
			{
				for (i=0; i<NV; i++)
				{
					nClusterIndiv_valid[i][k]/=(double)KeptSweep;
				}
			}
		}
		string clIndFile=FOFN+"PPM.txt";
		SaveToMyFileClumppFormat(nClusterIndiv,clIndFile,true,"Predicted Clustering Probabilities for Samples");
		string ClInd_barplot=FOFN+"PPM.png";
		PlotAP(ClInd_barplot.c_str(),nClusterIndiv, "Predicted Assignment Probabilities");
		
		runCriteria.cor = Correlation(cpAverage,nClusterIndiv);
	
		if (bValidation && bSameLoci)
		{
			string clIndFile_valid=FOFN+"PPM_valid.txt";
			SaveToMyFileClumppFormat(nClusterIndiv_valid,clIndFile_valid,true,"Predicted Clustering Probabilities for Samples");
			string ClInd_barplot_valid=FOFN+"PPM_valid.png";
			PlotAP(ClInd_barplot_valid.c_str(),nClusterIndiv_valid, "Predicted Assignment Probabilities");
		}
		
		// Predictions averaged among Class of the Qualitative Variables (eg Linguistic Groups for the variable 'Languages')
		if (nbQualVar>0)
		  {
		    for (var=0; var < nbQualVar; var++)
		      {
			for (k=0; k<K; k++)
			  {
			    for (int level=0;level<nQualLevels[var];level++)
			      {
				nClusterQualVar[var][level][k]/=(double)nb_predict*(double)LevelPopSize[var][level]*((double)KeptSweep/(double)pas_predict);
				if (bValidation)
				  {
				    nClusterQualVar_valid[var][level][k]/=(double)LevelPopSize_valid[var][level]*((double)KeptSweep);
				  }
			      }
			  }
			
			string svar;
			stringstream out;
			out << var+1;
			svar = out.str();
			
			// PPM = Posterior Predictive Membership
			string clFile=FOFN+"PPMvar"+svar+".txt";
			SaveToMyFileClumppFormat(nClusterQualVar[var],clFile);
			//string ClQualVar_barplot=FOFN+"PPM"+svar+".png";
			//PlotAP(ClQualVar_barplot.c_str(),nClusterQualVar[var]);
			if (bValidation && bSameLoci) 
			  {
			    string clFile_valid=FOFN+"PPM"+svar+"_valid.txt";
			    SaveToMyFileClumppFormat(nClusterQualVar_valid[var],clFile_valid);
			    //string ClQualVar_barplot_valid=FOFN+"PPM"+svar+"_valid.png";
			    //PlotAP(ClQualVar_barplot_valid.c_str(),nClusterQualVar_valid[var]);
			  }
		      }
		    string ppmqualFile = FOFN+"PPMqual.png";
		    PlotPPMqual(ppmqualFile.c_str(),nClusterQualVar);
		    if (bValidation && bSameLoci) 
		      {
			string ClQualVar_barplot_valid=FOFN+"PPMqual"+"_valid.png";
			PlotPPMqual(ClQualVar_barplot_valid.c_str(),nClusterQualVar_valid);
		      }
		  }
	}
	else
	  {
	    DBL_VCTR_2D randomAssignment;
	    InitVctr2D(randomAssignment,N,K,1.0/K);
	    runCriteria.cor = Correlation(cpAverage,randomAssignment);
	  }

	cout << "Correlation between estimated and predicted membership coeff = " << runCriteria.cor << endl;
	string strCOR = FOFN + "COR.txt";
	ofstream ofsCOR(strCOR.c_str());
	ofsCOR << runCriteria.cor;
	ofsCOR.close();
		
		

	// Log the results to a text file.
	string strTR = FOFN + "TR.txt";
	ofstream ofsTR(strTR.c_str());

	// Estimated Clustering Probabilities for Samples
	ofsTR << "Estimated Clustering Probabilities for Samples" << endl;
	WriteInOfstreamClummpFormat(cpAverage, ofsTR);
	if (beta_length!=0)
	{
		ofsTR << "Predicted Clustering Probabilities for Samples" << endl;
		WriteInOfstreamClummpFormat(nClusterIndiv,ofsTR);
		if (nbQualVar>0)
		{
			for (var=0; var < nbQualVar; var++)
			{
				ofsTR << "Predicted Clustering Probabilities Averaged over Individuals from the Same Class of the Qualitative Variable n°" << var+1 << endl;
				WriteInOfstreamClummpFormat(nClusterQualVar[var],ofsTR);
			}
		}
	}
	/*ofsTR << "Sample\t";
	for (k = 0; k < K; k++)
	{
		ofsTR << "***Cluster  ";
		ofsTR.width(2);
		ofsTR.fill('0');
		// Make sure the first index is 1 not 0.
		ofsTR << k + 1 << "***\t";
	}
	ofsTR << endl;
	for (i = 0; i < N; i++)
	{
		ofsTR.width(6);
		ofsTR.fill('0');
		// Make sure the first index is 1 not 0.
		ofsTR << i + 1 << ":\t";
		for (k = 0; k < K; k++)
		{
			ofsTR.width(17);
			ofsTR.precision(15);
			ofsTR << fixed;
			ofsTR << cpAverage[i][k] << " ";
		}
		ofsTR << "\t" << 1 << endl;
	}
	ofsTR << endl << endl << endl;
	*/

	// Estimated Allele Frequencies
	ofsTR << "Estimated Allele Frequencies" << endl;
	for (k = 0; k < K; k++)
	{
		ofsTR << "Cluster ";
		ofsTR.width(2);
		ofsTR.fill('0');
		// Make sure the first index is 1 not 0.
		ofsTR << k + 1 << endl;
		for (l = 0; l < L; l++)
		{
			ofsTR << "Locus ";
			ofsTR.width(3);
			ofsTR.fill('0');
			// Make sure the first index is 1 not 0.
			ofsTR << l + 1 << endl;
			for (j = 0; j < J[l]; j++)
			{
				ofsTR << "Allele ";
				ofsTR.width(3);
				ofsTR.fill('0');
				ofsTR << alleleV[l][j] << "\t";
				ofsTR.width(17);
				ofsTR.precision(15);
				ofsTR << fixed;
				ofsTR << spAverage.p[k][l][j] << endl;
			}
			ofsTR << endl;
		}
		ofsTR << endl;
	}
	ofsTR << endl;

	// Estimated Mixing Proportions
	ofsTR << "Estimated Mixing Proportions" << endl;
	for (k = 0; k < K; k++)
	{
		ofsTR.width(17);
		ofsTR.precision(15);
		ofsTR << fixed;
		ofsTR << pi[k] << endl;
	}
	ofsTR << endl << endl << endl;
	
	// Estimated Coefficients of Regression Beta
	if (beta_length!=0)
	{
	  string strBHtext=FOFN+"BP.txt";
	  SaveBetaToMyFile(BH,strBHtext,BurnIn+BurnInStructure);
		
	  /* estimated beta saved in FOFN+betaHat.txt
	    ofsTR << "Estimated Beta:" << endl;
		for (k = 0; k < K-1; k++)
		{
			ofsTR << k+1 << ":" << "\t";
			for (b = 0; b < beta_length; b++)
			{
				ofsTR.width(17);
				ofsTR.precision(15);
				ofsTR << fixed;
				ofsTR << spAverage.beta[k][b] << "\t";
			}
			ofsTR << endl;
		}
		ofsTR << endl << endl << endl;
	  */

		ofsTR << "Estimated Beta and Significance Level:" << endl;
		for (k = 0; k < K-1; k++)
		{
			ofsTR << k+1 << ":" << "\t";
			for (b = 0; b < beta_length; b++)
			{
				ofsTR.width(17);
				ofsTR.precision(15);
				ofsTR << fixed;
				ofsTR << spAverage.beta[k][b] << " "<< Significance(&BH[k][b][BurnIn+BurnInStructure],KeptSweep)<< "\t";
			}
			ofsTR << endl;
		}

		ofsTR << endl <<"100\% '****' 99.9 \% '***' 99\% '**' 95\% '*' 90\% '-' 0\% " <<endl;
		ofsTR << endl << endl << endl;


		ofsTR << "95 % Bayesian Confidence Interval for Estimated Beta:" << endl;
		for (k = 0; k < K-1; k++)
		  {
		    ofsTR << k+1 << ":" << "\t";
		    for (b = 0; b < beta_length; b++)
		      {
			ofsTR.width(17);
			ofsTR.precision(15);
			ofsTR << fixed;
			ofsTR << CI(&BH[k][b][BurnIn+BurnInStructure],KeptSweep,0.025,0.975) << "\t";
		      }
		    ofsTR << endl;
		  }
		ofsTR << endl << endl << endl;

		/*
		ofsTR << "Approximate Confidence Interval for Estimated Beta:" << endl;
		for (k = 0; k < K-1; k++)
		{
			ofsTR << k+1 << ":" << "\t";
			for (b = 0; b < beta_length; b++)
			{
				ofsTR.width(17);
				ofsTR.precision(15);
				ofsTR << fixed;
				ofsTR << spAverage.beta[k][b] - 1.96*sqrt((double)(varBeta[k][b] - spAverage.beta[k][b] * spAverage.beta[k][b]))/sqrt(KeptSweep) << "  --  " << spAverage.beta[k][b] + 1.96*sqrt((double)(varBeta[k][b] - spAverage.beta[k][b] * spAverage.beta[k][b]))/sqrt(KeptSweep) << "\t";
			}
			ofsTR << endl;
		}
		ofsTR << endl << endl << endl;
		*/

		
		
		// Estimated W
		ofsTR << "Estimated Latent Variable W:" << endl;
		for (i = 0; i < N; i++)
		{
			ofsTR << i << ":" << "\t";
			for (k = 0; k < K-1; k++)
			{
				ofsTR.width(17);
				ofsTR.precision(15);
				ofsTR << fixed;
				ofsTR << spAverage.w[i][k] << "\t";
			}
			ofsTR << endl;
		}
		
	}
	ofsTR.close();


	// Calculate, Stock and Plot Estimated and Predicted coefficients, for Training and Validation Data

	string strAPstock = FOFN + "AP.txt";
	SaveToMyFileClumppFormat(cpAverage,strAPstock);
	string strAP = FOFN + "AP.png";
	PlotAP(strAP.c_str(), cpAverage, "Assignment Probabilities");

	
	// Cluster number of Each Individual at Each Sweep
	//string popFile=FOFN+"PopStock.txt";
	//SaveToMyFile(pop_stock,popFile);
	
	
	// Calculate Estimated Number of Clusters
	int KEst = K;
	for (k = 0; k < K; k++)
	{
		if (pi[k] < 1.0 / (8 * K))
		{
			KEst--;
		}
	}
	string strEN = FOFN + "EN.txt";
	ofstream ofsEN(strEN.c_str());
	ofsEN << KEst << endl;
	ofsEN.close();


	// Calculate and Stock Criteria

	// Average Log-Likelihood, Average AIC, DIC, Bayes Factor
	// Note that the last column of AIC contains the AIC for regressions over ALL cluster

	// LH of Training Data during the run
	string strLH = FOFN + "LH.png";
	PlotLH(strLH.c_str(), LH);
	string strLHstock = FOFN + "LHstock.txt";
	SaveToMyFile(LH,strLHstock);
	
	// LH of Validation Data during the run
	if (bValidation)
	{
		string strLH_valid = FOFN + "LH_valid.png";
		PlotLH(strLH_valid.c_str(), LH_valid);
		string strLHstock_valid = FOFN + "LHstock_valid.txt";
		SaveToMyFile(LH_valid,strLHstock_valid);
	}
	
	// Stock AIC during the run
	if (beta_length!=0)
	{
		string AICFile=FOFN+"AICStock.txt";
		SaveToMyFile(AIC,AICFile);
	}

	//int NoS = int(0.1 * KeptSweep);  // Number of Sweeps Used in Calculation
	int NoS = int(1 * KeptSweep);  // Number of Sweeps Used in Calculation
	if (NoS == 0)
	{
		NoS = 1;
	}
	double dAL = 0.0; // Average Log-Likelihood of Training Data
	double dAL_valid = 0.0; // Average Log-Likelihood of Validation Data
	double dAL_tot = 0.0; // Average Log-Likelihood of All Data
	double varLH = 0.0; // Variance of Log-Likelihood of Training Data
	double varLH_valid = 0.0; // Variance of Log-Likelihood of Validation Data
	double varLH_tot = 0.0; // Variance of Log-Likelihood of All Data
	DBL_VCTR_1D dAIC(K,0.0);
	DBL_VCTR_1D dRSS(K,0.0);
	DBL_VCTR_1D varRSS(K,0.0); // Variance of RSS of Training Data
	double BF = 0.0;  //(LH max)/p(Data) : to calculate Bayes Factor
	double LHmax=-infinite_value; // MAX Log-Likelihood for training data

	for (i = TotalSweep - NoS; i < TotalSweep; i++)
	{
		// Log-Likelihood max
		if (LH[i] > LHmax) 
		{
			LHmax = LH[i];
		}
	}
	//cout << "NoS=" << NoS << endl;
	//cout << "*lh valid*" << endl;
	for (i = TotalSweep - NoS; i < TotalSweep; i++)
	{
		dAL += LH[i];
		varLH += LH[i]*LH[i];
		dAL_valid += LH_valid[i - (TotalSweep-KeptSweep)];
		//cout << i - (TotalSweep-KeptSweep) << ": " << LH_valid[i - (TotalSweep-KeptSweep)] << endl;
		varLH_valid += LH_valid[i - (TotalSweep-KeptSweep)] * LH_valid[i - (TotalSweep-KeptSweep)];
		varLH_tot += (LH[i] + LH_valid[i-(TotalSweep-KeptSweep)]) * (LH[i] + LH_valid[i-(TotalSweep-KeptSweep)]);
		
		BF += 1.0/exp(LH[i]-LHmax);
		for (k = 0; k < K; k++)
		{
			dAIC[k] += AIC[i][k];
			dRSS[k] += RSS[i][k];
			varRSS[k] += RSS[i][k]*RSS[i][k];
		}
	}
	dAL /= NoS;
	varLH = varLH/(double)NoS - dAL*dAL;
	dAL_valid /= (double)NoS;
	varLH_valid = varLH_valid/(double)NoS - dAL_valid*dAL_valid;
	dAL_tot = dAL + dAL_valid;
	varLH_tot = varLH_tot/(double)NoS - dAL_tot*dAL_tot;
	BF /= ((double)NoS);
	for (k = 0; k < K; k++)
	{
		dAIC[k]/= NoS;
		dRSS[k]/=NoS;
		varRSS[k] = varRSS[k]/(double)NoS - dRSS[k]*dRSS[k];
	}

	string strAL = FOFN + "AL.txt";
	ofstream ofsAL(strAL.c_str());
	ofsAL << dAL;
	/* Comment, for compatibility with popsGUI
	ofsAL  << "\t Training Data" << endl;
	ofsAL << dAL_valid << "\t Validation Data" << endl;
	*/
	ofsAL.close();

	if (K>1)
	{
		string strAIC = FOFN + "AIC.txt";
		ofstream ofsAIC(strAIC.c_str());
		ofsAIC << "Cluster \t AIC " << endl;
		for (k = 0; k < (K-1); k++) {
			ofsAIC << k+1 << "\t";
			ofsAIC.width(17);
			ofsAIC.precision(15);
			ofsAIC << fixed;
			ofsAIC << dAIC[k] << endl;
		}
		ofsAIC << "ALL" << "\t";
		ofsAIC.width(17);
		ofsAIC.precision(15);
		ofsAIC << fixed; 
		ofsAIC << dAIC[K-1] << endl;
		ofsAIC.close();
	}
	

	string strBF = FOFN + "BF.txt";
	ofstream ofsBF(strBF.c_str());
	ofsBF << "(Likelihood Max)/P(Data)\t";
	ofsBF << BF << endl;
	ofsBF << "Log-Likelihood Max" << "\t";
	ofsBF << LHmax << endl;
	ofsBF.close();
	

	// Calculate DIC 
	double tmpDIC = 0.0; // will be calculated with estimated parameters
	for (i = 0; i < N; i++)
	{
		for (a = 0; a < A; a++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					if (X[i][a][l] == alleleV[l][j])
					{
						
						tmpDIC += log(spAverage.p[hc[i]][l][j]);
						break;
					}
				}
			}
		}
	}


	/*
	double tmpDIC_valid = 0.0; // will be calculated with estimated parameters
	// we should defined estimated parameters for valid set...
	for (i = 0; i < NV; i++)
	{
	for (a = 0; a < A; a++)
	{
		for (l = 0; l < LV; l++)
		{
		for (j = 0; j < J[l+L]; j++)
			{
			if (Xnewloci[i][a][l] == alleleV[l+L][j])
			{
				
				tmpDIC_valid += log(spAverage.p[hc[i]][l][j]);
				break;
			}
			}
		}
		}
	}
	*/

	// DIC = E[D(theta)] + penalty = E[D(theta)] + (E[D(theta)]- D(E[theta])) = E[D(theta)] + 0.5(Var[D(theta)])
	double DIC = -4*dAL + 2*tmpDIC; //-2*E[log lh] + (-2*E[log lh] + 2*log lh(estimated param))
	double DIC2 = -2*dAL + 2*varLH; //-2*E[log lh] + 0.5* 4*Var[log lh] 
	//double DIC2_valid = -2*dAL_valid + 2*varLH_valid;
	double DIC2_tot = -2*dAL_tot + 2*varLH_tot;
	string strDIC = FOFN + "DIC.txt";
	ofstream ofsDIC(strDIC.c_str());
	ofsDIC << DIC;
	/* Comment, for compatibility with popsGUI
	ofsDIC << << "\t (DIC1)" << endl;
	ofsDIC << DIC2 << "\t (DIC2 Training Data)" << endl;
	ofsDIC << DIC2_valid << "\t (DIC2 Validation Data)" << endl;
	ofsDIC << DIC2_tot << "\t (DIC2 All Data)" << endl;
	*/
	ofsDIC.close();


	// Calculate DICregr (DIC for the regresion)
	
	DBL_VCTR_1D DICregr;
	InitVctr1D(DICregr,K,0.0); //Last columns will contain the sum of DIC over the (K-1) regressions
	DBL_VCTR_1D DICregr2;
	InitVctr1D(DICregr2,K,0.0); //Last columns will contain the sum of DIC over the (K-1) regressions
	if (K>1)
	  {
	    if (beta_length >0)
	      {
		double RSSestimatedParam = 0.0; // = RSS(estimated beta)
		double penalty = 0.0; // = dRSS-RSS(estimated beta)
		double penalty2 = 0.0; // 0.5 * var(RSS)
		
		Mult_t(Pred,Xtild,spAverage.beta); //Xtild%*%t(estimated beta) Xtild= covariables
		
		string strCritSuiv = FOFN + "DICregr.txt";
		ofstream ofsCritSuiv(strCritSuiv.c_str());
		ofsCritSuiv << "Cluster \t DIC regression \t DIC2 regression \t dRSS \t RSSestimatedParam \t penalty \t penalty2"<<endl;
		
		bool NegativPenalty=false;
		for (k=0; k<(K-1); k++)
		  {
		    RSSestimatedParam=0.0;
		    for (i=0; i<N; i++)
		      {
			RSSestimatedParam += (spAverage.w[i][k]- Pred[i][k])*(spAverage.w[i][k]- Pred[i][k]);
		      }
		    penalty = dRSS[k]-RSSestimatedParam; // penalty = E[RSS] - RSS(estimated beta)
		    DICregr[k] = dRSS[k] + penalty; // DICregr = dRSS + penalty
		    DICregr[K-1] += DICregr[k];
		    //DICregr2 += E[RSS] + 0.5*var(RSS)
		    penalty2 = 0.5*varRSS[k];
		    DICregr2[k] = dRSS[k] + penalty2;
		    DICregr2[K-1] += DICregr2[k];
		    if (penalty<0.0 || penalty2<0.0) NegativPenalty=true;

	
		    ofsCritSuiv << k+1 << "\t";
		    ofsCritSuiv.width(17);
		    ofsCritSuiv.precision(15);
		    ofsCritSuiv << fixed;
		    ofsCritSuiv << DICregr[k] << "\t";
		    ofsCritSuiv.width(17);
		    ofsCritSuiv.precision(15);
		    ofsCritSuiv << fixed;
		    ofsCritSuiv << DICregr2[k] << "\t";
		    ofsCritSuiv.width(17);
		    ofsCritSuiv.precision(15);
		    ofsCritSuiv << fixed;
		    ofsCritSuiv << dRSS[k] << "\t";
		    ofsCritSuiv.width(17);
		    ofsCritSuiv.precision(15);
		    ofsCritSuiv << fixed;
		    ofsCritSuiv << RSSestimatedParam << "\t";
		    ofsCritSuiv.width(17);
		    ofsCritSuiv.precision(15);
		    ofsCritSuiv << fixed;
		    ofsCritSuiv << penalty <<endl;
		    ofsCritSuiv.width(17);
		    ofsCritSuiv.precision(15);
		    ofsCritSuiv << fixed;
		    ofsCritSuiv << penalty2 <<endl;
		 
		  }
		ofsCritSuiv << "ALL \t";
		ofsCritSuiv.width(17);
		ofsCritSuiv.precision(15);
		ofsCritSuiv << fixed;
		ofsCritSuiv << DICregr[K-1] << "\t";
		ofsCritSuiv.width(17);
		ofsCritSuiv.precision(15);
		ofsCritSuiv << fixed;
		ofsCritSuiv << DICregr2[K-1] << "\t";
		ofsCritSuiv.width(17);
		ofsCritSuiv.precision(15);
		ofsCritSuiv << fixed;
		ofsCritSuiv << dRSS[K-1] << endl;
		ofsCritSuiv.close();
		
		if (NegativPenalty)
		  {
		    string strDICregrError = FOFN + "DICregrError.txt";
		    ofstream ofsDICregrError(strDICregrError.c_str());
		    ofsDICregrError << "Error, negative penalty ! penalty1= " << penalty << "penalty2=" << penalty2 << endl;
		    ofsDICregrError.close();
		  }
	      }
	  }
	
	/*string strDICregr = FOFN + "DICregr.txt";
	//SaveToMyFile(DICregr,strDICregr);
	ofstream ofsDICregr(strDICregr.c_str());
	ofsDICregr << "Cluster \t DIC regression " << endl;
	for (k = 0; k < (K-1); k++) {
	ofsDICregr << k+1 << "\t";
	ofsDICregr.width(17);
	ofsDICregr.precision(15);
	ofsDICregr << fixed;
	ofsDICregr << DICregr[k] << endl;
	}
	ofsDICregr << "ALL" << "\t";
	ofsDICregr.width(17);
	ofsDICregr.precision(15);
	ofsDICregr << fixed; 
	ofsDICregr << DICregr[K-1] << endl;
	ofsDICregr.close();
	*/


	// Sum Up All Criteria in F+"_allCriteria" (see main)
	//1st line: "RUN \t  KEst \t LHtraining \t LHvalidation \t DIC \t DIC2 \t AIC \t DICregr \t DICregr2 \t DIC2_tot \t pseudoBF \t LHmax \t DICregr_k1 \t .. DICregr_(K-1) \t AIC_1 \t .. AIC_K-1 ";
	
	runCriteria.KEst=KEst;
	runCriteria.dAL=dAL;
	runCriteria.dAL_valid=dAL_valid;
	runCriteria.DIC=DIC;
	runCriteria.DIC2=DIC2;
	runCriteria.dAIC=dAIC;
	runCriteria.DICregr=DICregr;
	runCriteria.DICregr2=DICregr2;
	runCriteria.DIC2_tot=DIC2_tot;
	runCriteria.BF=BF;
	runCriteria.LHmax=LHmax;


	if (K>1)
	{
		ofsCriteria << KEst << "\t" << dAL << "\t" << dAL_valid << "\t" << DIC << "\t" << DIC2 << "\t" << runCriteria.cor << "\t" << dAIC[K-1] << "\t" << DICregr[K-1] << "\t" << DICregr2[K-1] << "\t" << DIC2_tot << "\t" << BF << "\t" << LHmax;
		for (k=0; k<(K-1); k++)
		{
			ofsCriteria << "\t" << DICregr[k];
		}
		for (k=0; k<(K-1); k++)
		{
			ofsCriteria << "\t" << dAIC[k];
		}
		if (bFIT) 
		{
			ofsCriteria << "\t" << trainingConfig.name << endl;
		}
		else
		{
			ofsCriteria << "\t AllData" << endl;
		}
	}
	else
	{
		ofsCriteria << KEst << "\t" << dAL << "\t" << dAL_valid << "\t" << DIC << "\t" << DIC2 << "\t" << runCriteria.cor << "\t" << "NA" << "\t" << "NA" << "\t" << "NA" << "\t" << DIC2_tot << "\t" << BF << "\t" << LHmax;
		if (bFIT) 
		{
			ofsCriteria << "\t" << trainingConfig.name << endl;
		} 
		else
		{
			ofsCriteria << "\t AllData" << endl;
		}	  
	}
	FreeRNG(rng);
	//if (loop_sup_PROPOSAL_MAX>0)
	//{
	if (sample_struct != NULL)
	 {
		gsl_ran_discrete_free(sample_struct);
		//free(sample_struct);
	}
	clock_t e_clock = clock();

	cout << fixed;
	cout.precision(3);
	cout << "This run takes " << static_cast<double>(e_clock - s_clock) /
	CLOCKS_PER_SEC << " seconds to complete on this computer." << endl;
	
	return 1;
}


int main(int argc, char* argv[]){
	int test_inf=1;
	int test_sup=1;
	double cvScore=0;
	bool bF=false;
	bool bK=false;
	bool bN=false;
	bool bL=false;
	bool bg=false;
	bool bh=false;
	bool bFIT=false;
	bool bValOnIndiv=false;
	bool b2fold=false;        // use of 2-fold Cross-Validation Methodology
	string Fline,F,FindexTraining,F_test,run,FoO;
	string emptyRunString="";
	int aF,i,K,N,L,runDone;
	vector <CTrainingConfig> vTrainingConfig,vCompTrainingConfig;
	CTrainingConfig emptyTrainingConfig,currentTConfig,currentCompTConfig;
	CRunCriteria runCriteria;

	// Print comand line
	if (argc > 0) 
	{
		for (int j = 0; j < argc; j++) 
		{
			cout << argv[j] << " ";
		}
		cout << endl;
		cout << "**************************************************************" << endl;
		
	}
	  

	//	if (strcmp((const char*) &argv[1][1],"predict")==0)
	if ( (argc > 1) && (strcmp(argv[1],"-predict")==0) )
	  {
	 
	    predictionFromCommandLine(argc,argv);
	    return 0;
	  }
	cout << "Regular Usage" << endl;


	if (argc < 9)
	{
	        cout << "Only " << argc-1 << " arguments given, this is not enough." << endl;
		ShowUsage();
		cout << "Only " << argc-1 << " arguments given, this is not enough." << endl;
		exit(EXIT_FAILURE);
	}


	// Parse the command line parameters.
		for(int ac = 1; ac < argc; ac++)
		{

			switch (argv[ac][1])
			{
			case 'F':
				// File Training
				Fline = (const char*) &argv[ac][0]; //All F line (-FT...)
				F = (const char*) &argv[ac][2]; //Dir+File Name
				aF=ac;
				bF = true;
				break;	      
			case 'K':
				K=atoi((const char*) &argv[ac][2]);
				bK=true;
				break;
			case 'N':
				N=atoi((const char*) &argv[ac][2]);
				bN=true;
				break;
			case 'L':
				L=atoi((const char*) &argv[ac][2]);
				bL=true;
				break;
				
			case 'v':
				// Run with Validation
				// File containing the Training Configuration(s)
				FindexTraining = (const char*) &argv[ac][2];
				bFIT = true;
				break;
			case '2':
				if (strcmp((const char*) &argv[ac][2],"fold")==0)
				{
					b2fold=true;
					break;
				}
				else 	
				{
					cout << "Unknown Option - " << argv[ac] << " For more details: ./pops " << endl;
					exit(EXIT_FAILURE);
				}
			case 'g':
				test_inf=atoi((const char*) &argv[ac][2]);
				bg=true;
				break;
			case 'h':
				test_sup=atoi((const char*) &argv[ac][2]);
				bh=true;
				break;
			
			case 'o':
			  if (argv[ac][2] == 'r' && argv[ac][3] == 'u' && argv[ac][4] == 'n') {
			    run = (const char*) &argv[ac][5];
			  } else {
			    FoO = (const char*) &argv[ac][2];
			  }
			  break;

			default:
				break;
			}
		}
	

	if (!bF || !bN || !bL || (bg&&!bh) || (bh&&!bg)){
		ShowUsage();
	        cout << "Mandatory argument -F, -L or -N is missing, or only one of the option -g -h was specified" << endl;
		exit(EXIT_FAILURE);
	}

	string FOFN = !FoO.empty() ? FoO + "/" + F : F;
	if (!run.empty())
	  {
	    FOFN=FOFN+"_RUN"+run+"_";
	  }
	stringstream allCriteria,FCVScore;
	if (bg)
	{
	  if (!run.empty())
	    {
	      cout<< "You can not use -orun parameter at the same time as -g -h parameters" <<endl;
	      exit(EXIT_FAILURE);
 	    }
	  allCriteria << FOFN << "_g" << test_inf << "h" << test_sup << "_allCriteria.txt";
	  FCVScore << FOFN << "_g" << test_inf << "h" << test_sup << "_cvScore.txt";
	  
	}
	else
	{
		allCriteria << FOFN << "_allCriteria.txt";
		FCVScore << FOFN << "_cvScore.txt";

	}
	ofstream ofsCriteria((allCriteria.str()).c_str());
	ofsCriteria << "RUN \t  KEst \t LHtraining \t LHvalidation \t DIC \t DIC2 \t COR \t AIC \t DICregr \t DICregr2 \t DIC2_tot \t pseudoBF \t LHmax";
	for (int k=1; k<K; k++) 
	{
		ofsCriteria << "\t DICregr_" << k; 
	}
	for (int k=1; k<K; k++) 
	{
		ofsCriteria << "\t AIC_" << k; 
	}
	ofsCriteria << "\t TrainingConfiguration" << endl;

	ofstream ofsCVScore((FCVScore.str()).c_str());
	ofsCVScore << "RUN \t CVscore \t 2FoldPartition" << endl;

	if (bFIT) 
	{
		LoadTraining(vTrainingConfig,vCompTrainingConfig,bValOnIndiv,N,L,FindexTraining);
		string ValidationMethod="loci";
		if (bValOnIndiv) ValidationMethod="individuals";
		cout << "Validation on " << ValidationMethod << ": " << vTrainingConfig.size() << " training configuration(s) loaded." << endl;

		/* for (int i=0; i<vTrainingConfig.size(); i++)
	{
	vTrainingConfig[i].Print();
	cout<<endl;
	vCompTrainingConfig[i].Print();
	cout<<endl;
	} */
	}

	if (!bg)
	{

		if (bFIT)
		{
			if (b2fold){
				
				cvScore=0.0;
				ofsCriteria << F << "a" << "\t";
				pops(argc,argv,"RUNa",ofsCriteria,runCriteria,bFIT,bValOnIndiv,vTrainingConfig[0]);
				cvScore += runCriteria.dAL_valid/(vCompTrainingConfig[0].size());

				ofsCriteria << F << "b" << "\t";
				pops(argc,argv,"RUNb",ofsCriteria,runCriteria,bFIT,bValOnIndiv,vCompTrainingConfig[0]);
				cvScore += runCriteria.dAL_valid/(vTrainingConfig[0].size());
				cvScore /= 2;
				
				ofsCVScore << F << "\t" << cvScore << "\t" << vTrainingConfig[0].name << endl;
			}
			else
			{
				ofsCriteria << F << "\t";
				pops(argc,argv,emptyRunString,ofsCriteria,runCriteria,bFIT,bValOnIndiv,vTrainingConfig[0]);
			}
		}
		else
		{
			pops(argc,argv,emptyRunString,ofsCriteria,runCriteria,bFIT,bValOnIndiv,emptyTrainingConfig);
		}
	}
	else
	{
		for (i=test_inf; i<=test_sup; i++)
		{
			string run;
			stringstream out;
			out << i;
			run = out.str();
			runDone=0;
			
			if (bFIT)
			{
				if (b2fold)
				{
					cvScore=0.0;
					currentTConfig = vTrainingConfig[(i-test_inf)%vTrainingConfig.size()];
					currentCompTConfig = vCompTrainingConfig[(i-test_inf)%vCompTrainingConfig.size()];

					ofsCriteria << i << "a" << "\t";
					while (runDone==0){
						runDone=pops(argc,argv,run+"a",ofsCriteria,runCriteria,bFIT,bValOnIndiv,currentTConfig);
					}
					runDone = 0;
					cvScore += runCriteria.dAL_valid/(currentCompTConfig.size());
					
					ofsCriteria << i << "b" << "\t";
					while (runDone==0){
						runDone=pops(argc,argv,run+"b",ofsCriteria,runCriteria,bFIT,bValOnIndiv,currentCompTConfig);
					}
					runDone = 0;
					cvScore += runCriteria.dAL_valid/(currentTConfig.size());
					cvScore /= 2;

					ofsCVScore << i << "\t" << cvScore << "\t" << currentTConfig.name << endl;
				}
				else
				{
					ofsCriteria << i << "\t";
					currentTConfig = vTrainingConfig[(i-test_inf)%vTrainingConfig.size()];
					while (runDone==0)
					{
						runDone=pops(argc,argv,run,ofsCriteria,runCriteria,bFIT,bValOnIndiv,currentTConfig);
					}
					runDone=0;
				}
			}
			else
			{
				ofsCriteria << i << "\t";
				while (runDone==0)
				{
					runDone=pops(argc,argv,run,ofsCriteria,runCriteria,bFIT,bValOnIndiv,emptyTrainingConfig);
				}
				runDone=0;
			}
		}
	}


	ofsCriteria.close();
	ofsCVScore.close();


	if (!b2fold) {
		const string stringFCVScore = FCVScore.str();
		remove(stringFCVScore.c_str());
	}

	return 0;

}
