#ifndef __IO_UTILITIES__
#define __IO_UTILITIES__

/*! \file IOUtilities.h
 *  \brief Header File for Input/Output Utility Routines
 */

#include <string>

#include "VectorUtilities.h"

using namespace std;

template < typename T > bool SaveVector1D(vector< T >& V, const string& File)
{
    if (V.size() == 0)
    {
        cerr << "Uninitialized mat - SaveDBLMatrix1D" << endl;
        exit(EXIT_FAILURE);
    }
    int L = static_cast<int>(V.size());
    int l;
    ofstream ofs(File.c_str());
    if (!ofs.good())
    {
        cerr << "Cannot open file " << File << " - SaveVector1D" << endl;
        exit(EXIT_FAILURE);
    }
    for (l = 0; l < L; l++)
    {
        ofs << V[l]<< '\t' ;
    }
    ofs.close();

    return true;
}

template < typename T > bool SaveVector2D(vector< vector< T > >& V, const string& File)
{
    if (V.size() == 0)
    {
        cerr << "Uninitialized mat - SaveVector2D" << endl;
        exit(EXIT_FAILURE);
    }
    int L = static_cast<int>(V.size());
    int C = static_cast<int>(V[0].size());
    int l,c;
    ofstream ofs(File.c_str());
    if (!ofs.good())
    {
        cerr << "Cannot open file " << File << " - SaveVector2D" << endl;
        exit(EXIT_FAILURE);
    }
    for (l = 0; l < L; l++)
    {
        for (c = 0; c < C; c++)
        {
            ofs << V[l][c] << '\t' ;
        }
        ofs << endl;
    }
    ofs.close();

    return true;
}

template < typename T > bool LoadVector2D(vector< vector< T > >& V, const string& File)
{
    if (V.size() == 0)
    {
        cerr << "Uninitialized mat - LoadVector2D" << endl;
        exit(EXIT_FAILURE);
    }
    int L = static_cast<int>(V.size());
    int C = static_cast<int>(V[0].size());
    int l,c;
    ifstream ifs(File.c_str());
    if (!ifs.good())
    {
        cerr << "Cannot open file " << File << " - LoadVector2D" << endl;
        exit(EXIT_FAILURE);
    }
    for (l = 0; l < L; l++)
    {
        for (c = 0; c < C; c++)
        {
            ifs >> V[l][c] ;
        }
    }
    ifs.close();

    return true;
}

//! Saving Genotype Data to a STRUCTURE Format File
/*! \param Xtild Individuals covariates
 *  \param X Genotype Data
 *  \param File The Output File Name
 *  \return Result of the Action (Success/Failure)
 *  \author Durand Eric
 */
bool SaveData(const DBL_VCTR_2D& Xtild, const INT_VCTR_3D& X, const string& File);



//! Saving the Posterior Distribution Regression Coefficients in a File (no more than 1000 sweeps are kept)
/*! \param Beta Vector containing the Regression Coefficients
 *  \param File The Output File Name
 *  \author Jay Flora
 */
void SaveBetaToMyFile(const DBL_VCTR_3D& Beta, const string& File, const int BurnInPeriod);


//! Saving an object of type DBL_VTR_2D in a File in a Format compatible with CLUMPP
/*! \param V Vector
 *  \param File The Output File Name
 *  \param header boolean: Write or not an header (title + cluster labels)
 *  \param title
 *  \author Jay Flora
*/
void SaveToMyFileClumppFormat(const DBL_VCTR_2D& V, const string& File, const bool header=false, const string title="");


//! Writting an object of type DBL_VTR_2D in an Output File Stream in a Format compatible with CLUMPP
/*! \param V Vector
 *  \param ofs Output File Stream 
 *  \author Jay Flora
*/
void WriteInOfstreamClummpFormat(const DBL_VCTR_2D V, ofstream& ofs);





//! Loading Genotype and Covariates Data
/*! \param bbase boolean=true if user chose to assign some individuals to the base cluster (Kth cluster)
 *  \param isBase isBase[i]=1 if individual i is compulsorily assigned to the base cluster (if bbase=false isBase[i]=0 for all i)
 *  \param C Coordinate of Individuals
 *  \param Xtild Environmental Data
 *  \param XQual Qualitative Variables (eg Language, Habitat)
 *  \param LevelPopSize Number of Individuals in Each Class of Each Qualitative Variable
 *  \param nQualLevels
 *  \param X Genotype Data
 *  \param recAllele
 *  \param ER Number of Extra Rows
 *  \param EC Number of Extra Columns
 *  \param File Input Data
 *  \param specialFormat If true, individuals are stored in "one line per individual" format
 *  \param nbQuantVar Number of Quantitative Variable (other than spatial coordinates)
 *  \param nb_var_regr Number of Spatial Coordinates used for the Trend(0,1,2)
 *  \param isRecAllele
 *  \param trendDegree Degree of the Spatial Trend (if -1 data do NOT contained spatial coord.)
 *  \return Result of the Action (Success/Failure)
 *  \author Durand Eric, Jay Flora 
 */
bool LoadData(const bool bbase, INT_VCTR_1D& isBase,DBL_VCTR_2D& C, DBL_VCTR_2D& Xtild, INT_VCTR_2D& XQual, INT_VCTR_2D& LevelPopSize, INT_VCTR_1D& nQualLevels, INT_VCTR_3D& X, INT_VCTR_1D& recAllele, int ER, int EC, const string& File, const bool& specialFormat, const int& nbQuantVar, const int& nb_var_regr, const bool& isRecAllele, const int& trendDegree);


//! To Interpret Input Data As Qualitative Covariate Variables
/*! 
 *  \author Jay Flora
 */

bool TransformQualData(INT_VCTR_2D& XQual, DBL_VCTR_2D& XQuant, DBL_VCTR_2D& Xtild, INT_VCTR_1D& nQualLevels, INT_VCTR_2D& LevelPopSize);



//! Saving Configurations for "Continuing Previous Run"
/*! \param z Membership of Individuals (Hard Clustering)
 *  \param p Estimated Allele Frequencies
 *  \param File The Output File Name
 *  \return Result of the Action (Success/Failure)
 *  \author Durand Eric
 */

bool SaveCF(const DBL_VCTR_1D& sigma, const DBL_VCTR_2D& beta, const DBL_VCTR_2D& alpha, const DBL_VCTR_2D& q, const INT_VCTR_3D& zadmix, const DBL_VCTR_3D& p, const string& File);

bool SaveML(const int& trendDeg, const double& minDIC, const string& File);
bool LoadML(int& trendDeg, double& minDIC ,const string& File,bool quiet);

//! Loading Configurations for "Continuing Previous Run"
/*! \param z Membership of Individuals (Hard Clustering)
 *  \param p Estimated Allele Frequencies
 *  \param File The Input File Name
 *  \return Result of the Action (Success/Failure)
 *  \author Durand Eric
 */
bool LoadCF(DBL_VCTR_1D& sigma, DBL_VCTR_2D& beta, DBL_VCTR_2D& alpha, DBL_VCTR_2D& q, INT_VCTR_3D& zadmix, DBL_VCTR_3D& p, const string& File);
#endif //__IO_UTILITIES__
