#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "MatrixUtilities.h"
#include <fstream>
#include <iostream>

//A optimiser
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_permutation.h>

void DimensionErrorMsg() {
  printf("Error in dimensions, most likely due to a mismatch between the input datafile format and command line parameters. \nPlease check whether you have correctly input:\n");
  printf("%s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n %s\n",
	 "Number of Individuals",
	 "Number of Dummy Individuals",
	 "Number of Loci",
	 "Number of Qualitative Covariates",
	 "Number of Quantitative Covariates (except spatial coordinates)",
	 "Presence or Absence of Spatial Coordinates",
	 "Number of Extra Columns",
	 "Number of Extra Rows",
	 "Ploidy",
	 "Special Format Option for Data Stored in 1 Line per Individual instead of 2");
  printf( "For more details: ./popsAdm or ./pops \n");
}


void CopyVctr(const DBL_VCTR_2D& v1,DBL_VCTR_2D& v2)
{
  	// initialise seulement si v2 n'avait pas encore ete initialise
  	// sinon copie en ecrivant "par dessus" v2
	int i;
	if ((int)v2.size()==0)
	{
		for (i = 0; i < (int)v1.size(); i++)
		{
      			//printf("CopyVctr : v2 initialisation\n");
			v2.push_back(v1[i]);
		}
	}
	else
	{
		if (v2.size()!=v1.size()||v2[0].size()!=v1[0].size())
		{
			printf("CopyVctr : error, dimensions don't match. Exiting...\n");
			DimensionErrorMsg();
			exit(EXIT_FAILURE);
		}
    		//printf("CopyVctr : v2 rewritted\n");
		for (i = 0; i < (int)v1.size(); i++) 
		{
			v2[i]=v1[i];
		}
	}
}

void choldc(DBL_VCTR_2D& matrix, const DBL_VCTR_2D mat){
	int i,j,k;                         /* loop variables */
  	double sum;                         /* Temp variable */
  	
	int n=mat.size();
  	if (n != (int)mat[0].size())
  	{
    		printf("choldc : error, wrong matrix dimension");
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}
  	if ( ((int)matrix.size()!=0) && (((int)matrix.size()!=n)||((int)matrix[0].size()!=n)) )
  	{
    		printf("choldc : error, cholesky matrix is already initialized with wrong dimensions. Exiting...\n");
		DimensionErrorMsg();
    		exit(EXIT_FAILURE);
  	}
  	CopyVctr(mat,matrix);
  	for( i=0; i<n; i++ )
  	{
  		for( j=i; j<n; j++ )
		{
			sum = matrix[i][j];
			k = i;
	  		while ( --k >= 0 )
	  		{
	    			sum -= matrix[i][k] * matrix[j][k];
		  	}
		  	if ( i==j )
	   		{
	      			if (sum <= 0.0)
				{
		  			printf("Cholesky Decomposition Failed! Consider lowering the trend degree. Exiting...\n");
					DimensionErrorMsg();
		  			exit(EXIT_FAILURE);
				}
	      			else 
				{
		  			matrix[i][i] = sqrt( sum );
				}
	    		}
	  		else
	    		{
	      			matrix[j][i] = sum / matrix[i][i];
	    		}
		}
  	}
  	for(i=0; i<n; i++ )
  	{
    		for(j=i+1; j < n; j++)
    		{
      			matrix[i][j]=0;     
    		}
  	}
}


void buildNeighborhoodMatrix(DBL_VCTR_2D& W, const INT_VCTR_2D& NS)
{
	if (W.size() == 0)
	{
		cerr << "Uninitialized W - buildNeighborhoodMatrix" << endl;
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	if (NS.size() == 0)
	{
		cerr << "Uninitialized NS - buildNeighborhoodMatrix" << endl;
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	
	
	int N = static_cast<int>(W.size());
	int i,j;
	for (i = 0; i < N; i++)
	{
		int J = static_cast<int>(NS[i].size());
		W[i][i] = 0.0;
		for (j = 0; j < J; j++)
		{
			W[i][NS[i][j]] = 1.0;
		}
	}
}

void buildNeighborhoodMatrix(DBL_VCTR_2D& W, const INT_VCTR_2D& NS, const DBL_VCTR_2D& geo, const double theta)
{
	if (W.size() == 0)
	{
		cerr << "Uninitialized W - buildNeighborhoodMatrix" << endl;
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	if (NS.size() == 0)
	{
		cerr << "Uninitialized NS - buildNeighborhoodMatrix" << endl;
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	
	int N = static_cast<int>(W.size());
	int i,j;
	for (i = 0; i < N; i++)
	{
		int J = static_cast<int>(NS[i].size());
		W[i][i] = 0.0;
		for (j = 0; j < J; j++)
		{
			W[i][NS[i][j]] = (double)(exp(-geo[i][NS[i][j]]/theta));
		}
	}
}

void computeEigenvalues(DBL_VCTR_1D& eigenVal, const DBL_VCTR_2D& A)
{
	if (A.size() == 0)
	{
		cerr << "Uninitialized A - computeEigenvalues" << endl;
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	
	if (eigenVal.size() == 0)
	{
		cerr << "Uninitialized eigenVal - computeEigenvalues" << endl;
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	
	int N = static_cast<int>(A.size());
	gsl_matrix* gslMat = gsl_matrix_alloc(N,N);
	gsl_vector* gslVect = gsl_vector_alloc(N);
	int i,j;
	
	//copy A to gslMat
	for (i = 0; i < N; i++)
	{
		for (j = 0; j < N; j++)
		{
			double x = A[i][j];
			gsl_matrix_set(gslMat,i,j,x);
		}
	}
	
	gsl_eigen_symm_workspace* wsp = gsl_eigen_symm_alloc(N);
	gsl_eigen_symm(gslMat,gslVect,wsp);
	gsl_sort_vector(gslVect);
	
	for (i = 0; i < N; i++)
	{
		eigenVal[i] = gsl_vector_get(gslVect,i);
	}
	
	gsl_matrix_free(gslMat);
	gsl_vector_free(gslVect);
	gsl_eigen_symm_free(wsp);
}

void computeSymmMatrixStat(double& stat, const DBL_VCTR_2D& Matrix, const string option)
{
	if (Matrix.size() == 0)
	{
		cerr << "Uninitialized Matrix - computeSymmMatrixStat" << endl;
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	
	
	int N = static_cast<int>(Matrix.size());
	int i,j;
	int cpt = 0;
	int size = (N*(N+1)/2); //size of upper triangle
	gsl_vector* gslVect = gsl_vector_alloc(size);
	for (i = 0; i < N; i++)
	{
		for (j = i; j < N; j++)
		{
			double tmp = Matrix[i][j];
			gsl_vector_set(gslVect,cpt,tmp);
			cpt +=1;
		}
	}
	
	bool goodOpt = false;
	if (option == "m") 
	{
		stat = gsl_stats_mean(gslVect->data,gslVect->stride,size);
		goodOpt = true;
	}
	
	if (option == "d")
	{
		gsl_sort_vector(gslVect);
		stat = gsl_stats_median_from_sorted_data(gslVect->data,gslVect->stride,size);
		goodOpt = true;
	}
	
	if (option == "x")
	{
		stat = gsl_stats_max(gslVect->data,gslVect->stride,size);
		goodOpt = true;
	}
	
	if (!goodOpt)
	{
		cerr << "Unknown option for scale parameter - computeSymmMatrixStat" << endl;
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	gsl_vector_free(gslVect);
	
}

void PrintVctr(const DBL_VCTR_2D& V)
{
  	int i,j;
  	printf("\n\n");
  	for (i=0;i<(int)V.size();i++)
  	{
    		for (j=0;j<(int)V[0].size();j++)
    		{
      			printf("%f  ",V[i][j]);
    		}
    		printf("\n");
  	}
  	printf("\n");
  	cout<<" "<<endl;
}
void PrintVctr(const INT_VCTR_1D& V)
{
  	int i;
  	printf("\n\n");
  	for (i=0; i<(int)V.size(); i++)
  	{
    		printf("%d\n",V[i]);
  	}
}
void PrintVctr(const DBL_VCTR_1D& V)
{
  	int i;
  	printf("\n\n");
  	for (i=0; i<(int)V.size(); i++)
  	{
    		printf("%f  \n",V[i]);
  	}
}

void MultivariateGauss(DBL_VCTR_1D& mu, DBL_VCTR_2D& v_chol, DBL_VCTR_1D& Y, gsl_rng*& rng)
{
  	DBL_VCTR_1D X; 
  	int k,i,j;
  	int n=v_chol.size();
  	for(k=0; k<n; k++)
	{
    		X.push_back(gsl_ran_ugaussian(rng));
	}
  	if ((int)Y.size()==0)
  	{
    		// Y has not been initialized yet
    		for (i =0;i<n;i++)
    		{
      			Y.push_back(mu[i]);
      			for (j=0;j<n;j++)
      			{
				Y[i]+=v_chol[i][j]*X[j];
      			}
    		}
  	}
  	else
  	{
    		for (i =0;i<n;i++)
    		{
      			Y[i]=mu[i];
      			for (j=0;j<n;j++)
      			{
				Y[i]+=v_chol[i][j]*X[j];
      			}
    		}
  	}
}

void Mult(DBL_VCTR_1D& Res, const DBL_VCTR_2D& A, const DBL_VCTR_1D& B)
{
	  // A%*%B
 	 double sum;
 	 int i,j,k;
 	 if (A[0].size()!=B.size())
	 {
    		cout<<"Mult: Dimensions des matrices a multiplier inadequates. Exiting..."<<endl;
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
 	 }
  	int L=(int)A.size();
  	if ((int)Res.size()==0)
	{
    		// Res was not initialized
    		for (k=0; k<L; k++)
		{
      			Res.push_back(0.0);
		}
  	}   
  	else if((int)Res.size()!=L)
	{
  		cout<<"Mult: Dimensions de la matrice resultat inadequates. Exiting..."<<endl;
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}
  	for (i=0; i<L; i++)
	{
    		sum=0.0;
    		for (j=0; j<(int)B.size(); j++)
		{
      			sum+=A[i][j]*B[j];
    		}
    		Res[i]=sum;
  	}
}
void Mult(DBL_VCTR_2D& Res, const DBL_VCTR_2D& A, const DBL_VCTR_2D& B)
{
	  // A%*%B
	  int lig,col,tmp;
  	if (A[0].size()!=B.size())
	{ 
    		cout<<A[0].size()<<" vs "<<B.size()<<endl;
    		cout<<"Mult: Dimensions des matrices a multiplier inadequates. Exiting..."<<endl;
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}
  	int L=(int)A.size();
  	int C=(int)B[0].size();
  	int T=(int)B.size();
  	if ((int)Res.size()==0)
	{
    		// Res was not initialized
    		InitVctr2D(Res,L,C,0.0);
  	}
  	else if ( ((int)Res.size()!=L) || ((int)Res[0].size()!=C) )
	{
    		cout<<"Mult: Dimensions de la matrice resultat inadequates. Exiting..."<<endl;
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}
  	for (lig=0; lig<L; lig++)
	{
    		for (col=0; col<C; col++)
		{
      			Res[lig][col]=0;
      			for (tmp=0; tmp<T; tmp++)
			{
				Res[lig][col]+=A[lig][tmp]*B[tmp][col];
      			}
    		}
  	}  
}
void Mult_t(DBL_VCTR_2D& Res, const DBL_VCTR_2D& A, const DBL_VCTR_2D& B)
{
  	// A%*%t(B)
  	int lig,col,tmp;
	if (A[0].size()!=B[0].size()) //on multiplie avec la ligne puisqu'on veut la transposee
  	{ 
    		cout<<A[0].size()<<" vs "<<B[0].size()<<endl;
    		cout<<"Mult_t: Dimensions des matrices a multiplier inadequates. Exiting..."<<endl;
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}
  	int L=(int)A.size();
  	int C=(int)B.size();
  	int T=(int)B[0].size();
  	if ((int)Res.size()==0)
  	{
    		// Res was not initialized
    		InitVctr2D(Res,L,C,0.0);
  	}
  	else if ( ((int)Res.size()!=L) || ((int)Res[0].size()!=C) )
  	{
    		cout<<"Mult_t: Dimensions de la matrice resultat inadequates. Exiting..."<<endl;
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}

    	for (lig=0; lig<L; lig++)
    	{
      		for (col=0; col<C; col++)
      		{
			Res[lig][col]=0;
      			for (tmp=0; tmp<T; tmp++)
      			{
				Res[lig][col]+=A[lig][tmp]*B[col][tmp];
      			}
    		}
  	}  
}
void Mult_t_left(DBL_VCTR_2D& Res, const DBL_VCTR_2D& A, const DBL_VCTR_2D& B)
{
  // t(A)%*%B
  	int lig,col,tmp;
  	if (A.size()!=B.size())
  	{ 
    		cout<<A.size()<<" vs "<<B.size()<<endl;
    		cout<<"Mult_t_left: Dimensions des matrices a multiplier inadequates. Exiting..."<<endl;
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}
  	int L=(int)A[0].size();
  	int C=(int)B[0].size();
  	int T=(int)B.size();
  	if ((int)Res.size()==0)
  	{
    		// Res was not initialized
    		InitVctr2D(Res,L,C,0.0);
  	}
  	else if ( ((int)Res.size()!=L) || ((int)Res[0].size()!=C) )
  	{
    		cout<<"Mult_t_left: Dimensions de la matrice resultat inadequates. Exiting..."<<endl;
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}
  	for (lig=0; lig<L; lig++)
  	{
    		for (col=0; col<C; col++)
    		{
      			Res[lig][col]=0;
      			for (tmp=0; tmp<T; tmp++)
      			{
				Res[lig][col]+=A[tmp][lig]*B[tmp][col];
      			}
    		}
  	}  
}

void Mult_byConstant(DBL_VCTR_2D& result, const DBL_VCTR_2D& mat, const double constant)
{
	int NL = static_cast<int>(mat.size());
	int NC = static_cast<int>(mat[0].size());
	if (NL != static_cast<int>(result.size()) || NC != static_cast<int>(mat[0].size()))
	{
		printf("Mult_byConstant : error, dimensions don't match. Exiting...\n");
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	int i,j;
	for (i = 0; i < NL; i++)
	{
		for (j = 0; j < NC; j++)
		{
			result[i][j] = constant*mat[i][j];
		}
	}
}

//result = mat1 + mat2
void Sum(DBL_VCTR_2D& result, const DBL_VCTR_2D& mat1, const DBL_VCTR_2D& mat2)
{
	int NL = static_cast<int>(mat1.size());
	int NC = static_cast<int>(mat1[0].size());
	if (NL != static_cast<int>(result.size()) || NL != static_cast<int>(mat2.size()) || NC != static_cast<int>(result[0].size()) || NC != static_cast<int>(mat2[0].size()))
	{
		printf("Sum : error, dimensions don't match. Exiting...\n");
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	int i,j;
	for (i = 0; i < NL; i ++)
	{
		for (j = 0; j < NC; j ++)
		{
			result[i][j] = mat1[i][j] + mat2[i][j];
		}
	} 
	
}

//result = mat1 - mat2
void Diff(DBL_VCTR_2D& result, const DBL_VCTR_2D& mat1, const DBL_VCTR_2D& mat2)
{
	int NL = static_cast<int>(mat1.size());
	int NC = static_cast<int>(mat1[0].size());
	if (NL != static_cast<int>(result.size()) || NL != static_cast<int>(mat2.size()) || NC != static_cast<int>(result[0].size()) || NC != static_cast<int>(mat2[0].size()))
	{
		printf("Diff : error, dimensions don't match. Exiting...\n");
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	int i,j;
	for (i = 0; i < NL; i ++)
	{
		for (j = 0; j < NC; j ++)
		{
			result[i][j] = mat1[i][j] - mat2[i][j];
		}
	} 
	
}

//result = t(mat1) - (mat2)
void Diff_t_left(DBL_VCTR_2D& result, const DBL_VCTR_2D& mat1, const DBL_VCTR_2D& mat2)
{
	int NL = static_cast<int>(mat2.size());
	int NC = static_cast<int>(mat2[0].size());
	if (NL != static_cast<int>(result.size()) || NL != static_cast<int>(mat1[0].size()) || NC != static_cast<int>(result[0].size()) || NC != static_cast<int>(mat1.size()))
	{
		printf("Diff_t_left : error, dimensions don't match. Exiting...\n");
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	int i,j;
	for (i = 0; i < NL; i ++)
	{
		for (j = 0; j < NC; j ++)
		{
			result[i][j] = mat1[j][i] - mat2[i][j];
		}
	}
}

double FindBound(const DBL_VCTR_1D& Wi, int k)
{
  	//search for max(W[i,-k],0)
  	int j;
  	double max=0.0;
  	for (j=0; j<(int)Wi.size(); j++)
  	{
    		if ((j!=k) && (Wi[j]>max))
		{
      			max=Wi[j];
		}
  	}
  	return(max);
}

void Inverse(DBL_VCTR_2D& res, const DBL_VCTR_2D& A)
{
	if (A.size()!=A[0].size())
	{
		printf("Inverse : error, matrice is not square. Exiting...\n");
		DimensionErrorMsg();
		exit(EXIT_FAILURE);
	}
	int n=(int)A.size();

  	//printf("dimensions n=%d\n",n);
  	int i,j,k;
  	DBL_VCTR_2D MM,M1,Mi1;
  	CopyVctr(A,MM);
  	CopyVctr(A,M1);
  	//Mi1 and res initialized with identity matrice 
  	InitVctr2D(Mi1,n,n,0.0);
  	for (i=0; i<n; i++)
  	{
    		Mi1[i][i]=1.0;
  	}
  	CopyVctr(Mi1,res);
  	for (i=0;i<n;i++)
  	{
    		for (j=0;j<n;j++)
    		{
      			if (MM[i][i]==0)
      			{
				printf("Inverse : inversion with gauss method impossible, division by 0\n");

				DimensionErrorMsg();
				exit(EXIT_FAILURE);
      			}
      			M1[i][j] = MM[i][j]/MM[i][i];
      			Mi1[i][j] = res[i][j]/MM[i][i];
    		}
    		//printf("MM.size() %d", (int)MM.size());
    		CopyVctr(M1,MM);
    		//printf("res.size() %d", (int)res.size());
    		CopyVctr(Mi1,res);
		for (k=0;k<n;k++) // mise zéros de la colonne
		{
      			if (k!=i)
			{
				for (j=0;j<n;j++)
				{
	  				M1[k][j]=MM[k][j]-MM[i][j]*MM[k][i];
	  				Mi1[k][j]=res[k][j]-res[i][j]*MM[k][i];
				}
			}
		}
    		CopyVctr(M1,MM);
    		CopyVctr(Mi1,res);    
  	}
}

void InverseWithChol(DBL_VCTR_2D& res, const DBL_VCTR_2D& A)
{
  	if (A.size()!=A[0].size())
  	{
    		printf("InverseWithChol : error, matrice is not square. Exiting...\n");
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}
  	int N=(int)A.size();
	
	
	gsl_matrix* gslMat = gsl_matrix_alloc(N,N);
	gsl_matrix* gslResult = gsl_matrix_alloc(N,N);
	gsl_matrix_set_identity(gslResult);
	gsl_vector* gslVect = gsl_vector_alloc(N);
	int i,j;
	
	//copy A to gslMat
	for (i = 0; i < N; i++)
	{
		for (j = 0; j < N; j++)
		{
			double x = A[i][j];
			gsl_matrix_set(gslMat,i,j,x);
		}
	}
	
	gsl_linalg_cholesky_decomp(gslMat);
	
	for (i = 0; i < N; i ++)
	{
		for (j = 0; j < N; j ++)
		{
			gsl_vector_set(gslVect,j,gsl_matrix_get(gslResult,i,j));
		}
		gsl_linalg_cholesky_svx(gslMat, gslVect);
		for (j = 0; j < N; j ++)
		{
			gsl_matrix_set(gslResult,i,j,gsl_vector_get(gslVect,j));
		}
		
	}
	
	//Copy gslMat to result
	for (i = 0; i < N; i++)
	{
		for (j = 0; j < N; j ++)
		{
			res[i][j] = gsl_matrix_get(gslResult,i,j);
		}
	}	

	gsl_matrix_free(gslMat);
	gsl_vector_free(gslVect);
	gsl_matrix_free(gslResult);

}
void Transpose(DBL_VCTR_2D& res, const DBL_VCTR_2D& A)
{
	int i,j;
	int L =(int)A.size();
  	int C=(int)A[0].size();
  	if ((int)res.size()==0)
  	{
    		InitVctr2D(res,C,L,0.0);
  	}
  	else if ((int)res.size()!=C || (int)res[0].size()!=L)
  	{
    		printf("Transpose : wrong dimensions. Exiting...\n");
    		DimensionErrorMsg();
		exit(EXIT_FAILURE);
  	}
  	for (i=0; i<C; i++)
  	{
  		for (j=0; j<L; j++)
  		{
      			res[i][j]=A[j][i];
    		}
  	}
}

