#include "PredictUtilities.h"
#include "MatrixUtilities.h"
#include "RNGUtilities.h"

#include <gsl/gsl_randist.h>
#include <gsl/gsl_rng.h>
#include <cmath>
#include <fstream>
#include <iostream>
#include <gsl/gsl_cdf.h>
#include <math.h>
#include "ranker.h"

using namespace std;


// Warning if data contain supplementary covariates (not latitude or longitude) do not apply this function.
void predictFromBetaHat(DBL_VCTR_3D& prediction, const DBL_VCTR_2D& map, const DBL_VCTR_2D& beta, const double& noData)
{
	if (prediction.size() == 0)
	{
		cerr << "Uninitialized prediction -- predictFromBetaHat" << endl;
		exit(EXIT_FAILURE);
	}
	
	if (map.size() == 0)
	{
		cerr << "Uninitialized map -- predictFromBetaHat" << endl;
		exit(EXIT_FAILURE);
	}
	
	if (beta.size() == 0)
	{
		cerr << "Uninitialized beta -- predictFromBetaHat" << endl;
		exit(EXIT_FAILURE);
	}
	
	if (beta.size() != prediction.size())
	{
		cerr << "Internal error: beta.size and prediction.size don't agree -- predictFromBetaHat" << endl;
		exit(EXIT_FAILURE);
	}
	
	if (map.size() != (prediction[0].size()*prediction[0][0].size()))
	{
		cerr << "Internal error: map.size and prediction.size don't agree -- predictFromBetaHat" << endl;
		exit(EXIT_FAILURE);
	}
	
	int X = static_cast<int>(prediction[0].size());		//number of 1st coord point in the map
	int Y = static_cast<int>(prediction[0][0].size());	//number of 2nd coord point in the map
	
	int K = static_cast<int>(beta.size());		//Max number of clusters
	int B = static_cast<int>(beta[0].size()); 	//Number of predictors (depends on trend degree)
	
	//Allocation of Seed
	const gsl_rng_type *T;
	gsl_rng_env_setup();
	T = gsl_rng_default;
	gsl_rng* rng = NULL;
	AllocRNG(rng, T);
	
	DBL_VCTR_1D alpha,QPredict;
	InitVctr1D(alpha,K,0.0);	
	InitVctr1D(QPredict,K,0.0);
	
	
	int i,j,k;					//Loop index
	int cpt = 0;
	
	for (i = 0; i < X; i ++)
	{
		for (j= 0; j < Y; j ++)
		{
			double latitude = map[cpt][1];
			double longitude = map[cpt][0];
			cpt+=1;
			if (latitude != noData && longitude != noData)
			{
				
				for (k = 0; k < K; k ++)
				{
					//Mean
					double predictor_k = beta[k][0];
					//ASSUMING THE DATA SET COORDINATES WERE IN ORDER: Longitude - Latitude 
					if (B > 1)
					{
						//B == 3
						predictor_k += beta[k][1] * longitude + beta[k][2] * latitude;
						if (B > 3)
						{
							//B == 6
							predictor_k += beta[k][3] *longitude*longitude + beta[k][4] *latitude*latitude + beta[k][5]*latitude*longitude;
							if (B > 6)
							{
								//B == 10
								predictor_k += beta[k][6] *longitude*longitude*longitude + beta[k][7]*latitude *latitude*latitude + beta[k][8]*longitude*longitude*latitude + beta[k][9]*longitude*latitude*latitude;
							
							}
						}	
					}
				
					alpha[k] = exp(predictor_k);
				}
			
				//Q
				gsl_ran_dirichlet(rng, K, &alpha[0], &QPredict[0]);
			
				//prediction
				for (k = 0; k < K; k ++)
				{
					prediction[k][i][j] = QPredict[k];
				}				
			}
			else
			{
				for (k = 0; k < K; k ++)
				{
					prediction[k][i][j] = noData;
				}
			}
		}
	}
}


bool SavePredictionToMyFile(const DBL_VCTR_2D& prediction_clust, const string& File, const double& noData, const double& XLL, const double& YLL, const double& CellSize)
{
	if (prediction_clust.size() == 0)
	{
		cerr << "Uninitialized prediction -- SavePredictionToMyFile" << endl;
		exit(EXIT_FAILURE);
	}
	
	int X = static_cast<int>(prediction_clust.size());		//Number of longitude points
	int Y = static_cast<int>(prediction_clust[0].size());	//Number of latitude points
	
	int k,i,j;
	
	
	ofstream ofs(File.c_str());
	
	if (!ofs.good())
	{
		cerr << "Cannot open file " << File << " - SavePredictionToMyFile" << endl;
		exit(EXIT_FAILURE);
	}
	
	ofs << "#NCOLS   " << X << endl;
	ofs << "#NROWS   " << Y << endl;
	ofs << "#XLLCENTER   "<< XLL << endl;
	ofs << "#YLLCENTER   "<< YLL << endl;
	ofs << "#CELLSIZE   " << CellSize << endl;
	ofs << "#NODATA_VALUE   " << noData << endl;
	
	for (i = 0; i < X; i ++)
	{
		for (j = 0; j < Y; j ++)
		{
			ofs << prediction_clust[i][j] << "\t";
		}
		ofs << endl;
	}
	ofs << endl << endl << endl;
	
	ofs.close();
	return true;
}

bool LoadMap(DBL_VCTR_2D& map, double& NoData, double& XLL, double& YLL, double& CellSize, int& nrow, int& ncol, const string& File)
{
	//File MUST be in ASCII RASTER format
	ifstream ifs(File.c_str());
	
	if (!ifs.good())
	{
		cerr << "Cannot open file " << File << " - LoadMap" << endl;
		exit(EXIT_FAILURE);
	}
	
	int i,j;
		
	string discarded;
	ifs.precision(5);
	ifs >> discarded;
	ifs >> ncol;
	
	ifs >> discarded;
	ifs >> nrow;
	
	ifs >> discarded;
	ifs >> XLL;
	
	ifs >> discarded;
	ifs >> YLL;
	
	ifs >> discarded;
	ifs >> CellSize;
	
	ifs >> discarded;
	ifs >> NoData;
	
	int size = nrow*ncol;
	int cpt = 0;
	InitVctr2D(map,size,2,0.0);
	int Ymax = round(YLL + (nrow-1)*CellSize);
	
	for (i = 0; i < nrow; i ++)
	{
		for (j = 0; j < ncol; j ++)
		{
			double tmp = 0.0;
			ifs >> tmp;  
			if (tmp == NoData)
			{
				map[cpt][0] = NoData;
				map[cpt][1] = NoData;
			}
			else
			{
				map[cpt][0] = XLL + CellSize * j;	//longitude
				map[cpt][1] = Ymax - CellSize * i;	//latitude
			}
			cpt+=1;
		}
	}
}


