/*! \file popsAdm.cpp
*  \brief The POPS ADMIXTURE MCMC Algorithms
*  \author  Chibiao Chen and Eric Durand and Flora Jay
*/

#include <algorithm>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <sstream>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_exp.h>
#include <gsl/gsl_sf_log.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_histogram.h>
#include <gsl/gsl_math.h>
#include <iostream>
#include <set>
#include <string>

#include "IOUtilities.h"
#include "MiscUtilities.h"
#include "PlotUtilities.h"
#include "RNGUtilities.h"
#include "VectorUtilities.h"
#include "StatUtilities.h"
#include "MatrixUtilities.h"
#include "Voronoi.h"
#include "HierarchicalClustering.h"
#include "MatrixUtilities.h"
#include "PredictUtilities.h"

//! The Negative Limit of Logarithm (double)
#define N_LIMIT (-707.0)

//! The Positive Limit of Logarithm (double)
#define P_LIMIT (+707.0)

//! Sweep Parameters
class CSweepParams
{
public:
	DBL_VCTR_2D CAR;                          // CAR on admixture. Only used if CAR == true
	DBL_VCTR_2D beta;
	DBL_VCTR_1D sigma;                        // Variance of the K CARs
	DBL_VCTR_2D alpha;                        // Dirichlet parameters for admixture
	DBL_VCTR_2D q;                            // Admixture Proportions
	INT_VCTR_1D z;                            // Population of origin of individual i (for NJ init)
	INT_VCTR_3D zadMix;                       // Population of Origin of Allele Copy (case with admixture)
	DBL_VCTR_3D p;                            //!< Allele Frequencies
	DBL_VCTR_1D psi;
};

using namespace std;

//! Showing the Usage of the Program
/*! The compulsory/required options are indicated by upper case characters; the optional
*  options are indicated by lower case characters.
*  \author Chibiao Chen
*/
void ShowUsage()
{
	cout << "Typical Usage: popsAdm -FF -NN -AA -LL -KK -XX -XLXL -TT -PP -DD -SS -BB" << endl;
	cout << "Options can be specified in any order." << endl;
	cout << endl;
	cout << "Required Options:" << endl;
	cout << "-FF: F = File Name of Input Data File" << endl;
	cout << "     For file name contains spaces, input it as \"F\"." << endl;
	cout << "-NN: N = Number of Individuals" << endl;
	cout << "-AA: A = Ploidy (1 = Haploid, 2 = Diploid, ...)" << endl;
	cout << "-LL: L = Number of Loci" << endl;
	cout << "-KK: K = Number of Clusters" << endl;
	cout << "-XLXL: XL = Number of Qualitative Covariates" << endl;
	cout << "-XX: X = Number of Quantitative Covariates (other than spatial coordinates)" << endl;
	cout << "-TT: T = Degree of Trend" << endl;
	cout << "              0: Spatial Coord. NOT used for global trend regression (see doc for more details)" << endl;
	cout << "             {1,2,3}: Degree 1, 2 or 3" << endl;
	cout << "-PP: P = SPatial interaction parameter" << endl;
	cout << "-DD: D = Parameter of Dirichlet Allele Frequency Model (no F-model)" << endl;
	cout << "-SS: S = Total Number of Sweeps of MCMC" << endl;
	cout << "-BB: B = Burn In Number of Sweeps of MCMC" << endl;

	cout << endl;
	cout << "Optional Options:" << endl;
	cout << "-rr: r = Number of Extra Rows in Data File, Default: 0" << endl;
	cout << "-cc: c = Number of Extra Columns in Data File, Default: 0" << endl;
	cout << "-ii: i = Folder Name of Input Data File, Default: Current Folder" << endl;
	cout << "     For folder name contains spaces, input it as \"i\"." << endl;
	cout << "-oo: o = Folder Name of Output Result Files, Default: Current Folder" << endl;
	cout << "     For folder name contains spaces, input it as \"o\"." << endl;
	cout << "-orunorun: orun = String to Append to Output Result Files Names, e.g. a Run Number" << endl;
	cout << "-pp: p = Configuration File for Continue from Highest Likelihood Run, Default: NULL"<< endl;
	cout << "     For file name contains spaces, input it as \"p\"." << endl;
	cout << "-dd: d = Number of dummy individuals, Default: 0" << endl;
	cout << "-uu: u = Update sigma?" << endl;
	cout << "     u = y, Yes (Default)" << endl;
	cout << "     u = n, No" << endl;
	cout << "-upup: up = Update Psi?" << endl;
	cout << "     up = y, Yes" << endl;
	cout << "     up = n, No (Default)" << endl;
	cout << "-xx: x = Variance of CAR, Default: 30.0" << endl;
	cout << "-mm: m = Name of file containing map (ascii-raster format) to predict admixture coefficients on (optional)" << endl;
	cout << "     For file name contains spaces, input it as \"m\"." << endl;
	cout << "-nn: n = number of coordinates to use in regression. Default: 2." << endl;
	cout << "     If n = 1, the second coordinate is used" << endl;
	cout << "-ss: s = Shuffle update order for MCMC?" << endl;
	cout << "     s = y, Yes" << endl;
	cout << "     s = n, No (Default)" << endl;
	cout << "-jj: j = Init the run with Neighbor Joining tree" << endl;
	cout << "     j = y, Yes" << endl;
	cout << "     j = n, No (Default)" << endl;
	cout << "-gg: g = Name of file containing Geographic distances between individuals (optional) " << endl;
	cout << "     For file name contains spaces, input it as \"g\"." << endl;
	cout << "-rcrc: rc = Recessive Allele(s) in the data?" << endl;
	cout << "       rc = y, Yes" << endl;
	cout << "       rc = n, No (Default)" << endl;
	cout << "-scsc: sc = scale parameter for geographic distances. User predefined value or enter one. Ignored if no geographic distance file is given." << endl;
	cout << "       sc = m, Mean distance between individuals (default)" << endl;
	cout << "       sc = d, meDian distance between individuals" << endl;
	cout << "       sc = x, maX distance between individuals" << endl;
	cout << "       sc = value of the scale parameter (e.g. 10.4)" << endl;
	cout << "-spsp: sp = SPecial data format: one individual = one row" << endl;
	cout << "       sp = y, Yes" << endl;
	cout << "       sp = n, No (Default)" << endl;
	cout << "-bsbs: bs = Burn-In Structure, Default: 0" << endl;
}

//! Translating the XY Coordinates
/*! Translate the XY coordinates to all positive values to circumvent the quirks of
*  ChartDirector in the XYChart.
*  \param C Coordinates To Be Translated/Translated (Input & Output)
*  \return Result of the Action (Success/Failure)
*  \author Chibiao Chen
*/
bool TranslateCoordinates(DBL_VCTR_2D& C)
{
	int N = int(C.size());
	if (N == 0 || C[0].size() != 2)
	{
		return false;
	}
	double xMin = +1.0E+300;
	double xMax = -1.0E+300;
	double yMin = +1.0E+300;
	double yMax = -1.0E+300;
	int i;
	for (i = 0; i < N; i++)
	{
		if (C[i][0] < xMin)
		{
			xMin = C[i][0];
		}
		if (C[i][0] > xMax)
		{
			xMax = C[i][0];
		}
		if (C[i][1] < yMin)
		{
			yMin = C[i][1];
		}
		if (C[i][1] > yMax)
		{
			yMax = C[i][1];
		}
	}
	double xOffset = -xMin + (xMax - xMin) * 2.0;
	double yOffset = -yMin + (yMax - yMin) * 2.0;
	for (i = 0; i < N; i++)
	{
		C[i][0] += xOffset;
		C[i][1] += yOffset;
	}

	return true;
}

//! The POPS ADM MCMC Algorithms
int main(int argc, char* argv[])
{
	gsl_set_error_handler_off();                  // ERROR MUST BE HANDLED CASE BY CASE

	// Required Parameters
	string F;                                     // Name of Data File
	int N;                                        // Number of Samples/Individuals
	int A;                                        // Haploid, Diploid, ...
	int L;                                        // Number of Loci
	INT_VCTR_1D J;                                // Number of Distinct Alleles Observed at Every Locus
	int K;                                        // Number of Cluster
	int beta_length;
	double lambda;
	double psiUser = 0.6;
	int trendDegree = 1;                          //Degree of trend

	double sigmaQ = 10.0;                         //initial variances of CAR
	double sigmaQMean = 0.0;

	double minDIC = 1.7e+308;                     // min DIC of all runs; init with biggest double ever
	int MLtrendDegree;                            // trend degree of best run
	int TotalSweep;                               // Total Number of Sweeps
	int BurnIn;                                   // Burn In Number of Sweeps

	// Optional Parameters
	INT_VCTR_1D nQualLevels;        // Number of Categories for each Qualitative Variable	
	int nbQuantVar = 0;               // Number of Quantitative Covariates other than Spatial Coordinates
	string scOpt = "m";                           // How to initialize the scale parameter theta. Default: mean distance
	bool scSpecified = false;                     // Has user specified a value for the scale parameter. Default: no, use the mean distance
	double theta = 1.0;                           // scale parameter of geographic distances
	int ER = 0;                                   // Number of Extra Rows
	int EC = 0;                                   // Number of Extra Columns
	int nd = 0;                                   // Number of dummy individuals
	string FoI;                                   // Folder of Input Data File
	string FoO;                                   // Folder of Output Result Files
	string Voronoi;                               // Name of Voronoi File
	string run="";					 // Run Index for Output Files
	bool updateSigma = true;                      // Update variance of linear regression?
	bool updatePsi = false;
	bool shuffle = false;
	int burnStruct = 0;                           // Structure burn-in
	string geoFile;                               // Name of geographic distances file
	string mapFile;                               // Name of map (for prediction) file
	bool hasGeo;
	bool specialFormat = false;                   // Is the data in special format? (one line per individual, even diploid). Then diploid individual: one locus = 2 columns
	bool nj = false;                              // Init with Neighbor Joining tree ?
	string CF;
	bool isRecAllele = false;                     // Is there recessive alleles in the data set?
	bool bbase = false;                           //For compatibility only. Will be discarded anyway

	int nb_var_regr = 2;                          // How many variables (latitude,longitude) in linear regressions ?
	
	// Aux variables
	int pas_predict=10;                           // How many sweeps along MCMC should be used for prediction? (percentage)
	int nb_predict=10;                            // Each time we predict, how many values should we average?
	int nbQualVar=0;					// Number of Qualitative Covariates (= nQualLevels.size())
	//m int totalNbQualLevels=0;		// Number of Categories for ALL Qualitative Variable (tocalculate the size of Xtild)

	bool bF = false;
	bool bP = false;
	bool bN = false;
	bool bA = false;
	bool bL = false;
	bool bK = false;
	bool bT = false;
	bool bD = false;
	bool bS = false;
	bool bB = false;
	bool bX = false;
	bool bXL = false;

	if (argc < 10)
	{
	        cout << "Only " << argc-1 << " arguments given, this is not enough." << endl;
		ShowUsage();
		cout << "Only " << argc-1 << " arguments given, this is not enough." << endl;
		exit(EXIT_FAILURE);
	}
	else
	{
		// Parse the command line parameters.
		for (int ac = 1; ac < argc; ac++)
		{
			switch (argv[ac][1])
			{
			case 'X':
				if (argv[ac][2]=='L')
				{
				  //m nQualLevels.push_back(atoi((const char*) &argv[ac][3]));
				  //m totalNbQualLevels += atoi((const char*) &argv[ac][3]);
				  nbQualVar = atoi((const char*) &argv[ac][3]);
				  bXL = true;
				}
				else
				{
					nbQuantVar = atoi((const char*) &argv[ac][2]);
					bX = true;
				}
				break;

			case 'T':
				trendDegree = atoi((const char*) &argv[ac][2]);
				bT = true;
				break;

			case 'F':
				F = (const char*) &argv[ac][2];
				bF = true;
				break;

			case 'N':
				N = atoi((const char*) &argv[ac][2]);
				bN = true;
				break;

			case 'A':
				A = atoi((const char*) &argv[ac][2]);
				bA = true;
				break;

			case 'L':
				L = atoi((const char*) &argv[ac][2]);
				bL = true;
				break;

			case 'K':
				K = atoi((const char*) &argv[ac][2]);
				bK = true;
				break;

			case 'P':
				psiUser = atof((const char*) &argv[ac][2]);
				bP = true;
				break;

			case 'D':
				lambda = atof((const char*) &argv[ac][2]);
				bD = true;
				break;

			case 'S':
				TotalSweep = atoi((const char*) &argv[ac][2]);
				bS = true;
				break;

			case 'B':
				BurnIn = atoi((const char*) &argv[ac][2]);
				bB = true;
				break;

			case 'r':
				if (argv[ac][2] == 'e')
				// option -ref
				{
					if (argv[ac][3] == 'f')
					{
						bbase=true;    // For compatibility ; column for reference cluster (not used in admixture version)
					}
					else
					{
						cout << "Unknown Option - " << argv[ac] << " For more details: ./popsAdm " << endl;
						exit(EXIT_FAILURE);
					}
				}
				else if (argv[ac][2] == 'c')
				{
					if (argv[ac][3] == 'y')
					{
						isRecAllele = true;
					}
					if (argv[ac][3] == 'n')
					{
						isRecAllele = false;
					}
					if (argv[ac][3] != 'y' && argv[ac][3] != 'n')
					{
					        cout << "Unknown Option - " << argv[ac] << " For more details: ./popsAdm " << endl;
						exit(EXIT_FAILURE);
					}
				}
				else
				{
					ER = atoi((const char*) &argv[ac][2]);
				}
				break;

			case 'c':
				EC = atoi((const char*) &argv[ac][2]);
				break;

			case 'n':
				nb_var_regr = atoi((const char*) &argv[ac][2]);
				break;

			case 'p':
				CF = (const char*) &argv[ac][2];
				break;

			case 'j':
				if (argv[ac][2] != 'y' && argv[ac][2] != 'n')
				{
				        cout << "Unknown Option - " << argv[ac] << " For more details: ./popsAdm " << endl;
					exit(EXIT_FAILURE);
				}
				if (argv[ac][2] == 'y')
				{
					nj = true;
				}
				if (argv[ac][2] == 'n')
				{
					nj = false;
				}
				break;

			case 'd':
				nd = atoi((const char*) &argv[ac][2]);
				break;

			case 'x':
				sigmaQ = atof((const char*) &argv[ac][2]);
				break;

			case 'i':
				FoI = (const char*) &argv[ac][2];
				break;

			case 'o':
			  if (argv[ac][2] == 'r' && argv[ac][3] == 'u' && argv[ac][4] == 'n') {
			    run = (const char*) &argv[ac][5];
			  } else {
			    FoO = (const char*) &argv[ac][2];
			  }
			  break;


			case 'g':
				geoFile = (const char*) &argv[ac][2];
				break;

			case 'b':
				if (argv[ac][2] == 's')
				{
					burnStruct = atoi(&argv[ac][3]);
				}
				else
				{
				        cout << "Unknown Option - " << argv[ac] << " For more details: ./popsAdm " << endl;
					exit(EXIT_FAILURE);
				}
				break;

			case 'u':

				if (argv[ac][2] != 'y' && argv[ac][2] != 'n' && argv[ac][2] != 'p')
				{
					cout << "Unknown Option - " << argv[ac] << " For more details: ./popsAdm " << endl;
					exit(EXIT_FAILURE);
				}

				if (argv[ac][2] == 'p')
				{
					if (argv[ac][3] == 'y')
					{
						updatePsi = true;
					}
					else
					{
						updatePsi = false;
					}
				}
				else
				{
					if (argv[ac][2] == 'y')
					{
						updateSigma = true;
					}
					else
					{
						updateSigma = false;
					}
				}
				break;

			case 'v':
				Voronoi = (const char*) &argv[ac][2];
				break;

			case 'm':
				mapFile = (const char*) &argv[ac][2];
				break;

			case 's':

				if (argv[ac][2] == 'c')
				{
					if (argv[ac][3] == 'm' || argv[ac][3] == 'd' || argv[ac][3] == 'x')
					{
						scOpt = argv[ac][3];
					}
					else
					{
						theta= atof((const char*) &argv[ac][3]);
						scSpecified = true;
					}
					break;
				}

				if (argv[ac][2] == 'p')
				{
					//SPecial data format
					if (argv[ac][3] != 'y' && argv[ac][3] != 'n')
					{
					        cout << "Unknown Option - " << argv[ac] << " For more details: ./popsAdm " << endl;
						exit(EXIT_FAILURE);
					}
					if (argv[ac][3] == 'y')
					{
						specialFormat = true;
					}
					if (argv[ac][3] == 'n')
					{
						specialFormat = false;
					}
					break;
				}
				else
				{

					if (argv[ac][2] != 'y' && argv[ac][2] != 'n')
					{
						cout << "Unknown Option - " << argv[ac] << " For more details: ./popsAdm " << endl;
						exit(EXIT_FAILURE);
					}
					if (argv[ac][2] == 'y')
					{
						shuffle = true;
					}
					else
					{
						shuffle = false;
					}
					break;
				}

			default:                          // Unknown parameters.
			        cout << "Unknown Option - " << argv[ac] << " For more details: ./popsAdm " << endl;
				exit(EXIT_FAILURE);
			}
		}
	}

	if (!bF || !bN || !bA || !bL || !bK || !bD || !bS || !bB  || !bT || !bP )
	{

	        string missingopt="";
	        if (!bF) missingopt+=" -F" ;
		if (!bN) missingopt+=" -N" ;
		if (!bA) missingopt+=" -A" ;
		if (!bL) missingopt+=" -L" ;
		if (!bK) missingopt+=" -K" ;
		if (!bD) missingopt+=" -D" ;
		if (!bS) missingopt+=" -S" ;
		if (!bB) missingopt+=" -B" ;
		if (!bT) missingopt+=" -T" ;
		if (!bP) missingopt+=" -P" ;

		cout << "\n Mandatory arguments are missing, please check details for: " << missingopt << " with popsAdm | more " << endl;
		exit(EXIT_FAILURE);
	}

	if (isRecAllele && A != 2)
	{
	        cerr << "Dominant markers allowed for diploid individuals only ..." << endl;
		exit(EXIT_FAILURE);


	}

	if (psiUser >= 1.0)
	{
		cerr << "The spatial interaction parameter must be stricly inferior to 1." << endl;
		exit(EXIT_FAILURE);
	}

	if (trendDegree < 0 || trendDegree > 3)
	{
	        cerr << "The trend degree must be in {0,1,2,3}. It is mandatory that the input file contains spatial coordinates." << endl;
		exit(EXIT_FAILURE);
	}

	// Expected total number of lines given command line
	int expectedLines, expectedCols;
	int rec = isRecAllele? 1:0;	
	if (!specialFormat)
	{
	        // A-ploids individuals stored on A lines.
	        expectedLines = rec + ER + A*N + nd;
	}
	else
	{
		// individuals stored on one line.
		expectedLines = rec + ER + N + nd;
	}

	// Expected total number of columns given command line
	int coord =  (trendDegree ==-1)? 0:2; //Two Columns for XY Coordinates (for popsAdm trendDegree always !=1)
	int refCol = bbase? 1:0;
	if (!specialFormat)
	{
		// One column per locus
		expectedCols = EC + refCol + coord + nbQuantVar + nbQualVar + L;
	}
	else
	{
		// A column(s) per locus for A-ploids individuals
		expectedCols = EC + refCol + coord + nbQuantVar + nbQualVar + A*L;
	}
	printf("\n**Comment:**\n According to your parameter settings, the input datafile should have *%d* lines and *%d* columns.\nIf this is not correct please check your command line again.\n\n",expectedLines,expectedCols);

	string FIFN = !FoI.empty() ? FoI + "/" + F : F;
	string FOFN = !FoO.empty() ? FoO + "/" + F : F;
	if (!run.empty())
	  {
	    FOFN=FOFN+"_RUN"+run+"_";
	    cout<<"RUN= "<<run<<endl;
	  }

	int NumOfDigits = (int) log10((double) TotalSweep) + 1;

	// Loop Index Variables
	int i, a, l, j, k, b, predict, var;

	//m nbQualVar = nQualLevels.size();
	InitVctr1D(nQualLevels,nbQualVar,0);

	//cout << "nbQuantVar = " << nbQuantVar << "  nbQualVar = " << nbQualVar << endl;

	int trendSize = 1;
	//Degree of trend
	// The intercept is taken into account in trendSize
	switch(nb_var_regr)
	{
	case 0:
		trendSize = 1;
		break;
	case 1:
		trendSize = trendDegree + 1;
		break;
	case 2:
		trendSize = (trendDegree +1)*(trendDegree +2)/2;
		break;

	default:
		cerr << "Number of variables in the regression must be in {0,1,2}" << endl;
		ShowUsage();
		cout << "Number of variables in the regression must be in {0,1,2}" << endl;
		exit(EXIT_FAILURE);
	}

	//m beta_length = trendSize + nbQuantVar + totalNbQualLevels-nbQualVar;

	//geographic file available ?
	hasGeo = !geoFile.empty();

	DBL_VCTR_2D geo;                              //the geographic distances matrix
	if (hasGeo)
	{
		InitVctr2D(geo,N,N,0.0);
		LoadVector2D(geo,geoFile);
		if (!scSpecified)
		{
			computeSymmMatrixStat(theta,geo,scOpt);
		}
	}


	INT_VCTR_2D XQual; 	// Classes of Each Individual for Each Qual. Variable 
	                        // eg XQual[i][n]: class of individual i for the nth qual. variable
	InitVctr2D(XQual,N,nbQualVar,0);

	INT_VCTR_2D LevelPopSize; 	// Number of Individuals in Each Class for Each Qual. Variable
	//m InitVctr2D(LevelPopSize,nbQualVar,nQualLevels,0);
		
	INT_VCTR_3D X;
	InitVctr3D(X, N, A, L, 0);                    // L Loci

	INT_VCTR_1D recAllele;
	INT_VCTR_3D Xobs;                             //Observed genotypes
	if (isRecAllele)
	{
		InitVctr1D(recAllele,L,0);
		InitVctr3D(Xobs,N,A,L,0);
	}

	DBL_VCTR_2D Xtild;
	//m InitVctr2D(Xtild,N,beta_length,1.0);
	INT_VCTR_1D isBase;  //For compatibility with pops
	InitVctr1D(isBase,N,0);

	INT_VCTR_2D vvnNS;

	DBL_VCTR_2D C;
	InitVctr2D(C, N + nd, 2, 0.0);
	LoadData(bbase, isBase, C, Xtild, XQual, LevelPopSize, nQualLevels, X, recAllele, ER, EC, FIFN, specialFormat, nbQuantVar, nb_var_regr, isRecAllele, trendDegree);
	
	if (Xtild.size() == 0)
	  // if no covariates, (Xtild uninitialized in this case)
	  {
	    beta_length = 1;
	    InitVctr2D(Xtild,N,1,1.0); // no covariate but the intercept will be used (-> constant trend)
	  } 
	else 
	  {
	    beta_length=Xtild[0].size();
	  }


	// string saveTest="bla.txt";
	//    SaveData(Xtild,X,saveTest);
	
	if (isRecAllele)
	{
		for (i = 0; i < N; i ++)
		{
			for (a = 0; a < A; a ++)
			{
				for (l = 0; l < L; l ++)
				{
					Xobs[i][a][l] = X[i][a][l];
				}
			}
		}
	}

	clock_t s_clock = clock();

	//Allocation of Seed
	const gsl_rng_type *T;
	gsl_rng_env_setup();
	T = gsl_rng_default;
	gsl_rng* rng = NULL;
	AllocRNG(rng, T);

	TranslateCoordinates(C);
	vector< SPoint > vptC;
	vector< bool > vbKeep;
	vector< double > vdDataMinMax;
	vector< double > vdDataScale;
	vector< CEdge > veVoronoi;

	//init or load Voronoi diagram
	if (Voronoi == "" || !CVoronoi::LoadVoronoi(vptC, vbKeep, vdDataMinMax, vdDataScale, veVoronoi, vvnNS, Voronoi))
	{
		//adding small perturbation to coordinates to ensure one sample per location
		int nbDigitMax = 0;
		int nbDigit;
		string xCoord = "";
		string yCoord = "";

		for (i = 0; i < N; i ++)
		{
			stringstream tmp1,tmp2;
			tmp1 << C[i][0];
			xCoord = tmp1.str();
			tmp2 << C[i][1];
			yCoord = tmp2.str();
			nbDigit = max((xCoord.size()-xCoord.find('.')),(yCoord.size()-yCoord.find('.'))) - 1;

			if (nbDigit > nbDigitMax)
			{
				nbDigitMax = nbDigit;
			}
			tmp1.clear();
			tmp2.clear();
		}

		double perturb = gsl_pow_int(10,nbDigitMax);
		perturb = 1/(10.0*perturb);

		for (i = 0; i < N; i++)
		{
			C[i][0] += gsl_ran_gaussian(rng,perturb);
			C[i][1] += gsl_ran_gaussian(rng,perturb);
		}

		cout << "Generating Voronoi Tessellation & Neighborhood System... ";
		vptC.clear();
		SPoint ptThis;
		for (i = 0; i < N + nd; i++)
		{
			vptC.push_back(ptThis);
			vptC[i].m_nIndex = i;
			vptC[i].m_psXY.m_dX = C[i][0];
			vptC[i].m_psXY.m_dY = C[i][1];
		}
		CVoronoi::GetVoronoi(vptC, vbKeep, vdDataMinMax, vdDataScale, veVoronoi);
		string strVD = FIFN + "VD.png";
		CVoronoi::DrawVoronoi(vptC, N, vbKeep, vdDataMinMax, veVoronoi, false, strVD);
		CVoronoi::GetNeighborhoodSystem(N, veVoronoi, vvnNS);
		string strND = FIFN + "ND.png";
		CVoronoi::DrawNeighborhoodSystem(vptC, vdDataScale, vvnNS, true, strND);
		string strVC = FIFN + "VC";
		CVoronoi::SaveVoronoi(vptC, vbKeep,
		vdDataMinMax, vdDataScale, veVoronoi, vvnNS, strVC);
		cout << "Done" << endl;
	}

	// Determine the actual J from the data X.
	vector< set< int > > alleleS;
	set< int > emptyS;
	for (l = 0; l < L; l++)
	{
		alleleS.push_back(emptyS);
	}
	for (l = 0; l < L; l++)
	{
		for (i = 0; i < N; i++)
		{
			for (a = 0; a < A; a++)
			{
				alleleS[l].insert(X[i][a][l]);
			}
		}
	}

	INT_VCTR_2D alleleV;
	INT_VCTR_1D emptyV;

	for (l = 0; l < L; l++)
	{
		alleleV.push_back(emptyV);
	}
	for (l = 0; l < L; l++)
	{
		set< int >::const_iterator csiB = alleleS[l].begin();
		set< int >::const_iterator csiE = alleleS[l].end();
		// Discard the missing allele.
		// Missing allele is represented by a negative integer.
		if (*csiB < 0)
		{
			++csiB;
		}
		set< int >::const_iterator csi;
		for (csi = csiB; csi != csiE; ++csi)
		{
			alleleV[l].push_back(*csi);
		}
	}

	J.clear();
	for (l = 0; l < L; l++)
	{
		J.push_back(static_cast<int>(alleleV[l].size()));
	}

	//Dirichlet Parameters for allele frequencies
	DBL_VCTR_2D DirParams;
	InitVctr2D(DirParams, L, J, lambda);

	CSweepParams spSweep;

	DBL_VCTR_2D logAlpha;                         //log of alpha (to update beta)

	INT_VCTR_2D m;
	DBL_VCTR_2D logq;
	DBL_VCTR_1D dpq;                              //dirichlet parameters for admixture
	DBL_VCTR_4D cpAverageAdmix;                   //for admixture save of the maxL run
	DBL_VCTR_4D Pz;                               //for updating Zadmix (admixture case)

	//plot of the initial membership
	INT_VCTR_1D plotbidon;
	InitVctr1D(plotbidon,N,0);

	DBL_VCTR_2D cpBidon;
	InitVctr2D(cpBidon,N,K,0.0);

	//Init allele frequencies vector
	InitVctr3D(spSweep.p, K, L, J, 0.0);

	// Initialize auxiliary variable n.
	INT_VCTR_3D n;
	InitVctr3D(n, K, L, J, 0);

	// Initialize auxiliary variable logp.
	DBL_VCTR_3D logp;
	InitVctr3D(logp, K, L, J, 0.0);

	// Initialize clustering probabilities cpSweep and cpAverage.
	DBL_VCTR_2D cpAverage;
	InitVctr2D(cpAverage, N, K, 0.0);

	//per allele hard clustering
	INT_VCTR_3D hcAdmix;

	//init the vectors (admixture case)
	InitVctr3D(hcAdmix,N,A,L,0);

	// Initialize auxiliary variable m. (For Updating q)
	InitVctr2D(m, N, K, 0);

	//Initialize auxiliary variable logq (for updating q)
	InitVctr2D(logq,N,K,0.0);

	//Initialize auxiliary variable logAlpha (for updating alpha)
	InitVctr2D(logAlpha, N, K,0.0);

	//dpq
	InitVctr1D(dpq,K,1.0);

	//alpha
	InitVctr2D(spSweep.alpha, N, K,0.0);

	//q
	InitVctr2D(spSweep.q, N, K, 0.0);

	//Pz
	InitVctr4D(Pz,N,A,L,K,0.0);

	//Auxiliary: Pred = Xtild t(Beta)
	DBL_VCTR_2D Pred;
	InitVctr2D(Pred, N, K, 0.0);

	//Sigma
	InitVctr1D(spSweep.sigma,K,sigmaQ);

	//beta
	InitVctr2D(spSweep.beta, K, beta_length, 0.0);

	InitVctr1D(spSweep.z, N, 0);
	//zAdmix
	InitVctr3D(spSweep.zadMix, N, A, L, 0);

	//Init Z, P with NJ tree ? (individuals level)
	if (nj)
	{
		INT_VCTR_1D InitLabels = HClustering(X, WAR, K, true);
		for (i = 0; i < N; i++)
		{
			spSweep.z[i]=InitLabels[i];
		}

		INT_VCTR_1D OI(N, 0);
		for (i = 0; i < N; i++)
		{
			OI[i] = i;
		}
		int SN;
		if (L < 120)
		{
			SN = static_cast<int>(N * 0.25);
		}
		else
		{
			SN = static_cast<int>(N * 0.05);
		}
		if (SN == 0)
		{
			SN = 1;
		}
		INT_VCTR_1D SI(SN, 0);
		gsl_ran_choose(rng, &SI[0], SN, &OI[0], N, sizeof(int));
		INT_VCTR_1D KI(K, 0);
		for (k = 0; k < K; k++)
		{
			KI[k] = k;
		}
		for (i = 0; i < SN; i++)
		{
			int knew;
			do
			{
				gsl_ran_choose(rng, &knew, 1, &KI[0], K, sizeof(int));
			}
			while (knew == InitLabels[SI[i]]);
			InitLabels[SI[i]] = knew;
		}
		for (i = 0; i < N; i++)
		{

			spSweep.z[i]=InitLabels[i];
		}

		//now let's initialize p
		for (k = 0; k < K; k++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					int nSum = 0;
					int nInd = 0;
					for (i = 0; i < N; i++)
					{
						for (a = 0; a < A; a++)
						{
							if (spSweep.z[i] == k)
							{
								nInd++;
								if (X[i][a][l] == alleleV[l][j])
								{
									nSum++;
								}
							}
						}
					}
					if (nInd == 0)
					{
						spSweep.p[k][l][j] = 0.0;
					}
					else
					{
						spSweep.p[k][l][j] = (double)(nSum/(double)(nInd));
					}
				}
			}
		}
		string strIM = FOFN + "IM.png";
		PlotHC(strIM.c_str(), vptC, vbKeep, vdDataMinMax, veVoronoi, spSweep.z);
	}

	cout << "Inverting Covariance Matrices ..." << endl;
	//Matrix computation
	//CAR parameters
	DBL_VCTR_2D W;                                //W: W is the weight matrix (Wii = 0, Wij = 1 or exp(-dij/theta) if i~j and 0 otherwise)
	InitVctr2D(W,N,N,0.0);

	if (hasGeo)
	{
		buildNeighborhoodMatrix(W,vvnNS,geo,theta);
	}
	else
	{
		buildNeighborhoodMatrix(W,vvnNS);
	}
	//cout << "build neighborhood done" << endl;

	//Compute eigenvalues of R, because psi must be stricly lower than 1/(largest eigenvalue of R)
	DBL_VCTR_1D eigenVal;
	InitVctr1D(eigenVal,N,0.0);
	computeEigenvalues(eigenVal,W);
	//cout << "compute eigenval done" << endl;


	//eigenVal[N-1] is the largest eigenvalue of W
	//Assuming the user entered psiUser = psi*L, where L is the largest eigenvalue. We must have psiUser < 1;
	InitVctr1D(spSweep.psi,K,0.0);
	for (k = 0; k < K; k ++)
	{
		spSweep.psi[k] = double(psiUser/eigenVal[N-1]);
	}

	DBL_VCTR_3D R;                                //R = I - psi*W
	InitVctr3D(R,K,N,N,0.0);                      //R, the neighbourhood matrix for CAR
	for (k = 0; k < K; k ++)
	{
		for (i= 0; i < N; i++)
		{
			for (j = 0; j < N; j++)
			{
				if (i == j)
				{
					R[k][i][j] = 1.0;
				}
				else
				{
					R[k][i][j] = -spSweep.psi[k]*W[i][j];
				}
			}
		}
	}

	InitVctr2D(spSweep.CAR,K,N,0.0);              //spSweep.CAR[k][i] = log(alpha[i][k]); for computation facilities
	DBL_VCTR_3D Rinv,RinvChol;
	DBL_VCTR_3D CARcovarMatrix;
	InitVctr3D(RinvChol,K,N,N,0.0);
	InitVctr3D(CARcovarMatrix,K,N,N,0.0);
	InitVctr3D(Rinv,K,N,N,0.0);
	

	for (k = 0; k < K; k ++)
	{
		InverseWithChol(Rinv[k],R[k]);
		choldc(RinvChol[k],Rinv[k]);
	}

	//Beta parameters
	DBL_VCTR_3D Var_beta_chol;
	DBL_VCTR_3D Var_beta,V, M, Vinv, Vinv_M;      //V:matrice variance-covariance*1/sigma
	InitVctr3D(Vinv_M,K,beta_length,N,0.0);
	InitVctr3D(V,K,beta_length,beta_length,0.0);
	InitVctr3D(Vinv,K,beta_length,beta_length,0.0);
	InitVctr3D(Var_beta,K,beta_length,beta_length,0.0);
	InitVctr3D(Var_beta_chol,K,beta_length,beta_length,0.0);
	InitVctr3D(M,K,beta_length,N,0.0);

	DBL_VCTR_2D tXtild_Xtild, tXtild_W, tXtild_W_Xtild;
	DBL_VCTR_3D psi_tXtild_W,  psi_tXtild_W_Xtild;

	InitVctr3D(psi_tXtild_W, K, beta_length,N, 0.0);
	InitVctr3D(psi_tXtild_W_Xtild, K, beta_length, beta_length, 0.0);

	DBL_VCTR_1D Constant;                         //for initial mean of beta
	InitVctr1D(Constant,beta_length,0.0);

	DBL_VCTR_3D Mat_tmp;
	InitVctr3D(Mat_tmp,K,beta_length,K,0.0);

	DBL_VCTR_2D beta_mean;
	InitVctr2D(beta_mean, K, beta_length, 0.0);



	Mult_t_left(tXtild_Xtild,Xtild,Xtild);        //t(Xtild) %*% Xtild
	Mult_t_left(tXtild_W,Xtild,W);                //t(Xtild) %*% W
	Mult(tXtild_W_Xtild,tXtild_W, Xtild);         //t(XTild %*% W %*% Xtild
	
	// calcul of V, V_chol and V_tXtild
	for (k = 0; k < K; k ++)
	  {
	    Mult_byConstant(psi_tXtild_W_Xtild[k],tXtild_W_Xtild,spSweep.psi[k]);
	    Mult_byConstant(psi_tXtild_W[k],tXtild_W,spSweep.psi[k]);
	    //Vk = t(Xtild)%*%Xtild - psik * t(Xtild)%*%W%*%Xtild
	    Diff(V[k], tXtild_Xtild, psi_tXtild_W_Xtild[k]);
	    Diff_t_left(M[k], Xtild, psi_tXtild_W[k]);//Mk = t(Xtild) - psik * t(Xtild)%*%W
	    Inverse(Vinv[k],V[k]);
	    Mult(Vinv_M[k],Vinv[k],M[k]);             //utility matrix
	    choldc(Var_beta[k],Vinv[k]);
	    //V_chol = sqrt(sigmaQ)*Var_betak
	    Mult_byConstant(Var_beta_chol[k],Var_beta[k],sqrt(spSweep.sigma[k]));
	  }

	if (CF == "" || !LoadCF(spSweep.sigma, spSweep.beta, spSweep.alpha, spSweep.q,spSweep.zadMix,spSweep.p,CF))
	{

		for (k=0; k < K; k++)
		{
			MultivariateGauss(Constant,Var_beta_chol[k],spSweep.beta[k],rng);
		}

		// Initial auxiliary pred
		Mult_t(Pred,Xtild,spSweep.beta);          //Xtild * t(beta)

		for (k = 0; k < K; k ++)
		{
			Mult_byConstant(CARcovarMatrix[k],RinvChol[k],sqrt(spSweep.sigma[k]));
		}

		DBL_VCTR_1D ConstantCAR;                  //for initial mean of CAR
		InitVctr1D(ConstantCAR,N,0.0);

		for (k = 0; k < K; k++)
		{
			for (i = 0; i < N; i ++)
			{
				ConstantCAR[i] = Pred[i][k];
			}
			MultivariateGauss(ConstantCAR,CARcovarMatrix[k],spSweep.CAR[k],rng);
		}

		//Init alpha and q according to their prior
		for (i = 0; i < N; i ++)
		{
			for (k = 0; k < K; k++)
			{
				spSweep.alpha[i][k] = exp(spSweep.CAR[k][i]);
				// 				spSweep.alpha[i][k] = 1.0;
				// 				dpq[k] =spSweep.alpha[i][k];
			}
			if (!nj)
			{
				gsl_ran_dirichlet(rng, K, &dpq[0], &spSweep.q[i][0]);
			}
		}

		// 		string strInitQ = FOFN + "qInit.txt";
		// 		SaveVector2D(spSweep.q,strInitQ);

		//Init allele frequencies according to their prior and Init Z according to its prior
		if (!nj)
		{
			for (k = 0; k < K; k++)
			{
				for (l = 0; l < L; l++)
				{
					gsl_ran_dirichlet(rng, J[l], &DirParams[l][0], &spSweep.p[k][l][0]);
				}
			}

			for (i = 0; i < N; i++)
			{
				for (a = 0; a < A; a++)
				{
					for (l = 0; l < L; l++)
					{
						spSweep.zadMix[i][a][l] = SelectANum(rng, spSweep.q[i]);
					}
				}
			}
		}

		if (nj)
		{
			for (i = 0; i < N; i++)
			{
				for (k = 0; k < K; k++)
				{
					if (spSweep.z[i] == k)
					{
						spSweep.q[i][k] = 1.0;
						for (a= 0; a<A;a++)
						{
							for (l=0;l<L;l++)
							{
								spSweep.zadMix[i][a][l] = k;
							}
						}
					}
				}
			}
		}
	}

	else
	{
	  //Conf file has been loaded
		// Initial auxiliary pred
		Mult_t(Pred,Xtild,spSweep.beta);          //Xtild * t(beta)

		for (k = 0; k < K; k ++)
		{
			Mult_byConstant(CARcovarMatrix[k],RinvChol[k],sqrt(spSweep.sigma[k]));
		}
		DBL_VCTR_1D ConstantCAR;                  //for initial mean of CAR
		InitVctr1D(ConstantCAR,N,0.0);

		for (k = 0; k < K; k++)
		{
			for (i = 0; i < N; i ++)
			{
				ConstantCAR[i] = Pred[i][k];
			}
			MultivariateGauss(ConstantCAR,CARcovarMatrix[k],spSweep.CAR[k],rng);
		}

	}

	if (!nj)
	{
	  		//cpBidon & plotbidon to draw initial membership
		for (i = 0; i < N; i++)
		{
			for (a = 0; a < A; a++)
			{
				for (l=0;l<L;l++)
				{
					for (k = 0; k < K; k++)
					{
						if (spSweep.zadMix[i][a][l] == k)
						{
							cpBidon[i][k] +=1.0;
						}
					}
				}
			}
		}
	 
		for (i = 0; i < N; i++)
		{
			double dMaxBidon = -1.0;
			for (k = 0; k < K; k++)
			{
				if (cpBidon[i][k] > dMaxBidon)
				{
					dMaxBidon = cpBidon[i][k];
					plotbidon[i] = k;
				}
			}
		}

		string strIM = FOFN + "IM.png";
		PlotHC(strIM.c_str(), vptC, vbKeep, vdDataMinMax, veVoronoi, plotbidon);
	}

	cout << "Completed" << endl;

	DBL_VCTR_1D LH(TotalSweep, 0.0);              // Log-Likelihood History
	DBL_VCTR_3D BH;
	InitVctr3D(BH,K,beta_length,TotalSweep,0.0);  //beta history

	double varDeviance = 0.0;                     //Variance of Deviance

	DBL_VCTR_2D RSSH;
	InitVctr2D(RSSH,TotalSweep,K,0.0);

	//log likelihood for regressions
	DBL_VCTR_2D rLH;
	InitVctr2D(rLH,TotalSweep,K,0.0);
	
	//Cluster prediction with beta only
	DBL_VCTR_3D nClusterQualVar;
	InitVctr3D(nClusterQualVar,nbQualVar,nQualLevels,K,0.0);
	DBL_VCTR_2D nb_cluster_individual;
	InitVctr2D(nb_cluster_individual,N,K,0.0);
	DBL_VCTR_4D nb_cluster_loci;
	InitVctr4D(nb_cluster_loci,N,A,L,K,0.0);
	DBL_VCTR_2D QPredict;
	InitVctr2D(QPredict,N,K,0.0);
	DBL_VCTR_1D dpqPredict;
	InitVctr1D(dpqPredict,K,0.0);

	//FOR SHUFFLING INDIVIDUALS AND POPULATIONS
	int shuffledInd[N];
	int shuffledClust[K];
	int shuffledLoc[L];
	int shuffledPloidy[A];
	for (i = 0; i < N; i++)
	{
		shuffledInd[i] = i;
	}

	for (k = 0; k < K; k++)
	{
		shuffledClust[k] = k;
	}

	for (l = 0; l < L; l++)
	{
		shuffledLoc[l] = l;
	}

	for (a = 0; a < A; a++)
	{
		shuffledPloidy[a] = a;
	}

	//averages
	CSweepParams spAverage;
	InitVctr1D(spAverage.psi, K, 0.0);
	InitVctr3D(spAverage.p, K, L, J, 0.0);
	InitVctr1D(spAverage.sigma,K,0.0);
	InitVctr2D(spAverage.q,N,K,0.0);
	InitVctr4D(cpAverageAdmix,N,A,L,K,0.0);
	InitVctr2D(spAverage.beta,K,beta_length,0.0);
	InitVctr2D(spAverage.alpha,N,K,0.0);

	//variance
	DBL_VCTR_2D varBeta;
	InitVctr2D(varBeta,K,beta_length,0.0);
	
	DBL_VCTR_2D logAlphaAverage;
	InitVctr2D(logAlphaAverage,N,K,0.0);

	// Now, let's start our Markov chain Monte Carlo.
	int KeptSweep = TotalSweep - BurnIn;
	int spr;
	int status;
	double LCurr,vk;
	gsl_sf_result logpTmp, logqTmp;
	int clust,ind,ploidy,loc;

	//AD-MIXTURE CASE
	for (int s = 0; s < TotalSweep; s++)
	{
		//New sweep: new shuffle
		if (shuffle)
		{
			gsl_ran_shuffle(rng,shuffledInd,N,sizeof(int));
			gsl_ran_shuffle(rng,shuffledClust,K,sizeof(int));
			gsl_ran_shuffle(rng,shuffledLoc,L,sizeof(int));
			gsl_ran_shuffle(rng,shuffledPloidy,A,sizeof(int));
		}

		//Update X if there are recessive alleles in the data.
		if (isRecAllele)
		{
			int jr = 0;
			//All loci are dealt with independently
			for (l = 0; l < L; l ++)
			{

				int alleleRec = recAllele[l];
				if (alleleRec < 0)
				{
					//No recessive allele observed at locus l, skipping
					break;
				}

				//Finding the index of the recessive allele in the unique set of alleles
				for (jr = 0; jr < J[l]; jr ++)
				{
					if (alleleRec == alleleV[l][jr])
					{
						//jr is the recessive allele index.
						break;
					}
				}

				for (i = 0; i < N; i ++)
				{
					if (Xobs[i][0][l] != alleleRec)
					{
						//Observed genotype is ambiguous. Looking for dominant allele index in the data
						for (j = 0; j < J[l]; j++)
						{
							if (Xobs[i][0][l] == alleleV[l][j])
							{
								//j is the dominant allele index, coded as homozygote
								break;
							}
						}

						int kCurr0 = spSweep.zadMix[i][0][l];
						int kCurr1 = spSweep.zadMix[i][1][l];

						double den = spSweep.p[kCurr0][l][j]*spSweep.p[kCurr1][l][jr] + spSweep.p[kCurr1][l][j]*spSweep.p[kCurr0][l][jr] + spSweep.p[kCurr0][l][j]*spSweep.p[kCurr1][l][j];

						double p_hetero0 = spSweep.p[kCurr0][l][j]*spSweep.p[kCurr1][l][jr]/den;
						double p_hetero1 = spSweep.p[kCurr1][l][j]*spSweep.p[kCurr0][l][jr]/den;

						double u = gsl_rng_uniform(rng);
						if (u < p_hetero0)
						{
							X[i][0][l] = alleleV[l][j];
							X[i][1][l] = alleleV[l][jr];
						}
						else if (u < p_hetero0 + p_hetero1)
						{
							X[i][1][l] = alleleV[l][j];
							X[i][0][l] = alleleV[l][jr];
						}
						else
						{
							X[i][0][l] = alleleV[l][j];
							X[i][1][l] = alleleV[l][j];
						}

					}
				}
			}
		}

		// Update p.
		for (k = 0; k < K; k++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					int nSum = 0;
					for (i = 0; i < N; i++)
					{
						for (a = 0; a < A; a++)
						{
							if (X[i][a][l] == alleleV[l][j] && spSweep.zadMix[i][a][l] == k)
							{
								nSum++;
							}
						}
					}
					n[k][l][j] = nSum;
				}
			}
		}

		for (k = 0; k < K; k++)
		{
			clust = shuffledClust[k];
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					DirParams[l][j] = lambda + n[clust][l][j];
				}
				gsl_ran_dirichlet(rng, J[l], &DirParams[l][0], &spSweep.p[clust][l][0]);
			}
		}

		// Update logp.
		for (k = 0; k < K; k++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					status = gsl_sf_log_e(spSweep.p[k][l][j], &logpTmp);
					if (status)
					{
						logp[k][l][j] = N_LIMIT;
						cout << "warning: N_LIMIT used for logp" << endl;
					}
					else
					{
						logp[k][l][j] = logpTmp.val;
					}
				}
			}
		}

		// UPDATE Q for admixture model: Gibbs Sampler
		for (i = 0; i < N; i++)
		{
			for (k = 0; k < K; k++)
			{
				int nSum = 0;
				for (a = 0; a < A; a++)
				{
					for (l = 0; l < L; l++)
					{
						if (spSweep.zadMix[i][a][l] == k)
						{
							nSum++;
						}
					}
				}
				m[i][k] = nSum;
			}
		}

		for (i = 0; i < N; i++)
		{
			ind = shuffledInd[i];
			for (k = 0; k < K; k++)
			{
				dpq[k] = spSweep.alpha[ind][k] + m[ind][k];
			}
			gsl_ran_dirichlet(rng, K, &dpq[0], &spSweep.q[ind][0]);
		}

		//update logQ
		for (i = 0; i < N; i++)
		{
			for (k = 0; k < K; k++)
			{
				status = gsl_sf_log_e(spSweep.q[i][k], &logqTmp);
				if (status)
				{
					logq[i][k] = N_LIMIT;
				}
				else
				{
					logq[i][k] = logqTmp.val;
				}
			}
		}

		if (s > burnStruct)
		{
			//Update logAlpha: Metropolis-Hastings
			for (i = 0; i < N; i ++)
			{
				double eSum = 0.0;
				ind = shuffledInd[i];
				for (k = 0; k < K; k++)
				{
					eSum += exp(spSweep.CAR[k][ind]);
				}

				double size = static_cast<int>(vvnNS[ind].size()); // neighbors of individual ind 

				for (k = 0; k < K; k ++)
				{
					clust = shuffledClust[k];
					double lnAPu = 0.0;
					double mean = Pred[ind][clust];  // Pred = Xtild %*% t(beta) (mean response ie global trend)

					for (j = 0; j < size; j++)
					{
						if (!hasGeo)
						{
							mean += spSweep.psi[clust]*(spSweep.CAR[clust][vvnNS[ind][j]] - Pred[vvnNS[ind][j]][clust]);
							// spSeep.CAR - Pred = y  where y is the pure CAR term ?
							// for all j neighbors of i, wij = 1 ?
						}
						else
						{
						  mean += spSweep.psi[clust]*exp(-geo[ind][vvnNS[ind][j]]/theta)*(spSweep.CAR[clust][vvnNS[ind][j]] - Pred[vvnNS[ind][j]][clust]);
						  // wij = exp(-dij/theta)
						}
					}

					//Proposal = conditional prior
					double CAR_n = mean + gsl_ran_gaussian(rng,sqrt(spSweep.sigma[clust]));
					double eSum_n = eSum - exp(spSweep.CAR[clust][ind]) + exp(CAR_n);

					lnAPu = gsl_sf_lngamma(exp(spSweep.CAR[clust][ind])) - gsl_sf_lngamma(exp(CAR_n)) + (exp(CAR_n) - exp(spSweep.CAR[clust][ind]))*logq[ind][clust] + gsl_sf_lngamma(eSum_n) - gsl_sf_lngamma(eSum);

					double u = gsl_rng_uniform(rng);
					if (u < exp(lnAPu))
					{
						spSweep.CAR[clust][ind] = CAR_n;
					}
				}
			}

			//Update alpha for computation facilities
			for (i = 0; i < N; i ++)
			{
				for (k = 0; k < K; k ++)
				{
					spSweep.alpha[i][k] = exp(spSweep.CAR[k][i]);
					logAlpha[i][k] = spSweep.CAR[k][i];
				}
			}

			// Update beta
			// in R software :beta_mean=V%*%t(Xtild)%*%logAlpha;
			// V_tXtild=v%*%t(Xtild) has already been calculated
			for (k = 0; k < K; k ++)
			{
				//Mat_tmp = Vinv*M*logAlpha
				Mult(Mat_tmp[k],Vinv_M[k],logAlpha);
			}

			for (k=0; k < K; k++)
			{
				Transpose(beta_mean,Mat_tmp[k]);  //beta_mean = t(Mat_tmp)

				if (spSweep.sigma[k] > 0)
				{
					MultivariateGauss(beta_mean[k],Var_beta_chol[k],spSweep.beta[k],rng);
				}
				else
				{
					for (b = 0; b < beta_length; b++)
					{
						spSweep.beta[k][b] = beta_mean[k][b];
					}
				}
			}

			//Update Xtild*t(beta) = Pred
			Mult_t(Pred,Xtild,spSweep.beta);      //Xtild * t(beta)

			//Beta history for graphical representation
			for (k = 0; k < K; k++)
			{
				for (b = 0; b < beta_length; b++)
				{
					BH[k][b][s] = spSweep.beta[k][b];
				}
			}

			//Update sigma
			if (updateSigma)
			{
				for (k = 0; k < K; k++)
				{
					clust = shuffledClust[k];
					vk = 0.0;
					for (i = 0; i < N; i++)
					{
						double size = static_cast<int>(vvnNS[i].size());
						double vj = 0.0;
						for (j = 0; j < size; j++)
						{
							if (hasGeo)
							{
								vj += exp(-geo[i][vvnNS[i][j]]/theta)*(spSweep.CAR[clust][vvnNS[i][j]] - Pred[vvnNS[i][j]][clust]);
							}
							else
							{
								vj += (spSweep.CAR[clust][vvnNS[i][j]] - Pred[vvnNS[i][j]][clust]);
							}
						}
						vk += gsl_pow_2(spSweep.CAR[clust][i] - Pred[i][clust]) - spSweep.psi[clust]*(spSweep.CAR[clust][i] - Pred[i][clust])*vj;
					}

					double par1 = (double)(N)/((double)(2));
					double par2 = 1.0/((double)(vk)/((double)(2)));
					double gammaVar = gsl_ran_gamma(rng,par1,par2);
					spSweep.sigma[clust] = 1.0/gammaVar;

					// Update Covariance Matrix
					//Var_beta_chol = sqrt(sigmaQ)*Var = sqrt(sigmaQ)*cholesky(t(Xtild)Xtild)
					Mult_byConstant(Var_beta_chol[clust],Var_beta[clust],sqrt(spSweep.sigma[clust]));
				}
			}

			if (updatePsi)
			{
				for (k = 0; k < K; k ++)
				{
					//Update psi : Metropolis-Hastings
					double lnAPpsi = 0.0;
					//Conditionnal random walk proposal
					double psi_n  = spSweep.psi[k] + gsl_ran_gaussian(rng,0.05);
					if (psi_n >= 0 && psi_n < 1/eigenVal[N-1])
					{
						for (i = 0; i < N; i ++)
						{
							double sumJ = 0.0;
							double size = static_cast<int>(vvnNS[i].size());
							for (j = 0; j < size; j++)
							{
								if (hasGeo)
								{
									sumJ +=  exp(-geo[i][vvnNS[i][j]]/theta)*(spSweep.CAR[k][vvnNS[i][j]] - Pred[vvnNS[i][j]][k]);
								}
								else
								{
									sumJ += (spSweep.CAR[k][vvnNS[i][j]] - Pred[vvnNS[i][j]][k]);
								}
							}

							lnAPpsi += 0.5*(gsl_sf_log(1.0 - psi_n*eigenVal[i]) - gsl_sf_log(1.0 - spSweep.psi[k]*eigenVal[i])) + (psi_n - spSweep.psi[k])*(spSweep.CAR[k][i] - Pred[i][k])*sumJ*(-0.5/spSweep.sigma[k]);
						}

						double u = gsl_rng_uniform(rng);
						if (u < exp(lnAPpsi))
						{
							spSweep.psi[k] = psi_n;

							//Update utility matrix! HEAVY !
							Mult_byConstant(psi_tXtild_W_Xtild[k],tXtild_W_Xtild,spSweep.psi[k]);
							Mult_byConstant(psi_tXtild_W[k],tXtild_W,spSweep.psi[k]);
							Diff(V[k], tXtild_Xtild, psi_tXtild_W_Xtild[k]);
							Diff_t_left(M[k], Xtild, psi_tXtild_W[k]);
							Inverse(Vinv[k],V[k]);
							choldc(Var_beta[k],Vinv[k]);
							Mult(Vinv_M[k],Vinv[k],M[k]);
							//V_chol = sqrt(sigmaQ)*Var_betak
							Mult_byConstant(Var_beta_chol[k],Var_beta[k],sqrt(spSweep.sigma[k]));
						}
					}
				}
			}
		}

		//Update z
		for (i = 0; i < N; i++)
		{
			for (a = 0; a < A; a++)
			{
				for (l = 0; l < L; l++)
				{
					if (X[i][a][l] < 0)
					{
						for (k = 0; k < K; k++)
						{
							Pz[i][a][l][k] = spSweep.q[i][k];
						}
					}
					else
					{
						for (j = 0; j < J[l]; j++)
						{
							if (X[i][a][l] == alleleV[l][j])
							{
								break;
							}
						}
						for (k = 0; k < K; k++)
						{
							Pz[i][a][l][k] = spSweep.q[i][k] * spSweep.p[k][l][j];
						}
					}
				}
			}
		}

		for (i = 0; i < N; i++)
		{
			for (a = 0; a < A; a++)
			{
				for (l = 0; l < L; l++)
				{
					double dSum = 0.0;
					for (k = 0; k < K; k++)
					{
						dSum += Pz[i][a][l][k];
					}
					for (k = 0; k < K; k++)
					{
						Pz[i][a][l][k] /= dSum;
					}
				}
			}
		}

		for (i = 0; i < N; i++)
		{
			ind = shuffledInd[i];
			for (a = 0; a < A; a++)
			{
				ploidy = shuffledPloidy[a];
				for (l = 0; l < L; l++)
				{
					loc = shuffledLoc[l];
					spSweep.zadMix[ind][ploidy][loc] = SelectANum(rng, Pz[ind][ploidy][loc]);
				}
			}
		}

		// Update Log-Likelihood.
		LCurr = 0.0;
		for (i = 0; i < N; i++)
		{
			for (a = 0; a < A; a++)
			{
				for (l = 0; l < L; l++)
				{
					for (j = 0; j < J[l]; j++)
					{
						if (X[i][a][l] == alleleV[l][j])
						{
							LCurr += logp[spSweep.zadMix[i][a][l]][l][j];
							break;
						}
					}
				}
			}
		}
		LH[s] = LCurr;

		//Update rDIC
		for (k = 0; k < K; k++)
		{
			double RSScurr = 0.0;
			for (i = 0; i < N; i ++)
			{
				RSScurr += gsl_pow_2(spSweep.CAR[k][i] - Pred[i][k]);
			}

			rLH[s][k] = -(double)(N/((double)(2)))*gsl_sf_log(spSweep.sigma[k]) - 1/(2*spSweep.sigma[k])*RSScurr;
		}

		// Accumulate results after burn-in.
		if (s > BurnIn - 1)
		{
			// p
			for (k = 0; k < K; k++)
			{
				for (l = 0; l < L; l++)
				{
					for (j = 0; j < J[l]; j++)
					{
						spAverage.p[k][l][j] += spSweep.p[k][l][j];
					}
				}
			}

			//Sigma
			for (k = 0; k < K; k++)
			{
				spAverage.sigma[k] += spSweep.sigma[k];
				sigmaQMean += spSweep.sigma[k];
			}

			//psi
			for (k = 0; k < K; k ++)
			{
				spAverage.psi[k] += spSweep.psi[k];
			}

			//q
			for (i=0;i<N;i++)
			{
				for (k=0;k<K;k++)
				{
					spAverage.q[i][k] += spSweep.q[i][k];
				}
			}

			//Alpha
			for (i=0;i<N;i++)
			{
				for (k=0;k<K;k++)
				{
					spAverage.alpha[i][k] += spSweep.alpha[i][k];
				}
			}

			//log Alpha for computation of regression DIC
			for (i=0;i<N;i++)
			{
				for (k=0;k<K;k++)
				{
					logAlphaAverage[i][k] += spSweep.CAR[k][i];
				}
			}

			// cp
			for (i = 0; i < N; i++)
			{
				for (l=0;l<L;l++)
				{
					for (a=0;a<A;a++)
					{
						for (k = 0; k < K; k++)
						{
							if (spSweep.zadMix[i][a][l] == k)
							{
								cpAverage[i][k] += 1.0;
								cpAverageAdmix[i][a][l][k] += 1.0;
								break;
							}
						}
					}
				}
			}

			// beta
			for (k=0; k < K; k++)
			{
				for (b = 0; b < beta_length ; b++)
				{
					spAverage.beta[k][b] += spSweep.beta[k][b];
					varBeta[k][b] += spSweep.beta[k][b] * spSweep.beta[k][b];

				}
			}

			//DIC variance
			varDeviance += 4*(LH[s])*(LH[s]);     //(-2log(L))^2

			// Clustering probabilities without genetic information
			if (s % pas_predict == 0 && beta_length != 0)
			{
				for (predict = 0; predict < nb_predict; predict++)
				{
					for (i = 0; i < N; i++)
					{
						for (k = 0; k < K; k++)
						{
							dpqPredict[k] = exp(Pred[i][k] + gsl_ran_gaussian(rng,sqrt(spSweep.sigma[k])));
						}
						gsl_ran_dirichlet(rng, K, &dpqPredict[0], &QPredict[i][0]);

						//Now QPredict[i,] is fully computed. We predict every loci' population based on Qik
						for (a = 0; a < A; a ++)
						{
							for (l = 0; l < L; l ++)
							{
								int zial = SelectANum(rng, QPredict[i]);
								nb_cluster_loci[i][a][l][zial]++;
								nb_cluster_individual[i][zial]++;
								if (nbQualVar>0)
								{
								  for (var=0; var < nbQualVar; var++)
								    {
								      nClusterQualVar[var][XQual[i][var]][zial]++;
								    }								
								}
							}
						}
					}
				}
			}

		}

		spr = s + 1;

		if (spr % 100 == 0)
		{
			cout.width(NumOfDigits);
			cout.fill('0');
			cout << spr << " sweeps completed." << endl;
		}

	}

	string strLH = FOFN + "LH.png";
	PlotLH(strLH.c_str(), LH);

	string strBH = FOFN + "BH.png";
	PlotBetaHist(strBH.c_str(),BH, nQualLevels, nbQuantVar, trendDegree);

	string strBHtext = FOFN + "BP.txt";
	SaveBetaToMyFile(BH,strBHtext,BurnIn);

	// Calculate the Monte Carlo averages after MCMC.

	// Predictions averaged among Class of the Qualitative Variables (eg Linguistic Groups for the variable 'Languages')
	if (nbQualVar>0)
	  {
	    for (var=0; var < nbQualVar; var++)
	      {
		for (k=0; k<K; k++)
		  {
		    for (int level=0;level<nQualLevels[var];level++)
		      {
			nClusterQualVar[var][level][k]/=(double)A*L*(double)nb_predict*(double)LevelPopSize[var][level]*((double)KeptSweep/(double)pas_predict);
		      }
		  }
	      }
	  }

		

	//nb_cluster_loci
	for (i=0; i < N; i++)
	{
		for (k=0; k < K; k++)
		{
			nb_cluster_individual[i][k]/=(double)A*L*nb_predict*((double)KeptSweep/(double)pas_predict);
			for (a = 0; a < A; a++)
			{
				for (l = 0; l < L; l++)
				{
					nb_cluster_loci[i][a][l][k] /= (double)nb_predict*((double)KeptSweep/(double)pas_predict);
				}
			}
		}
	}
	// p
	for (k = 0; k < K; k++)
	{
		for (l = 0; l < L; l++)
		{
			for (j = 0; j < J[l]; j++)
			{
				spAverage.p[k][l][j] /= KeptSweep;
			}
		}
	}

	//cpAdmix
	for (i = 0; i < N; i++)
	{
		for (a = 0; a < A; a++)
		{
			for (l = 0; l < L; l++)
			{
				for (k = 0; k < K; k++)
				{
					cpAverageAdmix[i][a][l][k] /= (double)KeptSweep;
				}
			}
		}
	}

	for (i = 0; i < N; i++)
	{
		for (a = 0; a < A; a++)
		{
			for (l = 0; l < L; l++)
			{
				double dMax = -1.0;
				for (k = 0; k < K; k++)
				{
					if (cpAverageAdmix[i][a][l][k] > dMax)
					{
						dMax = cpAverageAdmix[i][a][l][k];
						hcAdmix[i][a][l] = k;
					}
				}
			}
		}
	}

	// q
	for (i=0; i<N; i++)
	{
		for (k=0; k<K; k++)
		{
			spAverage.q[i][k] /= KeptSweep;
		}
	}

	// sigma
	for (k = 0; k < K; k ++)
	{
		spAverage.sigma[k] /= KeptSweep;
	}

	sigmaQMean /= (KeptSweep * K);

	// psi
	for (k = 0; k < K; k ++)
	{
		spAverage.psi[k] /= KeptSweep;
	}

	// alpha
	for (i = 0; i < N; i++)
	{
		for (k = 0; k < K; k++)
		{
			spAverage.alpha[i][k] /= KeptSweep;
		}
	}

	// logAlpha
	for (i = 0; i < N; i++)
	{
		for (k = 0; k < K; k++)
		{
			logAlphaAverage[i][k] /= KeptSweep;
		}
	}

	// cp
	for (i = 0; i < N; i++)
	{
		for (k = 0; k < K; k++)
		{
			cpAverage[i][k] /= KeptSweep * A * L;
		}

	}

	//beta
	for (k = 0; k < K; k++)
	{
		for (b = 0; b < beta_length; b++)
		{
			spAverage.beta[k][b] /= KeptSweep;
			varBeta[k][b] /= KeptSweep;
		}
	}

	string BHat = FOFN + "betaHat.txt";
	SaveVector2D(spAverage.beta,BHat);

	// 	string Xest = FOFN + "Xest.txt";
	// 	SaveData(Xtild,X,Xest);

	// Hard Clustering
	INT_VCTR_1D hc(N, 0);
	for (i = 0; i < N; i++)
	{
		double dMax = -1.0;
		for (k = 0; k < K; k++)
		{
			if (cpAverage[i][k] > dMax)
			{
				dMax = cpAverage[i][k];
				hc[i] = k;
			}
		}
	}

	string strHC = FOFN + "HC.png";
	PlotHC(strHC.c_str(), vptC, vbKeep, vdDataMinMax, veVoronoi, hc);

	// pi
	DBL_VCTR_1D pi(K, 0.0);
	for (k = 0; k < K; k++)
	{
		for (i = 0; i < N; i++)
		{
			pi[k] += cpAverage[i][k];
		}
	}
	for (k = 0; k < K; k++)
	{
		pi[k] /= N;
	}

	// Log the results to a text file.
	string strTR = FOFN + "TR.txt";
	ofstream ofsTR(strTR.c_str());

	// Estimated Clustering Probabilities for Samples
	ofsTR << "Estimated Clustering Probabilities for Samples" << endl;
	ofsTR << "Sample\t";
	for (k = 0; k < K; k++)
	{
		ofsTR << "***Cluster  ";
		ofsTR.width(2);
		ofsTR.fill('0');
		// Make sure the first index is 1 not 0.
		ofsTR << k + 1 << "***\t";
	}
	ofsTR << endl;
	for (i = 0; i < N; i++)
	{
		ofsTR.width(6);
		ofsTR.fill('0');
		// Make sure the first index is 1 not 0.
		ofsTR << i + 1<<":"<< "\t";
		for (k = 0; k < K; k++)
		{
			ofsTR.width(17);
			ofsTR.precision(15);
			ofsTR << fixed;
			ofsTR << cpAverage[i][k] << "\t";
		}
		ofsTR << "\t" << "1";
		ofsTR << endl;
	}
	ofsTR << endl << endl << endl;

	// Estimated Allele Frequencies
	ofsTR << "Estimated Allele Frequencies" << endl;
	for (k = 0; k < K; k++)
	{
		ofsTR << "Cluster ";
		ofsTR.width(2);
		ofsTR.fill('0');
		// Make sure the first index is 1 not 0.
		ofsTR << k + 1 << endl;
		for (l = 0; l < L; l++)
		{
			ofsTR << "Locus ";
			ofsTR.width(3);
			ofsTR.fill('0');
			// Make sure the first index is 1 not 0.
			ofsTR << l + 1 << endl;
			for (j = 0; j < J[l]; j++)
			{
				ofsTR << "Allele ";
				ofsTR.width(3);
				ofsTR.fill('0');
				ofsTR << alleleV[l][j] << "\t";
				ofsTR.width(17);
				ofsTR.precision(15);
				ofsTR << fixed;
				ofsTR << spAverage.p[k][l][j] << endl;
			}
			ofsTR << endl;
		}
		ofsTR << endl;
	}
	ofsTR << endl;

	// Estimated Mixing Proportions
	ofsTR << "Estimated Mixing Proportions" << endl;
	for (k = 0; k < K; k++)
	{
		ofsTR.width(17);
		ofsTR.precision(15);
		ofsTR << fixed;
		ofsTR << pi[k] << endl;
	}
	ofsTR << endl << endl << endl;
	

	// Estimated Beta, Significance, and Confidence Interval
	ofsTR << "Estimated Beta and Significance Level:" << endl;
	for (k = 0; k < K; k++)
	  {
	    ofsTR << k+1 << ":" << "\t";
	    for (b = 0; b < beta_length; b++)
	      {
		ofsTR.width(17);
		ofsTR.precision(15);
		ofsTR << fixed;
		ofsTR << spAverage.beta[k][b] << " "<< Significance(&BH[k][b][BurnIn],KeptSweep)<< "\t";
	      }
	    ofsTR << endl;
	  }
	ofsTR << endl <<"100\% '****' 99.9 \% '***' 99\% '**' 95\% '*' 90\% '-' 0\% " <<endl;
	ofsTR << endl << endl << endl;
	
	ofsTR << "95 % Bayesian Confidence Interval for Estimated Beta:" << endl;
	for (k = 0; k < K; k++)
	  {
	    ofsTR << k+1 << ":" << "\t";
	    for (b = 0; b < beta_length; b++)
	      {
		ofsTR.width(17);
		ofsTR.precision(15);
		ofsTR << fixed;
		ofsTR << CI(&BH[k][b][BurnIn],KeptSweep,0.025,0.975) << "\t";
	      }
	    ofsTR << endl;
	  }
	ofsTR << endl << endl << endl;	
	
	// Estimated variance
	ofsTR << "Estimated sigma: " << endl;
	for (k = 0; k < K; k++)
	  {
	    ofsTR << "k: "<< k << '\t' << spAverage.sigma[k] << endl;
	  }

	ofsTR << endl << endl << endl;

	ofsTR << "Estimated psi: " << endl;
	for (k = 0; k < K; k ++)
	  {
	    ofsTR << "k: "<< k << '\t' << spAverage.psi[k]*eigenVal[N-1] << endl;
	  }

	ofsTR << endl << endl << endl;

	ofsTR.close();

	string strAP = FOFN + "AP.png";
	PlotAP(strAP.c_str(), cpAverage, "Assignment Probabilities");
	
	if (beta_length != 0)
	  // There are Covariates -> Predicted coefficients
	  {
	    //PPM: Posterior Predictive Memberships
	    string ClInd_barplot=FOFN+"PPM.png";
	    string ClInd_str = FOFN+"PPM.txt";
	    SaveToMyFileClumppFormat(nb_cluster_individual,ClInd_str,true,"Predicted Clustering Probabilities for Samples");
	    PlotAP(ClInd_barplot.c_str(),nb_cluster_individual,"Predicted Assignment Probabilities");
	    double cor =  Correlation(cpAverage,nb_cluster_individual);
	    cout << "Correlation between estimated and predicted membership coeff = " << cor << endl;
	    string strCOR = FOFN + "COR.txt";
	    ofstream ofsCOR(strCOR.c_str());
	    ofsCOR << cor;
	    ofsCOR.close();
	   
	    if (nbQualVar > 0)
	      {
		for (var=0; var < nbQualVar; var++)
		  {
		    
		    string svar;
		    stringstream out;
		    out << var+1;
		    svar = out.str();
		    
		    // PPM = Posterior Predictive Membership
		    string clFile=FOFN+"PPMvar"+svar+".txt";
		    SaveToMyFileClumppFormat(nClusterQualVar[var],clFile);
		  }
		string ClQualVar_barplot=FOFN+"PPMqual.png";
		PlotPPMqual(ClQualVar_barplot.c_str(),nClusterQualVar);	     
	      }
	  }
	
	// Calculate Average Log-Likelihood
	int NoS = int(0.1 * KeptSweep);               // Number of Sweeps Used in Calculation
	if (NoS == 0)
	{
		NoS = 1;
	}
	double dAL = 0.0;
	for (i = TotalSweep - NoS; i < TotalSweep; i++)
	{
		dAL += LH[i];
	}

	dAL /= NoS;
	string strAL = FOFN + "AL.txt";
	ofstream ofsAL(strAL.c_str());
	ofsAL << dAL << endl;
	ofsAL.close();

	// Calculate DIC
	double dALdic = 0.0;
	for (i = BurnIn; i < TotalSweep; i ++)
	{
		dALdic += LH[i];
	}
	dALdic /= KeptSweep;

	// Calculate variance of Deviance
	varDeviance /=KeptSweep;
	varDeviance = varDeviance - 4.0*dALdic*dALdic;

	double tmpDIC = 0.0;
	double tmpDIC2 = 0.0;
	double tmp = 0.0;
	for (i = 0; i < N; i++)
	{
		for (a = 0; a < A; a++)
		{
			for (l = 0; l < L; l++)
			{
				for (j = 0; j < J[l]; j++)
				{
					if (X[i][a][l] == alleleV[l][j])
					{
						tmp = 0.0;
						for (k =0; k < K; k ++)
						{
							tmp+=spAverage.p[k][l][j]*cpAverageAdmix[i][a][l][k];
						}
						tmpDIC += log(tmp);
						tmpDIC2 += log(spAverage.p[hcAdmix[i][a][l]][l][j]);
						break;
					}
				}
			}
		}
	}

	double pd = -2*dALdic + 2*tmpDIC;
	double pd2 = -2*dALdic + 2*tmpDIC2;

	double pd3 = (double)(0.5*varDeviance);

	double DIC = pd - 2*dALdic;
	double DIC2 = pd2 - 2*dALdic;

	double DIC3 = pd3 - 2*dALdic;

	string strPD = FOFN + "PD.txt";
	ofstream ofsPD(strPD.c_str());
	ofsPD << pd << endl;
	ofsPD.close();

	string strPD2 = FOFN + "PD2.txt";
	ofstream ofsPD2(strPD2.c_str());
	ofsPD2 << pd2 << endl;
	ofsPD2.close();

	string strDIC = FOFN + "DIC.txt";
	ofstream ofsDIC(strDIC.c_str());
	ofsDIC << DIC << endl;
	ofsDIC.close();

	string strDIC2 = FOFN + "DIC2.txt";
	ofstream ofsDIC2(strDIC2.c_str());
	ofsDIC2 << DIC2 << endl;
	ofsDIC2.close();
	
	string strDIC3 = FOFN + "DIC3.txt";
	ofstream ofsDIC3(strDIC2.c_str());
	ofsDIC3 << DIC3 << endl;
	ofsDIC3.close();


	//Store the min DIC on all runs
	//Restore the previous max likelihood
	string strML = FIFN + "MDIC_admCAR";
	if (LoadML(MLtrendDegree,minDIC,strML,true))
	{
		//A previous run has been saved. Have we done better ?
		if (DIC < minDIC)
		{
			//Yes, we did better. Saving the new DIC and the state reached by the markov chain.
			SaveML(trendDegree,DIC,strML);
			string strCF = FIFN + "CFadmCAR";
			SaveCF(spAverage.sigma, spAverage.beta, spAverage.alpha, spAverage.q, hcAdmix, spAverage.p,strCF);
		}
	}
	else
	{
		//No maxlikelihood has been stored until now. We save the DIC and the state reached by the markov chain.
		SaveML(trendDegree,DIC,strML);
		string strCF = FIFN + "CFadmCAR";
		SaveCF(spAverage.sigma, spAverage.beta,spAverage.alpha,spAverage.q,hcAdmix,spAverage.p,strCF);
	}

	FreeRNG(rng);

	clock_t e_clock = clock();

	cout << fixed;
	cout.precision(3);
	cout << "This run took " << static_cast<double>(e_clock - s_clock) /
	CLOCKS_PER_SEC << " seconds to complete on this computer." << endl<< endl<< endl;

	//PREDICTION ON A MAP
	if (!mapFile.empty())
	  {
	    cout << "Prediction on a map" << endl;
	    if (bX || bXL || (trendDegree==-1)) 
	      {
		cout << "Warning: POPS command line or gui do NOT predict on map if there are covariates other than spatial in the data file"<< endl;
		cout << "Option -m" << mapFile << "ignored" <<endl;
		cout << "Use R scripts provided along with the software to forecast membership/admixture coefficients on maps" << endl;
	      }
	    else
	      {

		s_clock = clock();
		cout << "Prediction started ..." << endl;

		DBL_VCTR_2D map;
		int ncol,nrow;                            //size of map stored in mapFile
		double noData;                            //Value of missing data in the ascii raster file
		double XLL, YLL, CellSize;
		//map has 2 columns: 1st: longitude, 2nd: Latitude
		LoadMap(map,noData,XLL,YLL,CellSize,nrow,ncol,mapFile);

		DBL_VCTR_3D prediction;
		InitVctr3D(prediction,K,nrow,ncol,0.0);


		predictFromBetaHat(prediction, map, spAverage.beta, noData);

		string predictStr[K];
		for (k = 0; k < K; k ++)
		  {
		    stringstream buffer;
		    buffer <<  FOFN << "Prediction_cluster_" << k+1 << ".txt";
		    predictStr[k] = buffer.str();
		    SavePredictionToMyFile(prediction[k],predictStr[k],noData, XLL, YLL, CellSize);
		  }

		e_clock = clock();
		cout << fixed;
		cout.precision(3);
		cout << "Prediction took " << static_cast<double>(e_clock - s_clock) /
		  CLOCKS_PER_SEC << " seconds to complete on this computer." << endl;
	      }
	  }

	return 0;
}
